/*
** File: avr2.c  
** Project: AVR2 API library
** Purpose: OS-independent code for AVR2 API library
**
** (C) Copyright Alpha Data 2016
*/

#include <avr2.h>
#include <portability.h>
#include <avr2_platform.h>
#include "protocol.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static bool_t
bIsControlCharacter(
  uint8_t character)
{
  return (character == SOH || character == EOT || character == ESCAPE) ? TRUE : FALSE;
}

void
avr2ProtocolStartRxPacket(
  RxEscapeState* pRxEscapeState,
  const uint8_t* pCommand)
{
  pRxEscapeState->bSOHDetected = FALSE;
  /* First two bytes of expected response are computed from first two bytes
  ** of command simply by flipping bit 7 of 2nd command byte. */
  pRxEscapeState->expectedResponse[0] = pCommand[0];
  pRxEscapeState->expectedResponse[1] = pCommand[1] ^ 0x80U;
}

bool_t /* TRUE => packet has been completed */
avr2ProtocolUnescapeRxCharacter(
  RxEscapeState* pRxEscapeState,
  uint8_t rawCh,
  bool_t* pbValid,
  uint32_t* pPosition)
{
  uint8_t unescaped;

  *pbValid = FALSE;

  if (rawCh == SOH) {
    /* If we get SOH at any point, go back to beginning of packet */
    pRxEscapeState->bSOHDetected = TRUE;
    pRxEscapeState->bLastWasEscape = FALSE;
    pRxEscapeState->position = 0;
    return FALSE; /* Packet not yet completed */
  }

  if (pRxEscapeState->bSOHDetected) {
    switch (rawCh) {
    case ESCAPE:
      /* Next character will need to be inverted to obtain unescaped character. */
      pRxEscapeState->bLastWasEscape = TRUE;
      break;

    case EOT:
      /* Packet completed */
      return TRUE;

    default:
      if (pRxEscapeState->bLastWasEscape) {
        /* Previous character was ESCAPE, so we must invert this one to obtain unescaped character. */
        unescaped = ~rawCh;
        pRxEscapeState->bLastWasEscape = FALSE;
      } else {
        unescaped = rawCh;
      }
      pRxEscapeState->unescapedCh = unescaped;

      if ((pRxEscapeState->position == 0 && unescaped != pRxEscapeState->expectedResponse[0]) ||
          (pRxEscapeState->position == 1 && unescaped != pRxEscapeState->expectedResponse[1]))
      {
        /* First two bytes of response are inconsistent with the command. Go back to waiting for SOH. */
        pRxEscapeState->bSOHDetected = FALSE;
        return FALSE; /* Packet not yet completed */
      }

      *pbValid = TRUE;
      *pPosition = pRxEscapeState->position++;
      break;
    }
  } else {
    /* Discard this character because SOH not received yet. */
  }

  return FALSE; /* Packet not yet completed */
}

void
avr2ProtocolStartTxPacket(
  TxEscapeState* pTxEscapeState)
{
  pTxEscapeState->buffer[0] = SOH;
  pTxEscapeState->nOccupied = 1;
  pTxEscapeState->bLastWasEscape = FALSE;
}

bool_t /* success => buffer full or buffer needs transmitted */
avr2ProtocolEscapeTxCharacter(
  TxEscapeState* pTxEscapeState,
  uint8_t character,
  bool_t *pbConsumed)
{
  switch (pTxEscapeState->bLastWasEscape) {
  case FALSE:
    if (bIsControlCharacter(character)) {
      pTxEscapeState->buffer[pTxEscapeState->nOccupied++] = 0x10U;
      pTxEscapeState->bLastWasEscape = TRUE;
      *pbConsumed = FALSE;
    } else {
      pTxEscapeState->buffer[pTxEscapeState->nOccupied++] = character;
      *pbConsumed = TRUE;
    }
    break;

  case TRUE:
    pTxEscapeState->buffer[pTxEscapeState->nOccupied++] = ~character;
    pTxEscapeState->bLastWasEscape = FALSE;
    *pbConsumed = TRUE;
    break;
  }

  return (ARRAY_LENGTH(pTxEscapeState->buffer) == pTxEscapeState->nOccupied) ? TRUE : FALSE;
}

void
avr2ProtocolEndTxPacket(
  TxEscapeState* pTxEscapeState)
{
  pTxEscapeState->buffer[pTxEscapeState->nOccupied++] = EOT;
}

AVR2_EXPORT AVR2_STATUS
AVR2_CALLING_CONVENTION
AVR2_Close(
  AVR2_HANDLE hDevice)
{
  return avr2Close(hDevice);
}

AVR2_EXPORT const char*
AVR2_CALLING_CONVENTION
AVR2_GetStatusStringA(
  AVR2_STATUS code,
  _AVR2_BOOL bShort)
{
  if (bShort) {
    switch (code) {
    case AVR2_SUCCESS:
      return "AVR2_SUCCESS";

    case AVR2_INTERNAL_ERROR:
      return "AVR2_INTERNAL_ERROR";

    case AVR2_UNEXPECTED_ERROR:
      return "AVR2_UNEXPECTED_ERROR";

    case AVR2_BAD_DRIVER:
      return "AVR2_BAD_DRIVER";

    case AVR2_NO_MEMORY:
      return "AVR2_NO_MEMORY";

    case AVR2_ACCESS_DENIED:
      return "AVR2_ACCESS_DENIED";

    case AVR2_DEVICE_NOT_FOUND:
      return "AVR2_DEVICE_NOT_FOUND";

    case AVR2_HARDWARE_ERROR:
      return "AVR2_HARDWARE_ERROR";

    case AVR2_INVALID_BUFFER:
      return "AVR2_INVALID_BUFFER";

    case AVR2_INVALID_FLAG:
      return "AVR2_INVALID_FLAG";

    case AVR2_INVALID_HANDLE:
      return "AVR2_INVALID_HANDLE";

    case AVR2_INVALID_INDEX:
      return "AVR2_INVALID_INDEX";

    case AVR2_INVALID_REGION:
      return "AVR2_INVALID_REGION";

    case AVR2_NULL_POINTER:
      return "AVR2_NULL_POINTER";

    case AVR2_CANCELLED:
      return "AVR2_CANCELLED";

    case AVR2_RESOURCE_LIMIT:
      return "AVR2_RESOURCE_LIMIT";

    case AVR2_REGION_TOO_LARGE:
      return "AVR2_REGION_TOO_LARGE";

    case AVR2_WRONG_MODE:
      return "AVR2_WRONG_MODE";

    case AVR2_NOT_SUPPORTED:
      return "AVR2_NOT_SUPPORTED";

    case AVR2_INVALID_PARAMETER:
      return "AVR2_INVALID_PARAMETER";

    case AVR2_TIMEOUT:
      return "AVR2_TIMEOUT";

    case AVR2_SHORT_COMMAND:
      return "AVR2_SHORT_COMMAND";

    default:
      return "AVR2_?";
    }
  } else {
    switch (code) {
    case AVR2_SUCCESS:
      return "No error";

    case AVR2_INTERNAL_ERROR:
      return "An error in the API logic was detected";

    case AVR2_UNEXPECTED_ERROR:
      return "An unexpected error caused the operation to fail";

    case AVR2_BAD_DRIVER:
      return "The driver may not be correctly installed";

    case AVR2_NO_MEMORY:
      return "Couldn't allocate memory required to complete operation";

    case AVR2_ACCESS_DENIED:
      return "The calling process does not have permission to perform the operation";

    case AVR2_DEVICE_NOT_FOUND:
      return "Failed to open the device with the specified index";

    case AVR2_HARDWARE_ERROR:
      return "An error in the hardware was detected";

    case AVR2_INVALID_BUFFER:
      return "The supplied buffer was invalid and could not be read/written";

    case AVR2_INVALID_FLAG:
      return "A flag was invalid or not recognized";

    case AVR2_INVALID_HANDLE:
      return "The device handle was invalid";

    case AVR2_INVALID_INDEX:
      return "The index parameter was invalid";

    case AVR2_INVALID_REGION:
      return "The offset and/or length parameters were invalid";

    case AVR2_NULL_POINTER:
      return "A NULL pointer was passed where non-NULL was required";

    case AVR2_CANCELLED:
      return "The operation was cancelled";

    case AVR2_RESOURCE_LIMIT:
      return "A resource limit was reached";

    case AVR2_REGION_TOO_LARGE:
      return "The specified region was too large for a single operation";

    case AVR2_WRONG_MODE:
      return "The operation was attempted when the device was not in the proper mode";

    case AVR2_NOT_SUPPORTED:
      return "The operation is not supported for this device";

    case AVR2_INVALID_PARAMETER:
      return "A parameter had an illegal value";

    case AVR2_TIMEOUT:
      return "The operation did not complete within the timeout period";

    case AVR2_SHORT_COMMAND:
      return "The length of the command was less than 2 bytes";

    case AVR2_SET_BAUD_FAILED:
      return "Failed to set the serial port to the requested baud rate";

    default:
      return "Unrecognized AVR2_STATUS code";
    }
  }
}

AVR2_EXPORT AVR2_STATUS
AVR2_CALLING_CONVENTION
AVR2_OpenPortA(
  const char* pPortName,
  AVR2_HANDLE* phDevice)
{
  return avr2OpenPortA(pPortName, phDevice);
}
