/*
** avr2util.c - utility to manipulate AVR2 microcontroller on certain reconfigurable
**              computing cards, at a low level.
**
** (C) Copyright 2016 - 2023 Alpha Data
**
** Supported models:
**
**   o ADM-AD01474
**   o ADM-PA100 / ADM-PA101
**   o ADM-SDEV-BASE
**   o ADM-PCIE-8K5 / ADM-PCIE-8K5-FH / ADM-PCIE-UCC
**   o ADM-PCIE-8V3
**   o ADM-PCIE-9H3
**   o ADM-PCIE-9H7
**   o ADM-PCIE-9V3 / ADM-PCIE-9V3G
**   o ADM-PCIE-9V5
**   o ADM-PCIE-9V6
**   o ADM-PCIE-9V7
**   o ADM-PCIE-9V8
**   o ADM-PCIE-TRC01
**   o ADM-VPX3-9V2
**   o ADM-VPX3-9Z2
**   o ADC-XMC-STANDALONE
**   o ADM-XRC-9R1
**   o ADM-XRC-9Z1
**   o ADM-XRC-KU1 / ADM-XRC-KU1-T8
**
** Exit codes: see EXIT_* #defines in exit_code.h
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tscanf_s scanf
# define _tcschr(s, c) strchr(s, c)
# define _tcslen(s) strlen(s)
# define _tcscpy_s(x, n, y) do { strncpy(x, y, n); (x)[(n) - 1] = '\0'; } while (0)
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# define _tcsstr(x, y) strstr(x, y)
# define _tmain main
# define _tprintf printf
# define _stscanf_s sscanf
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif
#include <assert.h>
#include <ctype.h>

#include <sdk_common.h>
#include <platform_endian.h>
#include <platform_time.h>

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#include "avr2util_version.h"
#include "boot_mode.h"
#include "brd_power.h"
#include "clobber.h"
#include "clock.h"
#include "clock_regmap.h"
#include "comm_test.h"
#include "connection.h"
#include "exit_code.h"
#include "firmware.h"
#include "fpga_info.h"
#include "i2c.h"
#include "mez.h"
#include "misc.h"
#include "model_info.h"
#include "product_id.h"
#include "sensor.h"
#include "smbus.h"
#include "spi_flash.h"
#include "srv_mode.h"
#include "fw_version.h"
#include "vpd.h"

#if defined(UPDATE_9V3G_U32)
# include "../update_9v3g_u32/update_9v3g_u32.h"
#endif

#if !defined(PROGRAM_NAME)
# ifdef _WIN32
#   define PROGRAM_NAME _T("avr2util.exe")
# else
#   if defined(AVR2_LIBRARY_STATIC)
#     define PROGRAM_NAME _T("avr2util-s")
#   else
#     define PROGRAM_NAME _T("avr2util")
#   endif /* defined(ADB3_LIBRARY_STATIC) */
# endif
#endif

/* Models that this utility knows about */
static struct _SupportedModels {
  const TCHAR* pszModelName;
} g_supportedModels[] = {
  { _T("ADM-AD01474") },
  { _T("ADM-PA100") },
  { _T("ADM-PA101") },
  { _T("ADM-PCIE-8V3") },
  { _T("ADM-PCIE-8K5") },
  { _T("ADM-PCIE-8K5-FH") },
  { _T("ADM-PCIE-UCC") },
  { _T("ADM-PCIE-9H3") },
  { _T("ADM-PCIE-9H7") },
  { _T("ADM-PCIE-9V3") },
  { _T("ADM-PCIE-9V3G") },
  { _T("ADM-PCIE-9V5") },
  { _T("ADM-PCIE-9V6") },
  { _T("ADM-PCIE-9V7") },
  { _T("ADM-PCIE-9V8") },
  { _T("ADM-PCIE-TRC01") },
  { _T("ADM-SDEV-BASE") },
  { _T("ADM-VPX3-9V2") },
  { _T("ADM-VPX3-9Z2") },
  { _T("ADC-XMC-STANDALONE") },
  { _T("ADM-XRC-9R1") },
  { _T("ADM-XRC-9Z1") },
  { _T("ADM-XRC-KU1") },
  { _T("ADM-XRC-KU1A-T8") }
};

static void
displaySupportedModels(
  void)
{
  unsigned int i;

  _tprintf(_T("Supported models:"));
  for (i = 0; i < ARRAY_LENGTH(g_supportedModels); i++) {
    _tprintf(_T(" %s"), g_supportedModels[i].pszModelName);
  }
  _tprintf(_T("\n"));
}

#if !(defined(__VXWORKS__) || defined(__vxworks))

static int stringToUInt(
  const TCHAR* pString,
  unsigned int* pValue)
{
  if (_tcslen(pString) >= 2 && pString[0] == _T('0') && pString[1] == _T('x')) {
    return _stscanf_s(pString + 2, _T("%x"), pValue);
  } else {
    return _stscanf_s(pString, _T("%u"), pValue);
  }
}

static int
stringToInt(
  const TCHAR* pString,
  int* pValue)
{
  bool_t bNegate = FALSE;
  unsigned int value;
  int nConverted;

  if (_tcslen(pString) >= 1) {
    if (pString[0] == _T('-')) {
      bNegate = TRUE;
      pString++;
    } else if (pString[0] == _T('+')) {
      pString++;
    } 
  }
  if (_tcslen(pString) >= 2 && pString[0] == _T('0') && pString[1] == _T('x')) {
    nConverted = _stscanf_s(pString + 2, _T("%x"), &value);
  } else {
    nConverted = _stscanf_s(pString, _T("%u"), &value);
  }
  if (nConverted >= 1) {
    if (bNegate) {
      *pValue = -(int)value;
    } else {
      *pValue = (int)value;
    }
  }
  return nConverted;
}

#if 0
static int
stringToFloat(
  const TCHAR* pString,
  float* pfValue)
{
  float fTmp = 0.0;
  int nConverted;

  nConverted = _stscanf_s(pString, _T("%f"), &fTmp);

  *pfValue = fTmp;

  return nConverted;
}
#endif

static int
stringToULong(
  const TCHAR* pString,
  unsigned long* pValue)
{
  if (_tcslen(pString) >= 2 && pString[0] == _T('0') && pString[1] == _T('x')) {
    return _stscanf_s(pString + 2, _T("%lx"), pValue);
  } else {
    return _stscanf_s(pString, _T("%lu"), pValue);
  }
}

#endif

#if (defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/

#define FLAG_BYSERIAL  (0x1U)   /* Interpret 'indexOrSerial' as a serial number */
#define FLAG_VERBOSE   (0x2U)   /* Verbose operation */
#define FLAG_CONFIRM   (0x100U) /* Don't prompt for confirmation before potentially dangerous operation */

static void
decodeFlags(
  unsigned int flags,
  bool_t* pbOpenBySerial,
  bool_t* pbVerbose)
{
  if (NULL != pbOpenBySerial) {
    *pbOpenBySerial = (flags & FLAG_BYSERIAL) ? TRUE : FALSE;
  }
  if (NULL != pbVerbose) {
    *pbVerbose = (flags & FLAG_VERBOSE) ? TRUE : FALSE;
  }
}

int
avr2utilHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  avr2utilBuildInfo()\n"));
  _tprintf(_T("  avr2utilHelp()\n"));
  _tprintf(_T("  avr2utilVersion(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilProductID(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilCommTest(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilEnterServiceMode(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilExitServiceMode(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilGetSMBUSAddr(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilGetSMBUSAddrNV(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilSetSMBUSAddr(indexOrSerial, flags, chipAddr7Bit)\n"));
  _tprintf(_T("  avr2utilSetSMBUSAddrNV(indexOrSerial, flags, chipAddr7Bit)\n"));
  _tprintf(_T("  avr2utilGetClk(indexOrSerial, flags, clockIndex)\n"));
  _tprintf(_T("  avr2utilSetClk(indexOrSerial, flags, clockIndex, frequencyHz)\n"));
  _tprintf(_T("  avr2utilSetClkRegMap(indexOrSerial, flags, clockIndex[, regMapFilename[, regMapTypeStr]])\n"));
  _tprintf(_T("    Register map types: lmk61e2\n"));
  _tprintf(_T("  avr2utilGetClkNV(indexOrSerial, flags, clockIndex)\n"));
  _tprintf(_T("  avr2utilSetClkNV(indexOrSerial, flags, clockIndex, frequencyHz)\n"));
  _tprintf(_T("  avr2utilSetClkNVRegMap(indexOrSerial, flags, clockIndex[, regMapFilename[, regMapTypeStr]])\n"));
  _tprintf(_T("  avr2utilI2cRead(indexOrSerial, flags, bus, device, address, count)\n"));
  _tprintf(_T("  avr2utilI2cReadToFile(indexOrSerial, flags, bus, device, address, count, pOutFilename)\n"));
  _tprintf(_T("  avr2utilI2cWrite(indexOrSerial, flags, bus, device, address, count, ...)\n"));
  _tprintf(_T("  avr2utilI2cWriteRead(indexOrSerial, flags, bus, device, writeCount, readCount, ...)\n"));
  _tprintf(_T("  avr2utilI2cWriteFromFile(indexOrSerial, flags, bus, device, address, pInFilename)\n"));
  _tprintf(_T("  avr2utilI2cVerifyWithFile(indexOrSerial, flags, bus, device, address, pInFilename)\n"));
  _tprintf(_T("  avr2utilSPIInfo(indexOrSerial, flags, chipIndex)\n"));
  _tprintf(_T("  avr2utilSPIRaw(indexOrSerial, flags, chipIndex, readCount, writeCount, ...)\n"));
  _tprintf(_T("  avr2utilSPICheckBlank(indexOrSerial, flags, chipIndex, address, count)\n"));
  _tprintf(_T("  avr2utilSPIErase(indexOrSerial, flags, chipIndex, address, count)\n"));
  _tprintf(_T("  avr2utilSPIProgramMCS(indexOrSerial, flags, chipIndex, pInFilename, translation)\n"));
  _tprintf(_T("  avr2utilSPIVerifyMCS(indexOrSerial, flags, chipIndex, pInFilename, translation)\n"));
  _tprintf(_T("  avr2utilSPISaveToMCS(indexOrSerial, flags, chipIndex, pOutFilename, address, count)\n"));
  _tprintf(_T("  avr2utilUpdateBrdCfg(indexOrSerial, flags, pBrdCfgFile)\n"));
  _tprintf(_T("  avr2utilUpdateFirmware(indexOrSerial, flags, pFirmwareFile)\n"));
  _tprintf(_T("  avr2utilUpdateVPD(indexOrSerial, flags, pVPDFile)\n"));
  _tprintf(_T("  avr2utilSaveBrdCfg(indexOrSerial, flags, pBrdCfgSaveFile)\n"));
  _tprintf(_T("  avr2utilSaveFirmware(indexOrSerial, flags, pFirmwareSaveFile)\n"));
  _tprintf(_T("  avr2utilSaveVPD(indexOrSerial, flags, pVPDSaveFile)\n"));
  _tprintf(_T("  avr2utilVerifyBrdCfg(indexOrSerial, flags, pBrdCfgFile)\n"));
  _tprintf(_T("  avr2utilVerifyFirmware(indexOrSerial, flags, pFirmwareFile)\n"));
  _tprintf(_T("  avr2utilVerifyVPD(indexOrSerial, flags, pVPDFile)\n"));
  _tprintf(_T("  avr2utilDisplayVPD(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilDisplayVPDRaw(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilDisplayBlackBox(indexOrSerial, flags[, types[, max]])\n"));
  _tprintf(_T("  avr2utilDisplaySensors(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilDisplaySensorsRaw(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilOverrideSensor(indexOrSerial, flags, sensorIndex, overrideValue)\n"));
  _tprintf(_T("  avr2utilReleaseSensor(indexOrSerial, flags, sensorIndex)\n"));
  _tprintf(_T("  avr2utilTestClobber(indexOrSerial, flags, sensorIndex)\n"));
  _tprintf(_T("  avr2utilTestTimeout(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilSetBrdPower(indexOrSerial, flags, bOnNotOff)\n"));
  _tprintf(_T("  avr2utilGetBrdPower(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilGetMezPower(indexOrSerial, flags, mezIndex)\n"));
  _tprintf(_T("  avr2utilGetMezPowerNV(indexOrSerial, flags, mezIndex)\n"));
  _tprintf(_T("  avr2utilForceMezPower(indexOrSerial, flags, reqMillivolt, minMillivolt, maxMillivolt, mezIndex)\n"));
  _tprintf(_T("  avr2utilForceMezPowerNV(indexOrSerial, flags, reqMillivolt, minMillivolt, maxMillivolt, mezIndex)\n"));
  _tprintf(_T("    !!! Incorrect use of Force Mezzanine Power verbs may damage hardware - use with care. !!!\n"));
  _tprintf(_T("  avr2utilSetBootMode(indexOrSerial, flags, pBootModeStr)\n"));
  _tprintf(_T("    ZynqMP boot modes: switch, 0, 1, ..., 15, jtag, qspi24, qspi32, sd0, nand, sd1_v2, emmc, usb0, pjtag0, pjtag1, sd1_v3\n"));
  _tprintf(_T("    Versal boot modes: switch, 0, 1, ..., 15, jtag, qspi24, qspi32, sd0_v3, sd1_v2, emmc1, ospi, selectmap, sd1_v3\n"));
  _tprintf(_T("                       \"switch\" means use switch setting on board\n"));
  _tprintf(_T("  avr2utilGetBootMode(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilSetBootModeNV(indexOrSerial, flags, pBootModeStr)\n"));
  _tprintf(_T("  avr2utilGetBootModeNV(indexOrSerial, flags)\n"));
  _tprintf(_T("  avr2utilRaw(indexOrSerial, flags, readCount, writeCount, ...)\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL 0x%X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_VERBOSE  0x%X => Verbose output\n"), FLAG_VERBOSE);
  _tprintf(_T("  FLAG_CONFIRM  0x%X => Don't prompt for confirmation before potentially dangerous operation\n"), FLAG_CONFIRM);
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
avr2utilBuildInfo(
  void)
{
  _tprintf(_T("%s %d.%d.%d - Microcontroller maintenance tool\n"), PROGRAM_NAME, AVR2UTIL_MAJOR, AVR2UTIL_MINOR, AVR2UTIL_BUGFIX);
  avr2utilDisplayBuildInfo();
  displaySupportedModels();
  return 0;
}

int
avr2utilProductID(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);

  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilProductIdCommand(pDevice, bVerbose, FALSE, NULL);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilVersion(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilVersionCommand(pDevice, bVerbose, FALSE, NULL);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

#if defined(AVR2UTIL_VXW_EXPERIMENTAL)

int
avr2utilUSBCOMVersion(
  const char* pszDevice,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bVerbose;
  int ret;
  
  decodeFlags(flags, NULL, &bVerbose);
  
  ret = avr2utilOpenDeviceUSBCom(pszDevice, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilVersionCommand(pDevice, bVerbose, FALSE, NULL);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

#endif

int
avr2utilCommTest(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilCommTestCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilEnterServiceMode(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("Nothing to do.\n"));
    goto done;
  }

  ret = avr2utilEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilExitServiceMode(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (!avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("Nothing to do.\n"));
    goto done;
  }

  ret = avr2utilEnterExitServiceMode(pDevice, bVerbose, FALSE, FALSE);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilGetSMBUSAddr(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetSMBUSChipAddrCommand(pDevice, bVerbose, FALSE);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilGetSMBUSAddrNV(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetSMBUSChipAddrCommand(pDevice, bVerbose, TRUE);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSetSMBUSAddr(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipAddr7Bit)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetSMBUSChipAddrCommand(pDevice, bVerbose, FALSE, chipAddr7Bit);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSetSMBUSAddrNV(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipAddr7Bit)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetSMBUSChipAddrCommand(pDevice, bVerbose, TRUE, chipAddr7Bit);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilGetClk(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int clockIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetClkCommand(pDevice, bVerbose, clockIndex);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSetClk(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int clockIndex,
  unsigned int frequencyHz)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetClkCommand(pDevice, bVerbose, clockIndex, (uint32_t)frequencyHz);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSetClkRegMap(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int clockIndex,
  const char* pRegMapFilename,
  const char* pRegMapTypeStr) /* Optional, may be NULL */
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetClkRegMapCommand(pDevice, bVerbose, clockIndex, PROGRAM_REGMAP_VOLATILE, pRegMapFilename, pRegMapTypeStr);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilGetClkNV(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int clockIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetClkNvCommand(pDevice, bVerbose, clockIndex);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSetClkNV(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int clockIndex,
  unsigned int frequencyHz)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetClkNvCommand(pDevice, bVerbose, clockIndex, (uint32_t)frequencyHz);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSetClkNVRegMap(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int clockIndex,
  const char* pRegMapFilename,
  const char* pRegMapTypeStr) /* Optional, may be NULL */
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetClkRegMapCommand(pDevice, bVerbose, clockIndex, PROGRAM_REGMAP_NONVOLATILE, pRegMapFilename, pRegMapTypeStr);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilI2cReadToFile(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  unsigned int count,
  const TCHAR* pOutFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilI2cReadToFileCommand(pDevice, bVerbose, bus, dev, address, count, pOutFilename);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilI2cWriteFromFile(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  const TCHAR* pInFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilI2cWriteFromFileCommand(pDevice, bVerbose, bus, dev, address, pInFilename);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilI2cVerifyWithFile(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  const TCHAR* pInFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilI2cVerifyWithFileCommand(pDevice, bVerbose, bus, dev, address, pInFilename);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilI2cRead(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  unsigned int count)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilI2cReadCommand(pDevice, bVerbose, bus, dev, address, count);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilI2cWrite(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  unsigned int count,
  ...)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  va_list args;
  uint8_t* pBytes = NULL;
  unsigned int i;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  pBytes = (uint8_t*)malloc(sizeof(uint8_t) * count);
  if (NULL == pBytes) {
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  va_start(args, count);
  for (i = 0; i < count; i++) {
    pBytes[i] = (uint8_t)va_arg(args, unsigned int);
  }
  va_end(args);

  ret = avr2utilI2cWriteCommand(pDevice, bVerbose, bus, dev, address, count, pBytes);

done:
  if (NULL != pBytes) {
    free(pBytes);
  }
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilI2cWriteRead(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int bus,
  unsigned int dev,
  unsigned int writeCount,
  unsigned int readCount,
  ...)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  va_list args;
  uint8_t* pWriteBytes = NULL;
  unsigned int i;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  pWriteBytes = (uint8_t*)malloc(sizeof(uint8_t) * writeCount);
  if (NULL == pWriteBytes) {
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  va_start(args, readCount);
  for (i = 0; i < writeCount; i++) {
    pWriteBytes[i] = (uint8_t)va_arg(args, unsigned int);
  }
  va_end(args);

  ret = avr2utilI2cWriteReadCommand(pDevice, bVerbose, bus, dev, writeCount, readCount, pWriteBytes);

done:
  if (NULL != pWriteBytes) {
    free(pWriteBytes);
  }
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPIRaw(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex,
  unsigned int readCount,
  unsigned int writeCount,
  ... /* Write bytes */)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  va_list args;
  uint8_t* pWriteBytes = NULL;
  unsigned int i;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  pWriteBytes = (uint8_t*)malloc(sizeof(uint8_t) * writeCount);
  if (NULL == pWriteBytes) {
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  va_start(args, writeCount);
  for (i = 0; i < writeCount; i++) {
    pWriteBytes[i] = (uint8_t)va_arg(args, unsigned int);
  }
  va_end(args);

  ret = avr2utilSpiRawCommand(pDevice, bVerbose, chipIndex, readCount, writeCount, pWriteBytes);

done:
  if (NULL != pWriteBytes) {
    free(pWriteBytes);
  }
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPIInfo(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSpiInfoCommand(pDevice, bVerbose, chipIndex);

done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPICheckBlank(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex,
  unsigned int address,
  unsigned int count)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSpiCheckBlankCommand(pDevice, bVerbose, chipIndex, address, count);

done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPIErase(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex,
  unsigned int address,
  unsigned int count)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSpiEraseCommand(pDevice, bVerbose, chipIndex, address, count);

done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPIProgramMCS(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex,
  const TCHAR* pInFilename,
  int translation)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSpiProgramIntelHexCommand(pDevice, bVerbose, chipIndex, pInFilename, translation);

done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPIVerifyMCS(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex,
  const TCHAR* pInFilename,
  int translation)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSpiVerifyIntelHexCommand(pDevice, bVerbose, chipIndex, pInFilename, translation);

done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSPISaveToMCS(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int chipIndex,
  const TCHAR* pOutFilename,
  unsigned int address,
  unsigned int count)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSpiSaveIntelHexCommand(pDevice, bVerbose, chipIndex, pOutFilename, address, count);

done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilUpdateFirmware(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pFilename) {
    _tprintf(_T("avr2utilUpdateFirmware requires the firmware filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilUpdateOrVerifyFirmwareCommand(pDevice, bVerbose, FALSE, pFilename);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilUpdateBrdCfg(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pFilename) {
    _tprintf(_T("avr2utilUpdateBrdCfg requires the board config. filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilUpdateOrVerifyBoardConfigCommand(pDevice, bVerbose, FALSE, pFilename);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilUpdateVPD(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pFilename) {
    _tprintf(_T("avr2utilUpdateVPD requires the VPD filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilUpdateOrVerifyVpdCommand(pDevice, bVerbose, FALSE, pFilename);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilSaveFirmware(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pSaveFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pSaveFilename) {
    _tprintf(_T("avr2utilSaveFirmware requires the firmware save filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSaveFirmwareCommand(pDevice, bVerbose, pSaveFilename);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilSaveBrdCfg(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pSaveFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pSaveFilename) {
    _tprintf(_T("avr2utilSaveBrdCfg requires the board config. save filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSaveBoardConfigCommand(pDevice, bVerbose, pSaveFilename);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilSaveVPD(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pSaveFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pSaveFilename) {
    _tprintf(_T("avr2utilSaveVPD requires the VPD save filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSaveVpdCommand(pDevice, bVerbose, pSaveFilename);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilVerifyBrdCfg(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pFilename) {
    _tprintf(_T("avr2utilVerifyBrdCfg requires the board config. filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilUpdateOrVerifyBoardConfigCommand(pDevice, bVerbose, TRUE, pFilename);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilVerifyFirmware(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pFilename) {
    _tprintf(_T("avr2utilVerifyFirmware requires the firmware filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilUpdateOrVerifyFirmwareCommand(pDevice, bVerbose, TRUE, pFilename);
  
done:
  avr2utilCloseDevice(pDevice);  
  return ret;
}

int
avr2utilVerifyVPD(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pFilename)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  if (NULL == pFilename) {
    _tprintf(_T("avr2utilVerifyVPD requires the VPD filename (.bin) as the first argument.\n"));
    ret = EXIT_INSUFFICIENT_ARGS;
    goto done;
  }
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilUpdateOrVerifyVpdCommand(pDevice, bVerbose, TRUE, pFilename);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilDisplayVPD(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilDisplayVpdCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilDisplayVPDRaw(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilDisplayVpdRawCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilDisplaySensors(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilDisplaySensorsCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilDisplaySensorsRaw(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilDisplaySensorsRawCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilOverrideSensor(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int sensorIndex,
  unsigned int overrideValue)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilOverrideSensorUnscaledCommand(pDevice, bVerbose, sensorIndex, overrideValue);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilReleaseSensor(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int sensorIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilReleaseSensorCommand(pDevice, bVerbose, sensorIndex);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilTestClobber(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilTestClobberCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilTestTimeout(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilTestTimeoutCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilDisplayBlackBox(
  unsigned int indexOrSerial,
  unsigned int flags,
  char* pTypes,
  unsigned int nRecordCount) /* 0 => all records, other => number of records to display */
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  unsigned int nTypes = 0x3; /* 0x1 => clobber records, 0x2 => noclobber records, 0x3 => both */
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  if (pTypes == NULL) {
    pTypes = "both";
  }

  if (_tcsicmp(pTypes, "both") == 0) {
    nTypes = 0x3;
  } else if (_tcsicmp(pTypes, "clobber") == 0) {
    nTypes = 0x1;
  } else if (_tcsicmp(pTypes, "noclobber") == 0) {
    nTypes = 0x2;
  } else {
      _tprintf(_T("*** Valid Black Box record types are: both, clobber, noclobber.\n"));
      ret = EXIT_INVALID_RECORD_TYPE;
      goto done;
  }

  if (0 == nRecordCount) {
    nRecordCount = 0xffffffff;
  }

  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilDisplayBlackBoxCommand(pDevice, bVerbose, nTypes, nRecordCount);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilSetBrdPower(
  unsigned int indexOrSerial,
  unsigned int flags,
  int bOnNotOff)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetBrdPowerCommand(pDevice, bVerbose, bOnNotOff ? TRUE : FALSE);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilGetBrdPower(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetBrdPowerCommand(pDevice, bVerbose);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilGetMezPowerPower(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int nMezIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetMezPowerCommand(pDevice, bVerbose, FALSE, nMezIndex);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilGetMezPowerPowerNV(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int nMezIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetMezPowerCommand(pDevice, bVerbose, TRUE, nMezIndex);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilForceMezPowerPower(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int nReqMillivolt,
  unsigned int nMinMillivolt,
  unsigned int nMaxMillivolt,
  unsigned int nMezIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose, bConfirmed = FALSE;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (flags & FLAG_CONFIRM) {
    bConfirmed = TRUE;
  }

  ret = avr2utilForceMezPowerCommand(pDevice, bVerbose, FALSE, bConfirmed, nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilForceMezPowerPowerNV(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int nReqMillivolt,
  unsigned int nMinMillivolt,
  unsigned int nMaxMillivolt,
  unsigned int nMezIndex)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose, bConfirmed = FALSE;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (flags & FLAG_CONFIRM) {
    bConfirmed = TRUE;
  }

  ret = avr2utilForceMezPowerCommand(pDevice, bVerbose, TRUE, bConfirmed, nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilSetBootMode(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pBootModeStr)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetBootModeCommand(pDevice, bVerbose, FALSE, pBootModeStr);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilGetBootMode(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetBootModeCommand(pDevice, bVerbose, FALSE);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilSetBootModeNV(
  unsigned int indexOrSerial,
  unsigned int flags,
  const TCHAR* pBootModeStr)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilSetBootModeCommand(pDevice, bVerbose, TRUE, pBootModeStr);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilGetBootModeNV(
  unsigned int indexOrSerial,
  unsigned int flags)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilGetBootModeCommand(pDevice, bVerbose, TRUE);
  
done:
  avr2utilCloseDevice(pDevice);
  return ret;
}

int
avr2utilRaw(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int readCount,
  unsigned int writeCount,
  ... /* Command bytes */)
{
  DeviceHandle* pDevice = NULL;
  bool_t bOpenBySerial, bVerbose;
  va_list args;
  uint8_t* pWriteBytes = NULL;
  unsigned int i;
  int ret;
  
  decodeFlags(flags, &bOpenBySerial, &bVerbose);
  
  ret = avr2utilOpenDeviceADB3(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, bVerbose, &pDevice);
  if (EXIT_OK != ret) {
    goto done;
  }

  pWriteBytes = (uint8_t*)malloc(sizeof(uint8_t) * writeCount);
  if (NULL == pWriteBytes) {
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  va_start(args, writeCount);
  for (i = 0; i < writeCount; i++) {
    pWriteBytes[i] = (uint8_t)va_arg(args, unsigned int);
  }
  va_end(args);

  ret = avr2utilRawCommand(pDevice, bVerbose, readCount, writeCount, pWriteBytes);

done:
  if (NULL != pWriteBytes) {
    free(pWriteBytes);
  }
  avr2utilCloseDevice(pDevice);  
  return ret;
}

#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/

int _tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = EXIT_OK;
  DeviceHandle* pDevice = NULL;
  Arguments arguments = g_initArguments;
  Option options[11];
  OptionType paramType[1];
  TCHAR* paramInfo[1];
  TCHAR* paramMeaning[1];
  bool_t bIndexSpecified, bAdxdmaSpecified, bPsuartSpecified, bUsbcomSpecified;
  bool_t bSmbusSpecified, bIpmiSpecified, bSerialSpecified;
  unsigned int chipAddr, nOpenOptionSpecified = 0, timeoutMs;
  bool_t bVerbose, bConfirmed;
  TCHAR* pCommand;
  
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use (ADB3 Driver)");
  options[1].key = _T("adxdma");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("index of card to use (ADXDMA Driver)");
  options[2].key = _T("psuart");
  options[2].type = OptionString;
  options[2].def.pStringVal = _T("");
#ifdef _WIN32
  options[2].meaning = _T("Zynq UltraScale+ PS UART and baud rate to use, e.g. COM1,171428");
#else
  options[2].meaning = _T("Zynq UltraScale+ PS UART and baud rate to use, e.g. /dev/ttyPS1,171428");
#endif
  options[3].key = _T("usbcom");
  options[3].type = OptionString;
  options[3].def.pStringVal = _T("");
#ifdef _WIN32
  options[3].meaning = _T("USB serial port to use, e.g. COM1");
#else
  options[3].meaning = _T("USB serial port to use, e.g. /dev/ttyACM0");
#endif
  options[4].key = _T("smbus");
  options[4].type = OptionString;
  options[4].def.pStringVal = _T("");
#ifdef _WIN32
  options[4].meaning = _T("SMBUS device to use, e.g. ?");
#else
  options[4].meaning = _T("SMBUS device to use, e.g. /dev/i2c-0");
#endif
  options[5].key = _T("ipmi");
  options[5].type = OptionString;
  options[5].def.pStringVal = _T("");
  options[5].meaning = _T("Use IPMI; specifies IPMI options string");
  options[6].key = _T("sn");
  options[6].type = OptionUInt;
  options[6].def.uintVal = 0;
  options[6].meaning = _T("serial number of card to use");
  options[7].key = _T("verbose");
  options[7].type = OptionBoolean;
  options[7].def.booleanVal = 0;
  options[7].meaning = _T("TRUE => verbose output");
  options[8].key = _T("chipaddr");
  options[8].type = OptionHex;
  options[8].def.uintVal = UC_SMBUS_CHIP_ADDR;
  options[8].meaning = _T("SMBUS address of the AVR2 microcontroller (hexadecimal)");
  options[9].key = _T("timeout");
  options[9].type = OptionUInt;
  options[9].def.uintVal = 1000;
  options[9].meaning = _T("Timeout for AVR2 microcontroller commands (milliseconds)");
  options[10].key = _T("confirm");
  options[10].type = OptionBoolean;
  options[10].def.booleanVal = 0;
  options[10].meaning = _T("TRUE => confirm potentially dangerous operation");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.bNoStdOptions = TRUE;
   
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] =
    _T("One of:\n")
    _T("\n     build-info")
    _T("\n     product-id")
    _T("\n     version")
    _T("\n     comm-test")
    _T("\n     enter-service-mode")
    _T("\n     exit-service-mode")
    _T("\n     get-smbus-addr")
    _T("\n     set-smbus-addr <7-bit SMBUS address>")
    _T("\n     get-smbus-addr-nv")
    _T("\n     set-smbus-addr-nv <7-bit SMBUS address>")
    _T("\n     getclk <index>")
    _T("\n     setclk <index> <Hz>")
    _T("\n     setclk-regmap <index> [reg. map filename] [reg. map type]")
    _T("\n       Register map types: lmk61e2")
    _T("\n     getclknv <index>")
    _T("\n     setclknv <index> <Hz>")
    _T("\n     setclknv-regmap <index> [reg. map filename] [reg. map type]")
    _T("\n     set-boot-mode <boot mode | switch>")
    _T("\n     set-boot-mode-nv <boot mode | switch>")
    _T("\n       ZynqMP boot modes: switch, 0, 1, ..., 15, jtag, qspi24, qspi32, sd0,")
    _T("\n                          nand, sd1_v2, emmc, usb0, pjtag0, pjtag1, sd1_v3")
    _T("\n       Versal boot modes: 0, 1, ..., 15, jtag, qspi24, qspi32, sd0_v3, sd1_v2,")
    _T("\n                          emmc1, ospi, selectmap, sd1_v3")
    _T("\n                          \"switch\" means use switch setting on board")
    _T("\n     get-boot-mode")
    _T("\n     get-boot-mode-nv")
    _T("\n     i2c-read-to-file <bus> <device> <address> <count> <output file.bin>")
    _T("\n     i2c-verify-with-file <bus> <device> <address> <input file.bin>")
    _T("\n     i2c-write-from-file <bus> <device> <address> <input file.bin>")
    _T("\n     i2c-read <bus> <device> <address> <count>")
    _T("\n     i2c-write <bus> <device> <address> <data byte> ...")
    _T("\n     i2c-write-read <bus> <device> <read count> <write address/data byte> ...")
    _T("\n     spi-raw <chip index> <read count> <write data byte> ...")
    _T("\n     spi-info <chip index>")
    _T("\n     spi-check-blank <chip index> <address> <count>")
    _T("\n     spi-erase <chip index> <address> <count>")
    _T("\n     spi-program-mcs <chip index> <filename.mcs> [translation]")
    _T("\n     spi-verify-mcs <chip index> <filename.mcs> [translation]")
    _T("\n     spi-save-to-mcs <chip index> <filename.mcs> <address> <count>")
    _T("\n     update-brdcfg <board configuration file.bin>")
    _T("\n     save-brdcfg <board configuration save file.bin>")
    _T("\n     verify-brdcfg <board configuration file.bin>")
    _T("\n     update-firmware <firmware file.bin>")
    _T("\n     save-firmware <firmware save file.bin>")
    _T("\n     verify-firmware <firmware file.bin>")
    _T("\n     display-vpd")
    _T("\n     display-vpd-raw")
    _T("\n     update-vpd <VPD file.bin>")
    _T("\n     save-vpd <VPD save file.bin>")
    _T("\n     verify-vpd <VPD file.bin>")
    _T("\n     display-black-box [clobber|noclobber|both] [count]")
    _T("\n     display-sensors")
    _T("\n     display-sensors-raw")
    _T("\n     override-sensor <index> <unscaled value>")
    _T("\n     release-sensor <index>")
    _T("\n     test-clobber")
    _T("\n     test-timeout")
    _T("\n     set-brd-power <on|off>")
    _T("\n     get-brd-power")
    _T("\n     get-mez-power <index>")
    _T("\n     get-mez-power-nv <index>")
    _T("\n     force-mez-power <requested mV> <min mV> <max mV> <index>")
    _T("\n     force-mez-power-nv <requested mV> <min mV> <max mV> <index>")
    _T("\n       INCORRECT USE OF FORCE MEZZANINE POWER COMMANDS MAY DAMAGE HARDWARE. USE WITH CARE.")
    _T("\n     raw <expected response length> <command bytes ...>")
  ;

  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  if (sampleParseCommandLine(PROGRAM_NAME, argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }

  bIndexSpecified = options[0].specified;

  bAdxdmaSpecified = options[1].specified;

  bPsuartSpecified = options[2].specified;

  bUsbcomSpecified = options[3].specified;

  bSmbusSpecified = options[4].specified;

  bIpmiSpecified = options[5].specified;

  bSerialSpecified = options[6].specified;

  bVerbose = options[7].value.booleanVal;

  chipAddr = options[8].value.hexVal;
  if (0 == chipAddr || chipAddr >= 0xFF) {
    _tprintf(_T("*** Value for '-chipaddr' option must be in the range 0x1 to 0x7F inclusive.\n"));
    ret = EXIT_BAD_CHIP_ADDR;
    goto done;
  }

  bConfirmed = options[10].value.booleanVal;

  pCommand = arguments.param[0].pStringVal;

  /* Check that no more than one open option has been passed on the command line */
  if (bIndexSpecified) {
    nOpenOptionSpecified++;
  }
  if (bAdxdmaSpecified) {
    nOpenOptionSpecified++;
  }
  if (bPsuartSpecified) {
    nOpenOptionSpecified++;
  }
  if (bUsbcomSpecified) {
    nOpenOptionSpecified++;
  }
  if (bSmbusSpecified) {
    nOpenOptionSpecified++;
  }
  if (bIpmiSpecified) {
    nOpenOptionSpecified++;
  }
  if (nOpenOptionSpecified > 1) {
    _tprintf(_T("*** Too many open options specifed; at most of one '-index', '-adxdma', '-usbcom', '-smbus' and '-ipmi' can be passed on the command line.\n"));
    ret = EXIT_TOO_MANY_COMM_OPTIONS;
    goto done;
  }

  if (_tcsicmp(pCommand, _T("build-info")) == 0) {
    _tprintf(_T("%s %d.%d.%d - Microcontroller maintenance tool\n"), PROGRAM_NAME, AVR2UTIL_MAJOR, AVR2UTIL_MINOR, AVR2UTIL_BUGFIX);
    avr2utilDisplayBuildInfo();
    displaySupportedModels();
    goto done;
  }

  /* Open the device using the chosen method of opening */
  if (bAdxdmaSpecified) {
    ret = avr2utilOpenDeviceADXDMA(
      options[1].value.uintVal,
      bVerbose,
      &pDevice);
  } else if (bPsuartSpecified) {
    ret = avr2utilOpenDevicePsuart(
      options[2].value.pStringVal,
      bVerbose,
      &pDevice);
  } else if (bUsbcomSpecified) {
    ret = avr2utilOpenDeviceUSBCom(
      options[3].value.pStringVal,
      bVerbose,
      &pDevice);
  } else if (bSmbusSpecified) {
    ret = avr2utilOpenDeviceSMBUS(
      options[4].value.pStringVal,
      (uint8_t)chipAddr, /* Cast safe due to range check above */
      bVerbose,
      &pDevice);
  } else if (bIpmiSpecified) {
    ret = avr2utilOpenDeviceIPMI(
      options[5].value.pStringVal,
      bVerbose,
      &pDevice);
  } else {
    ret = avr2utilOpenDeviceADB3(
      options[0].value.uintVal,
      bIndexSpecified,
      options[6].value.uintVal,
      bSerialSpecified,
      bVerbose,
      &pDevice);
  }

  if (EXIT_OK != ret) {
    goto done;
  }

  avr2utilCheckEnvironment(pDevice);

  if (options[9].specified) {
    timeoutMs = options[9].value.uintVal;

    if (timeoutMs < 1 || timeoutMs > 3600000) {
      _tprintf(_T("*** Value for '-timeout' option must be in the range 1 .. 3600000.\n"));
      ret = EXIT_BAD_TIMEOUT_VALUE;
      goto done;
    }

    avr2utilSetCommandTimeoutUs(pDevice, timeoutMs * 1000);
  }

  _tprintf(_T("%s Service Mode\n"), avr2utilIsInServiceMode(pDevice) ? _T("In") : _T("Not in"));
  
  if (_tcsicmp(pCommand, _T("version")) == 0) {
    ret = avr2utilVersionCommand(pDevice, bVerbose, FALSE, NULL);
  } else if (_tcsicmp(pCommand, _T("product-id")) == 0) {
    ret = avr2utilProductIdCommand(pDevice, bVerbose, FALSE, NULL);
  } else if (_tcsicmp(pCommand, _T("comm-test")) == 0) {
    ret = avr2utilCommTestCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("enter-service-mode")) == 0) {
    if (avr2utilIsInServiceMode(pDevice)) {
      _tprintf(_T("Nothing to do.\n"));
    } else {
      ret = avr2utilEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
    }
  } else if (_tcsicmp(pCommand, _T("exit-service-mode")) == 0) {
    if (!avr2utilIsInServiceMode(pDevice)) {
      _tprintf(_T("Nothing to do.\n"));
    } else {
      ret = avr2utilEnterExitServiceMode(pDevice, bVerbose, FALSE, FALSE);
    }
  } else if (_tcsicmp(pCommand, _T("get-smbus-addr")) == 0 ||
             _tcsicmp(pCommand, _T("get-smbus-addr-nv")) == 0)
  {
    bool_t bNonvolatile = (_tcsicmp(pCommand, _T("get-smbus-addr-nv")) == 0) ? TRUE : FALSE;

    ret = avr2utilGetSMBUSChipAddrCommand(pDevice, bVerbose, bNonvolatile);
  } else if (_tcsicmp(pCommand, _T("set-smbus-addr")) == 0 ||
             _tcsicmp(pCommand, _T("set-smbus-addr-nv")) == 0)
  {
    bool_t bNonvolatile = (_tcsicmp(pCommand, _T("set-smbus-addr-nv")) == 0) ? TRUE : FALSE;
    unsigned int nChipAddress;
    int nConverted;

    if (arguments.nParam < 2) {
      _tprintf(_T("*** '%s' command must be followed by a 7-bit SMBUS chip address.\n"), pCommand);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &nChipAddress);
    if (1 != nConverted) {
      _tprintf(_T("*** SMBUS chip address value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    
    ret = avr2utilSetSMBUSChipAddrCommand(pDevice, bVerbose, bNonvolatile, nChipAddress);
  } else if (_tcsicmp(pCommand, _T("getclk")) == 0) {
    unsigned int clockIndex;
    int nConverted;
    
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'getclk' command must be followed by a clock generator index.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &clockIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    
    ret = avr2utilGetClkCommand(pDevice, bVerbose, clockIndex);
  } else if (_tcsicmp(pCommand, _T("setclk")) == 0) {
    const TCHAR* pszFrequency;
    unsigned int clockIndex;
    unsigned long frequency;
    int nConverted;
    
    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'setclk' command must be followed by (1) a clock generator index and (2) a frequency (Hz).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &clockIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }
    if (clockIndex > 0xFFU) {
      _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    pszFrequency = arguments.param[2].pStringVal;
    nConverted = stringToULong(pszFrequency, &frequency);
    if (1 != nConverted) {
      _tprintf(_T("*** Frequency value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_FREQUENCY;
      goto done;
    }
    if (frequency > 0xFFFFFFFFU) {
      _tprintf(_T("*** Frequency must be in the range 0 to 4294967295.\n"));
      ret = EXIT_INVALID_FREQUENCY;
      goto done;
    }
    
    ret = avr2utilSetClkCommand(pDevice, bVerbose, clockIndex, (uint32_t)frequency);
  } else if (_tcsicmp(pCommand, _T("getclknv")) == 0) {
    unsigned int clockIndex;
    int nConverted;
    
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'getclknv' command must be followed by a clock generator index.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &clockIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }
    
    ret = avr2utilGetClkNvCommand(pDevice, bVerbose, clockIndex);
  } else if (_tcsicmp(pCommand, _T("setclknv")) == 0) {
    const TCHAR* pszFrequency;
    unsigned int clockIndex;
    unsigned long frequency;
    int nConverted;
    
    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'setclknv' command must be followed by (1) a clock generator index and (2) an override frequency (Hz).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &clockIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }
    if (clockIndex > 0xFFU) {
      _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    pszFrequency = arguments.param[2].pStringVal;
    nConverted = stringToULong(pszFrequency, &frequency);
    if (1 != nConverted) {
      _tprintf(_T("*** Override frequency value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_FREQUENCY;
      goto done;
    }
    if (frequency > 0xFFFFFFFFU) {
      _tprintf(_T("*** Override frequency must be in the range 0 to 4294967295.\n"));
      ret = EXIT_INVALID_FREQUENCY;
      goto done;
    }
    
    ret = avr2utilSetClkNvCommand(pDevice, bVerbose, clockIndex, (uint32_t)frequency);
  } else if (_tcsicmp(pCommand, _T("setclk-regmap")) == 0 || _tcsicmp(pCommand, _T("setclknv-regmap")) == 0) {
    unsigned int nClockIndex;
    const TCHAR* pszRegMapPath = NULL;
    const TCHAR* pszRegMapType = NULL;
    uint32_t nFlags;
    bool_t bNonVolatile;
    int nConverted;
    
    bNonVolatile = (_tcsicmp(pCommand, _T("setclknv-regmap")) == 0);

    if (arguments.nParam < 2) {
      _tprintf(_T("*** '%s' command must be followed by (1) a clock generator index, (2) optional path to register map file and (3) optional register map file type.\n"),
        bNonVolatile ? _T("setclknv-regmap") : _T("setclk-regmap"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &nClockIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }
    if (nClockIndex > 0xFFU) {
      _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    if (arguments.nParam >= 3) {
      pszRegMapPath = arguments.param[2].pStringVal;
    }
    
    if (arguments.nParam >= 4) {
      pszRegMapType = arguments.param[3].pStringVal;
    }
    
    nFlags = bNonVolatile ? PROGRAM_REGMAP_NONVOLATILE : PROGRAM_REGMAP_VOLATILE;

    ret = avr2utilSetClkRegMapCommand(pDevice, bVerbose, (uint8_t)nClockIndex, nFlags, pszRegMapPath, pszRegMapType);
  } else if (_tcsicmp(pCommand, _T("setclk")) == 0) {
    const TCHAR* pszFrequency;
    unsigned int clockIndex;
    unsigned long frequency;
    int nConverted;
    
    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'setclk' command must be followed by (1) a clock generator index and (2) a frequency (Hz).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &clockIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }
    if (clockIndex > 0xFFU) {
      _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    pszFrequency = arguments.param[2].pStringVal;
    nConverted = stringToULong(pszFrequency, &frequency);
    if (1 != nConverted) {
      _tprintf(_T("*** Frequency value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_FREQUENCY;
      goto done;
    }
    if (frequency > 0xFFFFFFFFU) {
      _tprintf(_T("*** Frequency must be in the range 0 to 4294967295.\n"));
      ret = EXIT_INVALID_FREQUENCY;
      goto done;
    }
    
    ret = avr2utilSetClkCommand(pDevice, bVerbose, clockIndex, (uint32_t)frequency);  } else if (_tcsicmp(pCommand, _T("i2c-read-to-file")) == 0) {
    unsigned int bus, dev, address;
    unsigned int count;
    int nConverted;

    if (arguments.nParam < 6) {
      _tprintf(_T("*** 'i2c-read-to-file' command must be followed by (1) the I2C bus number, (2) the I2C chip address, (3) the address, (4) the byte count, and (5) the filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &bus);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C bus number value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_I2C_BUS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &dev);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C chip address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_I2C_DEVICE;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** Address value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_I2C_ADDRESS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[4].pStringVal, &count);
    if (1 != nConverted) {
      _tprintf(_T("*** Count value '%s' not recognized as a number.\n"), arguments.param[4].pStringVal);
      ret = EXIT_INVALID_I2C_COUNT;
      goto done;
    }

    ret = avr2utilI2cReadToFileCommand(pDevice, bVerbose, bus, dev, address, count, arguments.param[5].pStringVal);
  } else if (_tcsicmp(pCommand, _T("i2c-write-from-file")) == 0) {
    unsigned int bus, dev, address;
    int nConverted;

    if (arguments.nParam < 5) {
      _tprintf(_T("*** 'i2c-write-from-file' command must be followed by (1) the I2C bus number, (2) the I2C chip address, (3) the address and (4) the filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &bus);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C bus number value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_I2C_BUS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &dev);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C chip address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_I2C_DEVICE;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** Address value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_I2C_ADDRESS;
      goto done;
    }

    ret = avr2utilI2cWriteFromFileCommand(pDevice, bVerbose, bus, dev, address, arguments.param[4].pStringVal);
  } else if (_tcsicmp(pCommand, _T("i2c-verify-with-file")) == 0) {
    unsigned int bus, dev, address;
    int nConverted;

    if (arguments.nParam < 5) {
      _tprintf(_T("*** 'i2c-verify-with-file' command must be followed by (1) the I2C bus number, (2) the I2C chip address, (3) the address and (4) the filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &bus);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C bus number value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &dev);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C chip address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** Address value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    ret = avr2utilI2cVerifyWithFileCommand(pDevice, bVerbose, bus, dev, address, arguments.param[4].pStringVal);
  } else if (_tcsicmp(pCommand, _T("i2c-read")) == 0) {
    unsigned int bus, dev, address;
    unsigned int count;
    int nConverted;

    if (arguments.nParam < 5) {
      _tprintf(_T("*** 'i2c-read' command must be followed by (1) the I2C bus number, (2) the I2C chip address, (3) the address, (4) the byte count\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &bus);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C bus number value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_I2C_BUS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &dev);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C chip address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_I2C_DEVICE;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** Address value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_I2C_ADDRESS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[4].pStringVal, &count);
    if (1 != nConverted) {
      _tprintf(_T("*** Count value '%s' not recognized as a number.\n"), arguments.param[4].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    ret = avr2utilI2cReadCommand(pDevice, bVerbose, bus, dev, address, count);
  } else if (_tcsicmp(pCommand, _T("i2c-write")) == 0) {
    unsigned int bus, dev, address;
    unsigned int count, i, byte;
    int nConverted;
    uint8_t* pBytes = NULL;

    if (arguments.nParam < 5) {
      _tprintf(_T("*** 'i2c-write' command must be followed by (1) the I2C bus number, (2) the I2C chip address, (3) the address, (4) one or more address/data bytes to write\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &bus);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C bus number value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_I2C_BUS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &dev);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C chip address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_I2C_DEVICE;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** Address value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_I2C_ADDRESS;
      goto done;
    }

    count = arguments.nParam - 4;

    pBytes = (uint8_t*)malloc(count * sizeof(uint8_t));
    if (NULL == pBytes) {
      _tprintf(_T("*** Failed to allocate byte buffer of %u byte(s).\n"), count);
      ret = EXIT_ALLOCATION_FAILED;
      goto done;
    }

    for (i = 0; i < count; i++) {
      nConverted = stringToUInt(arguments.param[i + 4].pStringVal, &byte);
      if (1 != nConverted) {
        free(pBytes);
        _tprintf(_T("*** Byte value '%s' not recognized as a number.\n"), arguments.param[i + 4].pStringVal);
        ret = EXIT_INVALID_I2C_BYTEVAL;
        goto done;
      }
      pBytes[i] = (uint8_t)byte;
    }

    ret = avr2utilI2cWriteCommand(pDevice, bVerbose, bus, dev, address, count, pBytes);

    free(pBytes);
  } else if (_tcsicmp(pCommand, _T("i2c-write-read")) == 0) {
    unsigned int bus, dev;
    unsigned int readCount, writeCount, i, byte;
    int nConverted;
    uint8_t* pWriteBytes = NULL;

    if (arguments.nParam < 4) {
      _tprintf(_T("*** 'i2c-write-read' command must be followed by (1) the I2C bus number, (2) the I2C chip address, (3) read byte count, (4) zero or more address/data bytes to write\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &bus);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C bus number value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_I2C_BUS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &dev);
    if (1 != nConverted) {
      _tprintf(_T("*** I2C chip address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_I2C_DEVICE;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &readCount);
    if (1 != nConverted) {
      _tprintf(_T("*** Read byte count value '%s' not recognized as a number.\n"), arguments.param[4].pStringVal);
      ret = EXIT_INVALID_I2C_COUNT;
      goto done;
    }

    writeCount = arguments.nParam - 4;

    pWriteBytes = (uint8_t*)malloc(writeCount * sizeof(uint8_t));
    if (NULL == pWriteBytes) {
      _tprintf(_T("*** Failed to allocate write byte buffer of %u byte(s).\n"), writeCount);
      ret = EXIT_ALLOCATION_FAILED;
      goto done;
    }

    for (i = 0; i < writeCount; i++) {
      nConverted = stringToUInt(arguments.param[i + 4].pStringVal, &byte);
      if (1 != nConverted) {
        free(pWriteBytes);
        _tprintf(_T("*** Byte value '%s' not recognized as a number.\n"), arguments.param[i + 4].pStringVal);
        ret = EXIT_INVALID_I2C_BYTEVAL;
        goto done;
      }
      pWriteBytes[i] = (uint8_t)byte;
    }

    ret = avr2utilI2cWriteReadCommand(pDevice, bVerbose, bus, dev, writeCount, readCount, pWriteBytes);

    free(pWriteBytes);
  } else if (_tcsicmp(pCommand, _T("spi-raw")) == 0) {
    unsigned int chipIndex, readCount, writeCount;
    unsigned int i, byte;
    int nConverted;
    uint8_t* pWriteBytes = NULL;

    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'spi-raw' command must be followed by (1) the SPI chip index, (2) number of bytes to read, (3) zero or more bytes to write\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &readCount);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI read count value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_SPI_READ_COUNT;
      goto done;
    }

    writeCount = arguments.nParam - 3;

    pWriteBytes = (uint8_t*)malloc(writeCount * sizeof(uint8_t));
    if (NULL == pWriteBytes) {
      _tprintf(_T("*** Failed to allocate write buffer of %u byte(s).\n"), writeCount);
      ret = EXIT_ALLOCATION_FAILED;
      goto done;
    }

    for (i = 0; i < writeCount; i++) {
      nConverted = stringToUInt(arguments.param[i + 3].pStringVal, &byte);
      if (1 != nConverted) {
        free(pWriteBytes);
        _tprintf(_T("*** Byte value '%s' not recognized as a number.\n"), arguments.param[i + 3].pStringVal);
        ret = EXIT_INVALID_SPI_BYTEVAL;
        goto done;
      }
      pWriteBytes[i] = (uint8_t)byte;
    }

    ret = avr2utilSpiRawCommand(pDevice, bVerbose, chipIndex, readCount, writeCount, pWriteBytes);

    free(pWriteBytes);
  } else if (_tcsicmp(pCommand, _T("spi-info")) == 0) {
    unsigned int chipIndex;
    int nConverted;

    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'spi-info' command must be followed by (1) the SPI chip index.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    ret = avr2utilSpiInfoCommand(pDevice, bVerbose, chipIndex);
  } else if (_tcsicmp(pCommand, _T("spi-check-blank")) == 0) {
    unsigned int chipIndex, address, count;
    int nConverted;

    if (arguments.nParam < 4) {
      _tprintf(_T("*** 'spi-check-blank' command must be followed by (1) the SPI chip index, (2) starting address, (3) number of bytes to erase\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_SPI_ADDRESS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &count);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI blank-check count value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_SPI_BYTE_COUNT;
      goto done;
    }

    ret = avr2utilSpiCheckBlankCommand(pDevice, bVerbose, chipIndex, address, count);
  } else if (_tcsicmp(pCommand, _T("spi-erase")) == 0) {
    unsigned int chipIndex, address, count;
    int nConverted;

    if (arguments.nParam < 4) {
      _tprintf(_T("*** 'spi-erase' command must be followed by (1) the SPI chip index, (2) starting address, (3) number of bytes to erase\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI address value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_SPI_ADDRESS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &count);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI erase count value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_SPI_BYTE_COUNT;
      goto done;
    }

    ret = avr2utilSpiEraseCommand(pDevice, bVerbose, chipIndex, address, count);
  } else if (_tcsicmp(pCommand, _T("spi-program-mcs")) == 0) {
    unsigned int chipIndex;
    TCHAR* pFilename;
    int translation = 0;
    int nConverted;

    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'spi-program-mcs' command must be followed by (1) the SPI chip index, (2) the filename of an Intel Hex (.mcs) file.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    pFilename = arguments.param[2].pStringVal;

    if (arguments.nParam >= 4) {
      nConverted = stringToInt(arguments.param[3].pStringVal, &translation);
      if (1 != nConverted) {
        _tprintf(_T("*** Translation value '%s' not recognized as a signed number.\n"), arguments.param[3].pStringVal);
        ret = EXIT_INVALID_TRANSLATION;
        goto done;
      }
    }

    ret = avr2utilSpiProgramIntelHexCommand(pDevice, bVerbose, chipIndex, pFilename, (int32_t)translation);
  } else if (_tcsicmp(pCommand, _T("spi-verify-mcs")) == 0) {
    unsigned int chipIndex;
    TCHAR* pFilename;
    int translation = 0;
    int nConverted;

    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'spi-verify-mcs' command must be followed by (1) the SPI chip index, (2) the filename of an Intel Hex (.mcs) file.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    pFilename = arguments.param[2].pStringVal;

    if (arguments.nParam >= 4) {
      nConverted = stringToInt(arguments.param[3].pStringVal, &translation);
      if (1 != nConverted) {
        _tprintf(_T("*** Translation value '%s' not recognized as a signed number.\n"), arguments.param[3].pStringVal);
        ret = EXIT_INVALID_TRANSLATION;
        goto done;
      }
    }

    ret = avr2utilSpiVerifyIntelHexCommand(pDevice, bVerbose, chipIndex, pFilename, translation);
  } else if (_tcsicmp(pCommand, _T("spi-save-to-mcs")) == 0) {
    unsigned int chipIndex, address, count;
    TCHAR* pFilename;
    int nConverted;

    if (arguments.nParam < 5) {
      _tprintf(_T("*** 'spi-save-to-mcs' command must be followed by (1) the SPI chip index, (2) the filename of an Intel Hex (.mcs) file, (3) starting address, (4) byte count\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &chipIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** SPI chip index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    pFilename = arguments.param[2].pStringVal;

    nConverted = stringToUInt(arguments.param[3].pStringVal, &address);
    if (1 != nConverted) {
      _tprintf(_T("*** Starting address value '%s' not recognized as a number.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[4].pStringVal, &count);
    if (1 != nConverted) {
      _tprintf(_T("*** Byte count value '%s' not recognized as a number.\n"), arguments.param[4].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    ret = avr2utilSpiSaveIntelHexCommand(pDevice, bVerbose, chipIndex, pFilename, address, count);
  } else if (_tcsicmp(pCommand, _T("update-firmware")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'update-firmware' command must be followed by the firmware filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilUpdateOrVerifyFirmwareCommand(pDevice, bVerbose, FALSE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("update-brdcfg")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'update-brdcfg' command must be followed by the board config. filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilUpdateOrVerifyBoardConfigCommand(pDevice, bVerbose, FALSE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("update-vpd")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'update-vpd' command must be followed by the VPD filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilUpdateOrVerifyVpdCommand(pDevice, bVerbose, FALSE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("save-firmware")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'save-firmware' command must be followed by the filename (.bin) into which the firmware is to be saved.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilSaveFirmwareCommand(pDevice, bVerbose, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("save-brdcfg")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'save-brdcfg' command must be followed by the filename (.bin) into which the board config. is to be saved.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilSaveBoardConfigCommand(pDevice, bVerbose, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("save-vpd")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'save-vpd' command must be followed by the filename (.bin) into which the VPD is to be saved.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilSaveVpdCommand(pDevice, bVerbose, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("verify-brdcfg")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'verify-brdcfg' command must be followed by the board config. filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilUpdateOrVerifyBoardConfigCommand(pDevice, bVerbose, TRUE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("verify-firmware")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'verify-firmware' command must be followed by the firmware filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilUpdateOrVerifyFirmwareCommand(pDevice, bVerbose, TRUE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("verify-vpd")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'verify-vpd' command must be followed by the VPD filename (.bin).\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilUpdateOrVerifyVpdCommand(pDevice, bVerbose, TRUE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("display-vpd")) == 0) {
    ret = avr2utilDisplayVpdCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("display-vpd-raw")) == 0) {
    ret = avr2utilDisplayVpdRawCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("display-black-box")) == 0) {
    unsigned int nTypes = 0x3; /* 0x1 => clobber records, 0x2 => noclobber records, 0x3 => both */
    unsigned int nRecordCount = 0; /* 0 => all records, other => number of records to display */
    unsigned int nConverted;
    TCHAR* pTypes;

    if (arguments.nParam >= 2) {
      pTypes = arguments.param[1].pStringVal;

      if (_tcsicmp(pTypes, _T("both")) == 0) {
        nTypes = 0x3;
      } else if (_tcsicmp(pTypes, _T("clobber")) == 0) {
        nTypes = 0x1;
      } else if (_tcsicmp(pTypes, _T("noclobber")) == 0) {
        nTypes = 0x2;
      } else {
        _tprintf(_T("*** Valid 'display-black-box' command record types are: clobber, noclobber.\n"));
        ret = EXIT_INVALID_RECORD_TYPE;
        goto done;
      }
    }

    if (arguments.nParam >= 3) {
      nConverted = stringToUInt(arguments.param[2].pStringVal, &nRecordCount);
      if (1 != nConverted) {
        _tprintf(_T("*** Black Box record count value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
        ret = EXIT_INVALID_RECORD_COUNT;
        goto done;
      }
    }

    ret = avr2utilDisplayBlackBoxCommand(pDevice, bVerbose, nTypes, nRecordCount);
  } else if (_tcsicmp(pCommand, _T("display-sensors")) == 0) {
    ret = avr2utilDisplaySensorsCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("display-sensors-raw")) == 0) {
    ret = avr2utilDisplaySensorsRawCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("override-sensor")) == 0) {
    unsigned int sensorIndex, overrideValue;
    int nConverted;

    if (arguments.nParam < 3) {
      _tprintf(_T("*** 'override-sensor-raw' must be followed by the (1) sensor index and (2) the unscaled override value.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &sensorIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Sensor index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &overrideValue);
    if (1 != nConverted) {
      _tprintf(_T("*** Override value '%s' not recognized as a number.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_OVERRIDE;
      goto done;
    }

    ret = avr2utilOverrideSensorUnscaledCommand(pDevice, bVerbose, sensorIndex, overrideValue);
  } else if (_tcsicmp(pCommand, _T("release-sensor")) == 0) {
    unsigned int sensorIndex;
    int nConverted;

    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'release-sensor' must be followed by the sensor index.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &sensorIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Sensor index value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    ret = avr2utilReleaseSensorCommand(pDevice, bVerbose, sensorIndex);
  } else if (_tcsicmp(pCommand, _T("set-brd-power")) == 0) {
    bool_t bPowerOn; 

    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'set-brd-power' command must be followed by 'on' or 'off'\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    if (0 == _tcsicmp(arguments.param[1].pStringVal, _T("on"))) {
      bPowerOn = TRUE;
    } else if (0 == _tcsicmp(arguments.param[1].pStringVal, _T("off"))) {
      bPowerOn = FALSE;
    } else {
      _tprintf(_T("*** '%s' is not a recognized board power state; 'on' or 'off'\n"), arguments.param[1].pStringVal);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilSetBrdPowerCommand(pDevice, bVerbose, bPowerOn);
  } else if (_tcsicmp(pCommand, _T("get-brd-power")) == 0) {
    ret = avr2utilGetBrdPowerCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("force-mez-power")) == 0 || _tcsicmp(pCommand, _T("force-mez-power-nv")) == 0) {
    bool_t bNonVolatile = _tcsicmp(pCommand, _T("force-mez-power-nv")) == 0 ? TRUE : FALSE;
    unsigned nReqMillivolt, nMinMillivolt, nMaxMillivolt;
    unsigned int nMezIndex;
    int nConverted;

    if (arguments.nParam < 5) {
      _tprintf(_T("*** '%s' command must be followed by (1) requested voltage (mV), (2) minumum allowed voltage (mV), (3) maximum allowed voltage (mV) and (4) Mezzanine index\n"), pCommand);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &nReqMillivolt);
    if (1 != nConverted) {
      _tprintf(_T("Requested voltage value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[2].pStringVal, &nMinMillivolt);
    if (1 != nConverted) {
      _tprintf(_T("Minimum allowed voltage value '%s' not recognized as an unsigned integer.\n"), arguments.param[2].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[3].pStringVal, &nMaxMillivolt);
    if (1 != nConverted) {
      _tprintf(_T("Maximum allowed voltage value '%s' not recognized as an unsigned integer.\n"), arguments.param[3].pStringVal);
      ret = EXIT_INVALID_SPI_INDEX;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[4].pStringVal, &nMezIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Mezzanine index value '%s' not recognized as an unsigned integer.\n"), arguments.param[4].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    ret = avr2utilForceMezPowerCommand(pDevice, bVerbose, bNonVolatile, bConfirmed, nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
  } else if (_tcsicmp(pCommand, _T("get-mez-power")) == 0 || _tcsicmp(pCommand, _T("get-mez-power-nv")) == 0) {
    bool_t bNonVolatile = _tcsicmp(pCommand, _T("get-mez-power-nv")) == 0 ? TRUE : FALSE;
    unsigned int nMezIndex;
    int nConverted;

    if (arguments.nParam < 2) {
      _tprintf(_T("*** '%s' command must be followed by (1) Mezzanine index\n"), pCommand);
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &nMezIndex);
    if (1 != nConverted) {
      _tprintf(_T("*** Mezzanine index value '%s' not recognized as an unsigned integer.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_INDEX;
      goto done;
    }

    ret = avr2utilGetMezPowerCommand(pDevice, bVerbose, bNonVolatile, nMezIndex);
  } else if (_tcsicmp(pCommand, _T("set-boot-mode")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'set-boot-mode' command must be followed by either a decimal or hexadecimal number between 0 and 15 inclusive, or the word 'switch'.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilSetBootModeCommand(pDevice, bVerbose, FALSE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("get-boot-mode")) == 0) {
    ret = avr2utilGetBootModeCommand(pDevice, bVerbose, FALSE);
  } else if (_tcsicmp(pCommand, _T("set-boot-mode-nv")) == 0) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** 'set-boot-mode-nv' command must be followed by either a decimal or hexadecimal number between 0 and 15 inclusive, or the word 'switch'.\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    ret = avr2utilSetBootModeCommand(pDevice, bVerbose, TRUE, arguments.param[1].pStringVal);
  } else if (_tcsicmp(pCommand, _T("get-boot-mode-nv")) == 0) {
    ret = avr2utilGetBootModeCommand(pDevice, bVerbose, TRUE);
  } else if (_tcsicmp(pCommand, _T("test-clobber")) == 0) {
    ret = avr2utilTestClobberCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("test-timeout")) == 0) {
    ret = avr2utilTestTimeoutCommand(pDevice, bVerbose);
  } else if (_tcsicmp(pCommand, _T("raw")) == 0) {
    unsigned int readCount, writeCount, i, byte;
    int nConverted;
    uint8_t* pWriteBytes = NULL;

    if (arguments.nParam < 4) {
      _tprintf(_T("*** 'raw' command must be followed by (1) the expected response length, (2) two or more command bytes to write\n"));
      ret = EXIT_INSUFFICIENT_ARGS;
      goto done;
    }

    nConverted = stringToUInt(arguments.param[1].pStringVal, &readCount);
    if (1 != nConverted) {
      _tprintf(_T("*** Expected response length value '%s' not recognized as a number.\n"), arguments.param[1].pStringVal);
      ret = EXIT_INVALID_RAW_READ_COUNT;
      goto done;
    }

    writeCount = arguments.nParam - 2;

    pWriteBytes = (uint8_t*)malloc(writeCount * sizeof(uint8_t));
    if (NULL == pWriteBytes) {
      _tprintf(_T("*** Failed to allocate write byte buffer of %u byte(s).\n"), writeCount);
      ret = EXIT_ALLOCATION_FAILED;
      goto done;
    }

    for (i = 0; i < writeCount; i++) {
      nConverted = stringToUInt(arguments.param[i + 2].pStringVal, &byte);
      if (1 != nConverted) {
        free(pWriteBytes);
        _tprintf(_T("*** Command byte value '%s' not recognized as a number.\n"), arguments.param[i + 2].pStringVal);
        ret = EXIT_INVALID_RAW_BYTEVAL;
        goto done;
      }
      pWriteBytes[i] = (uint8_t)byte;
    }

    ret = avr2utilRawCommand(pDevice, bVerbose, readCount, writeCount, pWriteBytes);

    free(pWriteBytes);
  } else {
    _tprintf(_T("*** Unrecognised command: %s\n"), pCommand);
    _tprintf(_T("    Use '%s -h' for help.\n"), PROGRAM_NAME);
    ret = EXIT_UNRECOGNIZED_COMMAND;
    goto done;
  }

done:
  avr2utilCloseDevice(pDevice);

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
