#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# define _tcslen(s) strlen(s)
# define _stscanf_s sscanf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#include "boot_mode.h"
#include "exit_code.h"
#include "fpga_info.h"
#include "model_info.h"
#include "vpd.h"

static int stringToUInt(
  const TCHAR* pString,
  unsigned int* pValue)
{
  if (_tcslen(pString) >= 2 && pString[0] == _T('0') && pString[1] == _T('x')) {
    return _stscanf_s(pString + 2, _T("%x"), pValue);
  } else {
    return _stscanf_s(pString, _T("%u"), pValue);
  }
}

static int
bootModeStringToNumber(
  uint32_t nDrawingNumber,
  const TCHAR* pszBootMode)
{
  switch (nDrawingNumber) {
  case DRAWING_NUM_ADMXRC9Z1:
    return avr2utilZynqMPBootModeStringToNumber(pszBootMode);

  case DRAWING_NUM_ADMPA100:
  case DRAWING_NUM_ADMPA101:
    return avr2utilVersalBootModeStringToNumber(pszBootMode);

  default:
    return -2; /* Model not known to support software-controlled boot mode */
  }
}

static const TCHAR*
bootModeNumberToString(
  uint32_t nDrawingNumber,
  unsigned int nBootMode)
{
  switch (nDrawingNumber) {
  case DRAWING_NUM_ADMXRC9Z1:
    return avr2utilVersalBootModeNumberToString(nBootMode);

  case DRAWING_NUM_ADMPA100:
  case DRAWING_NUM_ADMPA101:
    return avr2utilVersalBootModeNumberToString(nBootMode);

  default:
    return _T("UNKNOWN");
  }
}

int
avr2utilSetBootModeCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bNonVolatile,
  const TCHAR* pszBootMode)
{
  uint8_t vpdBuffer[256];
  uint32_t nDrawingNumber = 0;
  const TCHAR* pszModelName = NULL;
  bool_t bDrawingNumberValid = FALSE;
  uint8_t command[4];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int nConverted, iBootMode;
  unsigned int nBootMode;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The set-boot-mode%s command requires the device to not be in Service Mode.\n"), bNonVolatile ? _T("-nv") : _T(""));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  /* Attempt to get VPD for drawing number; if this fails, it's not a fatal error
  ** if the user specified the boot mode as 'switch' or numerically. */
  memset(vpdBuffer, 0, sizeof(vpdBuffer));
  ret = avr2utilRequestVpd(pDevice, bVerbose, vpdBuffer);
  if (EXIT_OK == ret) {
    nDrawingNumber = GET_VPD_VAL32(&vpdBuffer[4]);
    pszModelName = avr2utilGetModelName(nDrawingNumber);
    bDrawingNumberValid = TRUE;
  }

  if ((uintptr_t)pszBootMode <= 0xFF) {
    /* If invoked from VxWorks, boot mode might be a string pointer or a small integer */
    nBootMode = (int)(uintptr_t)pszBootMode;
  } else {
    nConverted = stringToUInt(pszBootMode, &nBootMode);
    if (1 != nConverted) {
      if (0 == _tcsicmp(pszBootMode, _T("switch"))) {
        nBootMode = 0xFF;
      } else {
        if (!bDrawingNumberValid) {
          _tprintf(_T("*** Failed to get VPD for determining FPGA family; cannot convert boot mode of '%s' to a number.\n"), pszBootMode);
          _tprintf(_T("    If you know the boot mode value, try using the literal value instead of its symbolic name.\n"));
          ret = EXIT_INVALID_BOOTMODE;
          goto done;
        }

        iBootMode = bootModeStringToNumber(nDrawingNumber, pszBootMode);
        if (iBootMode < 0) {
          _tprintf(_T("*** '%s' not recognized as a valid boot mode for product '%s'.\n"), pszBootMode, pszModelName);
          _tprintf(_T("    If you know the boot mode value, try using the literal value instead of its symbolic name.\n"));
          ret = EXIT_INVALID_BOOTMODE;
          goto done;
        }

        nBootMode = (unsigned int)iBootMode;
      }
    }
  }

  /* Now convert numerical boot mode back to a symbolic string for the requested boot mode, if it was a numerical string */
  pszBootMode = bDrawingNumberValid ? bootModeNumberToString(nDrawingNumber, nBootMode) : _T("UNKNOWN");

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x95; /* SetMode command */
  command[2] = (uint8_t)nBootMode;
  command[3] = (uint8_t)(bNonVolatile ? 0x02 : 0x01);
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x15; /* SetMode response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 0x1U:
      pReason = _T("Set Boot Mode message not supported by device.");
      break;

    case 0x2U:
      pReason = _T("Device reports invalid boot mode.");
      break;

    case 0x3U:
      pReason = _T("Device reports invalid volatile / nonvolatile flag.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  if (nBootMode == 0xFF) {
    if (bNonVolatile) {
      _tprintf(_T("Boot Mode at next power-on will be entirely under control of switches.\n"));
    } else {
      _tprintf(_T("Boot Mode is now entirely under control of switches.\n"));
    }
  } else {
    if (bNonVolatile) {
      _tprintf(_T("Boot Mode at next power-on will be '%s' (0x%X).\n"), pszBootMode, (unsigned int)nBootMode);
    } else {
      _tprintf(_T("Boot Mode is now '%s' (0x%X).\n"), pszBootMode, (unsigned int)nBootMode);
    }
  }

done:
  return ret;
}

int
avr2utilGetBootModeCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bNonVolatile)
{
  uint8_t vpdBuffer[256];
  uint32_t nDrawingNumber = 0;
  bool_t bDrawingNumberValid = FALSE;
  uint8_t command[2];
  uint8_t response[6];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint8_t nBootMode, nBootModeActual;
  const TCHAR* pszBootMode;
  const TCHAR* pszBootModeActual;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The get-boot-mode%s command requires the device to not be in Service Mode.\n"), bNonVolatile ? _T("-nv") : _T(""));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  /* Attempt to get VPD for drawing number. If this fails, it's not a fatal error, but it
  ** means we cannot convert the numerical boot mode value into a human-readable string. */
  memset(vpdBuffer, 0, sizeof(vpdBuffer));
  ret = avr2utilRequestVpd(pDevice, bVerbose, vpdBuffer);
  if (EXIT_OK == ret) {
    nDrawingNumber = GET_VPD_VAL32(&vpdBuffer[4]);
    bDrawingNumberValid = TRUE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x96; /* GetMode command */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x16; /* GetMode response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 0x1U:
      pReason = _T("Get Boot Mode message not supported by device.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  if (bNonVolatile) {
    nBootMode = response[4];
    if (nBootMode == 0xFF) {
      _tprintf(_T("Boot mode at next power-on is entirely under control of switches.\n"));
    } else {
      pszBootMode = bDrawingNumberValid ? bootModeNumberToString(nDrawingNumber, nBootMode) : _T("UNKNOWN");
      _tprintf(_T("Boot mode at next power-on is '%s' (0x%X).\n"), pszBootMode, (unsigned int)nBootMode);
    }
  } else {
    nBootMode = response[3];
    if (nBootMode == 0xFF) {
      _tprintf(_T("Boot mode currently entirely under control of switches.\n"));
    } else {
      pszBootMode = bDrawingNumberValid ? bootModeNumberToString(nDrawingNumber, nBootMode) : _T("UNKNOWN");
      _tprintf(_T("Boot mode currently driven is '%s' (0x%X).\n"), pszBootMode, (unsigned int)nBootMode);
    }

    nBootModeActual = response[5];
    pszBootModeActual = bDrawingNumberValid ? bootModeNumberToString(nDrawingNumber, nBootModeActual) : _T("UNKNOWN");
    _tprintf(_T("Actual boot mode pins are '%s' (0x%X).\n"), pszBootModeActual, (unsigned int)nBootModeActual);
  }

done:
  return ret;
}
