#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>

#include "brd_power.h"
#include "exit_code.h"

int
avr2utilSetBrdPowerCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bPowerOn)
{
  uint8_t command[3];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The set-brd-power command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x97; /* SetBrdPower command */
  command[2] = (uint8_t)(bPowerOn ? 0x00 : 0x01);
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x17; /* SetBrdPower response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 0x1U:
      pReason = _T("SetBrdPower message not supported by device.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  _tprintf(_T("Set SW board power request to %s.\n"), bPowerOn ? _T("ON") : _T("OFF"));

done:
  return ret;
}

int
avr2utilGetBrdPowerCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose)
{
  uint8_t command[2];
  uint8_t response[6];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status, soft, hard, actual;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The get-brd-power command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x98; /* GetBrdPower command */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x18; /* GetBrdPower response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 0x1U:
      pReason = _T("GetBrdPower message not supported by device.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  soft = response[3];
  hard = response[4];
  actual = response[5];
  _tprintf(_T("SW req. power %s, HW req. power %s, Actual PSU_OFF_L = %u\n"),
    soft ? _T("OFF") : _T("ON"), hard ? _T("OFF") : _T("ON"), actual ? 0 : 1);

done:
  return ret;
}
