#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>

#include "exit_code.h"
#include "clock.h"

int
avr2utilGetClkCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int clockIndex)
{
  uint8_t command[3];
  uint8_t response[7];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint32_t frequency;
  int ret = EXIT_OK;

  if (clockIndex > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The getclk command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x8F; /* GetFrequency query */
  command[2] = (uint8_t)clockIndex; /* Clock generator index; cast safe due to above range check. */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x0F; /* GetFrequency response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid clock generator index");
      break;

    case 2:
      pReason = _T("Simple clock generator programming not supported for this model.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned error status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  frequency = ((uint32_t)response[3] << 0) | ((uint32_t)response[4] << 8) | ((uint32_t)response[5] << 16) | ((uint32_t)response[6] << 24);
  _tprintf(_T("Frequency of clock generator %u is %lu Hz\n"),
    clockIndex, (unsigned long)frequency);

done:
  return ret;
}

int
avr2utilSetClkCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int clockIndex,
  uint32_t frequency)
{
  uint8_t command[7];
  uint8_t response[7];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  if (clockIndex > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The setclk command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x8E; /* SetFrequency query */
  command[2] = (uint8_t)clockIndex; /* Clock generator index; cast safe due to above range check. */
  command[3] = (uint8_t)((frequency >> 0) & 0xFFU); /* Frequency bits [7:0] */
  command[4] = (uint8_t)((frequency >> 8) & 0xFFU); /* Frequency bits [15:8] */
  command[5] = (uint8_t)((frequency >> 16) & 0xFFU); /* Frequency bits [23:16] */
  command[6] = (uint8_t)((frequency >> 24) & 0xFFU); /* Frequency bits [31:24] */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x0E; /* SetFrequency response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid clock generator index");
      break;

    case 2:
      pReason = _T("Simple clock generator programming not supported for this model.");
      break;

    case 17:
      pReason = _T("Requested frequency is out of range");
      break;

    case 33:
      pReason = _T("Microcontroller failed to communicate with SI5338");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  _tprintf(_T("Frequency for clock generator %u set to %lu Hz\n"), clockIndex, (unsigned long)frequency);

done:
  return ret;
}

int
avr2utilGetClkNvCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int clockIndex)
{
  uint8_t command[4];
  uint8_t response[7];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint32_t frequency;
  bool_t bIsOverridden;
  int ret = EXIT_OK;

  if (clockIndex > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The getclknv command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x84; /* GetFrequencyEx query */
  command[2] = (uint8_t)clockIndex; /* Clock generator index; cast safe due to above range check. */
  command[3] = 0x1; /* Flags: 0x1 => Get NV frequency value */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x04; /* GetFrequencyEx response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid clock generator index");
      break;

    case 2:
      pReason = _T("Simple clock generator programming not supported for this model.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned error status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  frequency = ((uint32_t)response[3] << 0) | ((uint32_t)response[4] << 8) | ((uint32_t)response[5] << 16) | ((uint32_t)response[6] << 24);
  bIsOverridden = (frequency != 0U && frequency != 0xFFFFFFFFU);
  if (bIsOverridden) {
    _tprintf(_T("NV override frequency for clock generator %u is %lu Hz\n"),
      clockIndex, (unsigned long)frequency);
  } else {
    _tprintf(_T("NV override frequency for clock generator %u is %lu(0x%lX), which will not result in programming at power-up.\n"),
      clockIndex, (unsigned long)frequency, (unsigned long)frequency);
  }

done:
  return ret;
}

int
avr2utilSetClkNvCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int clockIndex,
  uint32_t frequency)
{
  uint8_t command[8];
  uint8_t response[7];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  bool_t bIsOverridden;
  int ret = EXIT_OK;

  if (clockIndex > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The setclknv command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x83; /* SetFrequencyEx query */
  command[2] = (uint8_t)clockIndex; /* Clock generator index; cast safe due to above range check. */
  command[3] = (uint8_t)((frequency >> 0) & 0xFFU); /* Frequency bits [7:0] */
  command[4] = (uint8_t)((frequency >> 8) & 0xFFU); /* Frequency bits [15:8] */
  command[5] = (uint8_t)((frequency >> 16) & 0xFFU); /* Frequency bits [23:16] */
  command[6] = (uint8_t)((frequency >> 24) & 0xFFU); /* Frequency bits [31:24] */
  command[7] = 0x3; /* Flags: 0x1 => Set NV frequency value, 0x2 => Don't program clock generator now */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x03; /* SetFrequencyEx response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid clock generator index");
      break;

    case 2:
      pReason = _T("Clock generator frequency cannot be overridden");
      break;

    case 3:
      pReason = _T("Simple clock generator programming not supported for this model.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  bIsOverridden = (frequency != 0U && frequency != 0xFFFFFFFFU);

  if (bIsOverridden) {
    _tprintf(_T("NV override frequency for clock generator %u set to %lu Hz\n"),
      clockIndex, (unsigned long)frequency);
  } else {
    _tprintf(_T("NV override frequency for clock generator %u set to %lu(0x%lX), which will not result in programming at power-up.\n"),
      clockIndex, (unsigned long)frequency,  (unsigned long)frequency);
  }

done:
  return ret;
}
