#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif

#include <sdk_common.h>

#include "exit_code.h"
#include "clock_regmap.h"
#include "lmk61e2_regmap.h"

#define MAX_REGMAP_SIZE (508)

/* Flags understood by uC in flags byte of command message */
#define FLAG_NONVOLATILE     (0x1 << 0) /* Will write to clock generator in a nonvolatile way */
#define FLAG_FACTORY_DEFAULT (0x1 << 1) /* Will write factory default register map to clock generator */
#define FLAG_PACKET_FIRST    (0x1 << 6) /* First packet of register map data */
#define FLAG_PACKET_FINAL    (0x1 << 7) /* Final packet of register map data */

typedef enum _RegisterMapType {
  RegisterMapSI5338  = 0, /* Any variant (A, B etc.) of SI5338 */
  RegisterMapLMK61E2 = 1
} RegisterMapType;

static void messageCallbackRoutineLMK61E2(
  const TCHAR* pszMessage,
  LMK61E2RegMapMessageType nMessageType)
{
  const TCHAR* pszMessageType = _T("");

  switch (nMessageType) {
  case LMK61E2RegMapMessageInfo:
    pszMessageType = _T("Info");
    break;

  case LMK61E2RegMapMessageWarning:
    pszMessageType = _T("Warning");
    break;

  case LMK61E2RegMapMessageError:
    pszMessageType = _T("Error");
    break;
  }

  _tprintf(_T("%s: %s\n"), pszMessageType, pszMessage);
}

/* Attempt to determine type of register map file */
static int
probeRegisterMapFile(
  const TCHAR* pszFilename,
  RegisterMapType* peType)
{
  LMK61E2RegMapError nErrorLMK61E2;
  LMK61E2RegMap* pRegMapLMK61E2 = NULL;

  /* First try LMK61E2 format */

  /* For probing register map type, attempt to read LMK61E2 register map without any message display callback */
  nErrorLMK61E2 = LMK61E2RegMapRead(pszFilename, NULL, &pRegMapLMK61E2);
  if (LMK61E2RegMapOk == nErrorLMK61E2) {
    /* Succeeded in reading file in LMK61E2 format */
    *peType = RegisterMapLMK61E2;

    LMK61E2RegMapDestroy(pRegMapLMK61E2);

    return EXIT_OK;
  } else if (nErrorLMK61E2 != LMK61E2RegMapBadFile) {
    /* Some low level error occurred reading file - treat as fatal error */
    const TCHAR* pszReason = LM61E2RegMapGetErrorDesc(nErrorLMK61E2);

    _tprintf(_T("*** Failed to read file '%s': %s (error %d)\n"), pszFilename, pszReason, nErrorLMK61E2);

    return EXIT_READ_REG_MAP_FAILED;
  } else {
    /* File doesn't look like an LMK61E2 register map file - don't treat as fatal error */
  }

  /* No more formats to try */

  _tprintf(_T("*** Failed to automatically determine register map file format.\n"));

  return EXIT_PROBE_REG_MAP_FAILED;
}

static void unloadRegisterMapFile(
  RegisterMapType eType,
  void* pRegisterMap)
{
  switch (eType) {
  case RegisterMapLMK61E2:
    LMK61E2RegMapDestroy((LMK61E2RegMap*)pRegisterMap);
    break;

  default:
    /* How did we get here? */
    break;
  }
}

static void dumpRegMapLMK61E2(
  const LMK61E2RegMap* pRegMap)
{
  unsigned int i, n = pRegMap->nPairCount;

  _tprintf(_T("%u register map pair(s):\n"), n);

  for (i = 0; i < n; i++) {
    _tprintf(
      _T("#%02u index=0x%02X value=0x%02X\n"),
      i, pRegMap->pairs[i].nRegIndex, pRegMap->pairs[i].nRegValue);
  }
}

static int loadRegisterMapFile(
  const TCHAR* pszFilename,
  RegisterMapType eType,
  void** ppRegisterMap,
  void** ppRegisterMapData,
  size_t* pnRegisterMapSize,
  bool_t bVerbose)
{
  LMK61E2RegMap* pRegMapLMK61E2 = NULL;
  LMK61E2RegMapError nErrorLMK61E2;

  switch (eType) {
  case RegisterMapLMK61E2:
    /* Read register map with message display callback */
    nErrorLMK61E2 = LMK61E2RegMapRead(pszFilename, messageCallbackRoutineLMK61E2, &pRegMapLMK61E2);
    if (LMK61E2RegMapOk != nErrorLMK61E2) {
      const TCHAR* pszReason = LM61E2RegMapGetErrorDesc(nErrorLMK61E2);

      _tprintf(_T("*** Failed to read file '%s' in LMK61E2 register map format: %s (error %d)\n"), pszFilename, pszReason, nErrorLMK61E2);

      return EXIT_READ_REG_MAP_FAILED;
    }

    _tprintf(_T("File '%s' is an LMK61E2 register map with %u register / value pair(s)\n"), pszFilename, pRegMapLMK61E2->nPairCount);

    if (bVerbose) {
      dumpRegMapLMK61E2(pRegMapLMK61E2);
    }

    *ppRegisterMap = pRegMapLMK61E2;
    *ppRegisterMapData = &pRegMapLMK61E2->pairs[0];
    *pnRegisterMapSize = (size_t)pRegMapLMK61E2->nPairCount * 2;
    break;

  default:
    /* How did we get here? */
    return EXIT_READ_REG_MAP_FAILED;
  }

  return EXIT_OK;
}

extern int
avr2utilSetClkRegMapCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nClockIndex,
  uint32_t nFlags,
  const TCHAR* pszRegMapPath,
  const TCHAR* pszRegMapType)
{
  RegisterMapType eRegMapType = (RegisterMapType)0;
  void* pRegMap = NULL;
  void* pRegMapData = NULL;
  size_t nRegMapSize = 0;
  bool_t bNonVolatile = (nFlags & PROGRAM_REGMAP_NONVOLATILE) == PROGRAM_REGMAP_NONVOLATILE ? TRUE : FALSE;
  bool_t bFactoryDefault = FALSE;
  uint8_t command[7 + MAX_REGMAP_SIZE];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t commandLength, responseLength;
  uint8_t nCmdFlags, avr2Status;
  int ret = EXIT_OK;

  if (nClockIndex > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (NULL == pszRegMapPath) {
    bFactoryDefault = TRUE;

    _tprintf(_T("Register map to be used is factory default.\n"));
  } else {
    if (NULL == pszRegMapType || 0 == _tcsicmp(pszRegMapType, _T("auto"))) {
      /* Attempt to figure out register map type */
      ret = probeRegisterMapFile(pszRegMapPath, &eRegMapType);
      if (EXIT_OK != ret) {
        return ret;
      }
    } else {
      /* User has explicitly specified register map type */
      if (0 == _tcsicmp(pszRegMapType, _T("lmk61e2"))) {
        eRegMapType = RegisterMapLMK61E2;
      } else {
        _tprintf(_T("*** Unrecognized register map file type '%s'; expecting one of: lmk61e2\n"), pszRegMapType);
        ret = EXIT_BAD_REG_MAP_TYPE;
        goto done;
      }
    }

    ret = loadRegisterMapFile(pszRegMapPath, eRegMapType, &pRegMap, &pRegMapData, &nRegMapSize, bVerbose);
    if (EXIT_OK != ret) {
      return ret;
    }

    if (nRegMapSize > MAX_REGMAP_SIZE) {
      _tprintf(_T("*** Register map size of %llu(0x%llX) bytes is larger than maximum allowed size of %u(0x%X).\n"),
        (unsigned long long)nRegMapSize, (unsigned long long)nRegMapSize, MAX_REGMAP_SIZE, MAX_REGMAP_SIZE);
      ret = EXIT_BAD_REG_MAP;
      goto done;
    }
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The setclk%s-regmap command requires the device to not be in Service Mode.\n"), bNonVolatile ? _T("nv") : _T(""));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  nCmdFlags = FLAG_PACKET_FIRST | FLAG_PACKET_FINAL;
  if (bNonVolatile) {
    nCmdFlags |= FLAG_NONVOLATILE; /* Nonvolatile register map write to clock generator */
  }
  if (bFactoryDefault) {
    nCmdFlags |= FLAG_FACTORY_DEFAULT; /* Microcontroller will write factory default register map to clock generator */
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x9B; /* Write clock register map command */
  command[2] = nCmdFlags;
  command[3] = (uint8_t)nClockIndex; /* Clock generator index; cast safe due to above range check. */
  command[4] = (uint8_t)eRegMapType;
  command[5] = (uint8_t)(nRegMapSize & 0xFF);
  command[6] = (uint8_t)((nRegMapSize >> 8) & 0xFF);
  memcpy(&command[7], pRegMapData, nRegMapSize);
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x1B; /* Write clock register map response */

  commandLength = (uint32_t)(7 + nRegMapSize);

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    commandLength,
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    if (avr2Status >= 0x80) {
      uint8_t nChipError = (uint8_t)(avr2Status & 0x7F);

      _tprintf(_T("*** AVR2 returned status 0x%02X => chip-specific error 0x%X\n"), avr2Status, nChipError);
      ret = EXIT_AVR2_BAD_STATUS;
      goto done;
    } else {
      const TCHAR* pReason;

      switch (avr2Status) {
      case 1:
        pReason = _T("Register map clock generator programming not supported for this model");
        break;

      case 2:
        pReason = _T("Invalid clock generator index");
        break;

      case 3:
        pReason = _T("Invalid flag");
        break;

      case 4:
        pReason = _T("Invalid register map");
        break;

      default:
        pReason = _T("** UNEXPECTED ERROR CODE **");
        break;
      }

      _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), avr2Status, pReason);
      ret = EXIT_AVR2_BAD_STATUS;
      goto done;
    }
  }

  _tprintf(_T("Register map write (%s) completed successfully.\n"), bNonVolatile ? _T("nonvolatile") : _T("volatile"));

done:
  if (NULL != pRegMap) {
    unloadRegisterMapFile(eRegMapType, pRegMap);
  }

  return ret;
}
