#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>
#include <platform_time.h>

#include "exit_code.h"
#include "comm_test.h"

int
avr2utilCommTestCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose)
{
  const unsigned int testDurationMs = 2000;
  const unsigned int payloadLength = 32;
  const unsigned int maxErrorsDisplayed = 10;
  admxrc3_time_t duration, start, now, expire;
  uint8_t command[34];
  uint8_t response[34];
  uint8_t expResponse[2];
  uint32_t responseLength;
  size_t nLinkSent = 0, nLinkReceived = 0, nPayloadSent = 0, nPayloadReceived = 0;
  unsigned int i, errorCount = 0;
  uint8_t expected, actual, sequenceNumber;
  double elapsedSec;
  int ret = EXIT_OK;

  duration = doubleToTime((double)testDurationMs * 0.001);
  start = now = timeGet();
  expire = timeAdd(start, duration);

  memset(command, 0, sizeof(command));
  command[0] = 0x03; /* User command group */
  command[1] = 0x84; /* Loopback32 command */

  expResponse[0] = 0x03; /* User command group */
  expResponse[1] = 0x04; /* Loopback32 response */

  _tprintf(_T("Measuring data rate for %u ms using Loopback32 command...\n"), testDurationMs);

  sequenceNumber = 1;

  do {
    /* Set up loopback32 data. */
    for (i = 0; i < payloadLength; i++) {
      command[i + 2] = (uint8_t)(i + sequenceNumber);
    }

    /* Zero-out the response buffer. */
    memset(response, 0, sizeof(response));

    ret = avr2utilSendAvr2Command(
      pDevice,
      bVerbose,
      0,
      command,
      sizeof(command),
      expResponse,
      response,
      sizeof(response),
      &responseLength);
    if (EXIT_OK != ret) {
      return ret;
    }

    /* Verify loopback32 response data; should be the same as the command data. */
    for (i = 0; i < payloadLength; i++) {
      expected = (uint8_t)(i + sequenceNumber);
      actual = response[i + 2];
      if (actual != expected) {
        if (errorCount < maxErrorsDisplayed) {
          _tprintf(_T("*** Error in loopback32 response data byte %u: expected=0x%02X actual=0x%02X\n"),
            i, expected, actual);
        } else if (errorCount == maxErrorsDisplayed) {
          _tprintf(_T("*** More errors in loopback32 response data.\n"));
        }
        errorCount++;
      }
    }

    nLinkSent += sizeof(command);
    nLinkReceived += sizeof(response);
    nPayloadSent += sizeof(command) - 2;
    nPayloadReceived += sizeof(response) - 2;

    if (errorCount) {
      break;
    }

    /* sequenceNumber is always in the range 1 .. 255 */
    if (0xFF == sequenceNumber) {
      sequenceNumber = 1;
    } else {
      sequenceNumber++;
    }

    now = timeGet();
  } while (timeBefore(now, expire));

  elapsedSec = timeToDouble(timeSubtract(now, start));

  if (elapsedSec >= 0.01) {
    _tprintf(_T("Link sent / received rate = %.1f / %.1f B/s\n"),
      (double)nLinkSent / (double)elapsedSec, (double)nLinkReceived / (double)elapsedSec);
    _tprintf(_T("Payload sent / received rate = %.1f / %.1f B/s\n"),
      (double)nPayloadSent / (double)elapsedSec, (double)nPayloadReceived / (double)elapsedSec);

    if (errorCount) {
      _tprintf(_T("*** NOTE: The computed data rates are not reliable because at least one error was detected in a loopback32 response payload.\n"));
    }
  } else {
    _tprintf(_T("*** Data rates not computed because the test did not run for long enough.\n"));
  }

  return EXIT_OK;
}
