#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>
# define AVR2_VIA_USB_SUPPORTED
# define AVR2_VIA_ADXDMA_SUPPORTED

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>

#include "exit_code.h"
#include "firmware.h"
#include "srv_mode.h"

#define FLASH_PAGE_ORDER (9)
#define FLASH_PAGE_SIZE (1U << FLASH_PAGE_ORDER)
#define FLASH_PAGE_MASK (FLASH_PAGE_SIZE - 1U)

#define FIRMWARE_BASE_ADDR     (0x0U) /* Address in Application Flash where firmware code lives */
#define FIRMWARE_LENGTH_LIMIT  (0x1FC00U) /* Maximum possible size of firmware code */

#define BOARD_CONFIG_BASE_ADDR    (0x1FC00U) /* Address in Application Flash where board configuration block lives */
#define BOARD_CONFIG_LENGTH_LIMIT (512U)     /* Maximum possible size of board configuration block */

#define VPD_BASE_ADDR    (0x1FF00U) /* Address in Application Flash where VPD lives */
#define VPD_LENGTH_LIMIT (128U)     /* Maximum possible size of VPD block */

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

int
readApplicationFlashPage(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint32_t pageAddress,
  uint8_t* pPageBuffer)
{
  uint8_t command[5];
  uint8_t response[3 + FLASH_PAGE_SIZE];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint16_t pageIndex;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  _tprintf(_T("Reading Application Flash page @ 0x%lX...\n"), (unsigned long)pageAddress);

  pageIndex = (uint16_t)(pageAddress >> FLASH_PAGE_ORDER); /* Cast should be safe because range of pageAddress is limited by callers */

  command[0] = 0x06; /* Boot Flash command group */
  command[1] = 0x80; /* Flash Application Read command */
  command[2] = FLASH_PAGE_ORDER;
  command[3] = (uint8_t)((pageIndex >> 0) & 0xFFU);
  command[4] = (uint8_t)((pageIndex >> 8) & 0xFFU);
  expResponse[0] = 0x06; /* Boot Flash command group */
  expResponse[1] = 0x00; /* Flash Application Read response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Flash Application Read command\n"), avr2Status);
    return EXIT_AVR2_COMMAND_ERROR;
  }

  memcpy(pPageBuffer, &response[3], FLASH_PAGE_SIZE);

  return EXIT_OK;
}

int
writeApplicationFlashPage(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint32_t pageAddress,
  const uint8_t* pPageBuffer)
{
  uint8_t command[5 + FLASH_PAGE_SIZE];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint16_t pageIndex;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  _tprintf(_T("Writing Application Flash page @ 0x%lX...\n"), (unsigned long)pageAddress);

  pageIndex = (uint16_t)(pageAddress >> FLASH_PAGE_ORDER); /* Cast should be safe because range of pageAddress is limited by callers */

  command[0] = 0x06; /* Boot Flash command group */
  command[1] = 0x82; /* Flash Application Write command */
  command[2] = FLASH_PAGE_ORDER;
  command[3] = (uint8_t)((pageIndex >> 0) & 0xFFU);
  command[4] = (uint8_t)((pageIndex >> 8) & 0xFFU);
  memcpy(&command[5], pPageBuffer, FLASH_PAGE_SIZE);
  expResponse[0] = 0x06; /* Boot Flash command group */
  expResponse[1] = 0x02; /* Flash Application Write response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Flash Application Write command\n"), avr2Status);
    return EXIT_AVR2_COMMAND_ERROR;
  }

  return EXIT_OK;
}

int
avr2utilWriteApplicationFlash(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint32_t address,
  const uint8_t* pBuffer,
  uint32_t length)
{
  uint8_t pageBuffer[512];
  uint32_t pageAddress, pageOffset, chunk;
  int ret = EXIT_OK;

  while (length) {
    pageOffset = address & FLASH_PAGE_MASK;
    pageAddress = address & ~FLASH_PAGE_MASK;
    chunk = FLASH_PAGE_SIZE - pageOffset;
    if (chunk > length) {
      chunk = length;
    }

    if (chunk != FLASH_PAGE_SIZE) {
      ret = readApplicationFlashPage(pDevice, bVerbose, pageAddress, pageBuffer);
      if (EXIT_OK != ret) {
        return ret;
      }
    }
    memcpy(pageBuffer + pageOffset, pBuffer, chunk);

    ret = writeApplicationFlashPage(pDevice, bVerbose, pageAddress, pageBuffer);
    if (EXIT_OK != ret) {
      return ret;
    }

    length -= chunk;
    address += chunk;
    pBuffer += chunk;
  }

  return EXIT_OK;
}

int
avr2utilReadApplicationFlash(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint32_t address,
  uint8_t* pBuffer,
  uint32_t length)
{
  uint8_t pageBuffer[512];
  uint32_t pageAddress, pageOffset, chunk;
  int ret = EXIT_OK;

  while (length) {
    pageOffset = address & FLASH_PAGE_MASK;
    pageAddress = address & ~FLASH_PAGE_MASK;
    chunk = FLASH_PAGE_SIZE - pageOffset;
    if (chunk > length) {
      chunk = length;
    }

    ret = readApplicationFlashPage(pDevice, bVerbose, pageAddress, pageBuffer);
    if (EXIT_OK != ret) {
      return ret;
    }

    memcpy(pBuffer, pageBuffer + pageOffset, chunk);

    pBuffer += chunk;
    address += chunk;
    length -= chunk;
  }

  return ret;
}

int
avr2utilVerifyApplicationFlash(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint32_t address,
  const uint8_t* pBuffer,
  uint32_t length)
{
  const unsigned int maxErrorReported = 20;
  uint8_t pageBuffer[512];
  uint32_t i, pageAddress, pageOffset, chunk;
  uint32_t errorCount = 0;
  int ret = EXIT_OK;

  while (length) {
    pageOffset = address & FLASH_PAGE_MASK;
    pageAddress = address & ~FLASH_PAGE_MASK;
    chunk = FLASH_PAGE_SIZE - pageOffset;
    if (chunk > length) {
      chunk = length;
    }

    ret = readApplicationFlashPage(pDevice, bVerbose, pageAddress, pageBuffer);
    if (EXIT_OK != ret) {
      return ret;
    }

    for (i = 0; i < chunk; i++) {
      uint8_t expected = *pBuffer++;
      uint8_t actual = pageBuffer[pageOffset + i];

      if (actual != expected) {
        if (errorCount < maxErrorReported) {
          if (expected != actual) {
            _tprintf(_T("*** Verification error at address 0x%lX; expected=0x%02X actual=0x%02X\n"),
              (unsigned long)address, (unsigned int)expected, (unsigned int)actual);
          }
        } else if (errorCount == maxErrorReported) {
          _tprintf(_T("*** More than %u verification errors detected.\n"), maxErrorReported);
        }
        errorCount++;
      }
      address++;
    }

    length -= chunk;
  }

  if (errorCount) {
    _tprintf(_T("*** Total of %u verification errors detected.\n"), errorCount);
    ret = EXIT_VERIFY_FAILED;
  } else {
    _tprintf(_T("Verified OK.\n"));
  }

  return ret;
}

int
avr2utilUpdateOrVerifyFirmwareCommand(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bVerifyOnly,
  const TCHAR* pFilename)
{
  const unsigned int firmwareLengthLimit = FIRMWARE_LENGTH_LIMIT;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  long length;
  size_t nRead;
  bool_t bWasInServiceMode;
  int ret = EXIT_OK;

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, TRUE);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to open firmware file '%s' for reading.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of firmware file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of firmware file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of firmware file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if ((size_t)length > firmwareLengthLimit) { /* Cast to size_t safe due to positive check above */
    _tprintf(_T("*** Firmware file cannot be more than 0x%X bytes; actual length is 0x%llX.\n"),
      firmwareLengthLimit, (unsigned long long)length);
    ret = EXIT_FIRMWARE_TOO_LARGE;
    goto done;
  }
  _tprintf(_T("Firmware file '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  pBuffer = (uint8_t*)malloc(length);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate firmware file buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  nRead = fread(pBuffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from firmware file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  bWasInServiceMode = avr2utilIsInServiceMode(pDevice);

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (!bVerifyOnly) {
    ret = avr2utilWriteApplicationFlash(pDevice, bVerbose, FIRMWARE_BASE_ADDR, pBuffer, (uint32_t)length); /* Cast safe due to above check on 'length' */
  }
  if (EXIT_OK == ret) {
    ret = avr2utilVerifyApplicationFlash(pDevice, bVerbose, FIRMWARE_BASE_ADDR, pBuffer, (uint32_t)length); /* Cast safe due to above check on 'length' */
  }

  avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, bWasInServiceMode, FALSE);

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilUpdateOrVerifyBoardConfigCommand(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bVerifyOnly,
  const TCHAR* pFilename)
{
  const unsigned int boardConfigLengthLimit = BOARD_CONFIG_LENGTH_LIMIT;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  long length;
  size_t nRead;
  bool_t bWasInServiceMode;
  int ret = EXIT_OK;

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, TRUE);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to open board config. file '%s' for reading.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of board config. file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of board config. file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of board config. file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if ((size_t)length > boardConfigLengthLimit) { /* Cast to size_t safe due to positive check above */
    _tprintf(_T("*** Board config. file cannot be more than 0x%X bytes; actual length is 0x%llX.\n"),
      boardConfigLengthLimit, (unsigned long long)length);
    ret = EXIT_FIRMWARE_TOO_LARGE;
    goto done;
  }
  _tprintf(_T("Board config file '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  pBuffer = (uint8_t*)malloc(length);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate board config. file buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  nRead = fread(pBuffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from board config. file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  bWasInServiceMode = avr2utilIsInServiceMode(pDevice);

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (!bVerifyOnly) {
    ret = avr2utilWriteApplicationFlash(pDevice, bVerbose, BOARD_CONFIG_BASE_ADDR, pBuffer, (uint32_t)length); /* Cast safe due to above check on 'length' */
  }
  if (EXIT_OK == ret) {
    ret = avr2utilVerifyApplicationFlash(pDevice, bVerbose, BOARD_CONFIG_BASE_ADDR, pBuffer, (uint32_t)length); /* Cast safe due to above check on 'length' */
  }

  avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, bWasInServiceMode, FALSE);

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilUpdateOrVerifyVpdCommand(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bVerifyOnly,
  const TCHAR* pFilename)
{
  const unsigned int vpdLengthLimit = VPD_LENGTH_LIMIT;
  FILE* pFile = NULL;
  uint8_t buffer[VPD_LENGTH_LIMIT];
  long length;
  size_t nRead;
  bool_t bWasInServiceMode;
  int ret = EXIT_OK;

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, TRUE);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to VPD file '%s' for reading.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of VPD file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of VPD file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of VPD file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  if ((size_t)length > vpdLengthLimit) { /* Cast to size_t safe due to positive check above */
    _tprintf(_T("*** VPD file cannot be more than 0x%X bytes; actual length is 0x%llX.\n"),
      vpdLengthLimit, (unsigned long long)length);
    ret = EXIT_FIRMWARE_TOO_LARGE;
    goto done;
  }
  _tprintf(_T("VPD file '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  memset(buffer, 0, sizeof(buffer));
  nRead = fread(buffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from VPD file '%s'.\n"), pFilename);
    ret = EXIT_READ_FIRMWARE_FAILED;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  bWasInServiceMode = avr2utilIsInServiceMode(pDevice);

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  if (EXIT_OK != ret) {
    goto done;
  }

  if (!bVerifyOnly) {
    ret = avr2utilWriteApplicationFlash(pDevice, bVerbose, VPD_BASE_ADDR, buffer, (uint32_t)length); /* Cast safe due to above check on 'length' */
  }
  if (EXIT_OK == ret) {
    ret = avr2utilVerifyApplicationFlash(pDevice, bVerbose, VPD_BASE_ADDR, buffer, (uint32_t)length); /* Cast safe due to above check on 'length' */
  }

  avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, bWasInServiceMode, FALSE);

done:
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilSaveFirmwareCommand(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  const TCHAR* pFilename)
{
  const unsigned int firmwareLengthLimit = FIRMWARE_LENGTH_LIMIT;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  size_t nWritten;
  bool_t bWasInServiceMode;
  int ret = EXIT_OK;

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, TRUE);
  if (EXIT_OK != ret) {
    goto done;
  }

  pBuffer = (uint8_t*)malloc(firmwareLengthLimit);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to buffer for firmware save data.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  bWasInServiceMode = avr2utilIsInServiceMode(pDevice);

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilReadApplicationFlash(pDevice, bVerbose, FIRMWARE_BASE_ADDR, pBuffer, firmwareLengthLimit);

  avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, bWasInServiceMode, FALSE);

  if (EXIT_OK == ret) {
    if (0 != _tfopen_s(&pFile, pFilename, _T("wb"))) {
      _tprintf(_T("*** Failed to open firmware save file '%s' for writing.\n"), pFilename);
      ret = EXIT_WRITE_FIRMWARE_FAILED;
      goto done;
    }

    nWritten = fwrite(pBuffer, 1, firmwareLengthLimit, pFile);
    if (nWritten != (size_t)firmwareLengthLimit) {
      _tprintf(_T("*** Failed to write expected number of bytes to firmware save file '%s'.\n"), pFilename);
      ret = EXIT_WRITE_FIRMWARE_FAILED;
      goto done;
    }
  }

done:
  if (NULL != pFile) {
    fclose(pFile);
  }
  if (NULL != pBuffer) {
    free(pBuffer);
  }

  return ret;
}

int
avr2utilSaveBoardConfigCommand(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  const TCHAR* pFilename)
{
  const unsigned int boardConfigLengthLimit = BOARD_CONFIG_LENGTH_LIMIT;
  FILE* pFile = NULL;
  uint8_t buffer[BOARD_CONFIG_LENGTH_LIMIT];
  size_t nWritten;
  bool_t bWasInServiceMode;
  int ret = EXIT_OK;

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, TRUE);
  if (EXIT_OK != ret) {
    goto done;
  }

  bWasInServiceMode = avr2utilIsInServiceMode(pDevice);

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilReadApplicationFlash(pDevice, bVerbose, BOARD_CONFIG_BASE_ADDR, buffer, boardConfigLengthLimit);

  avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, bWasInServiceMode, FALSE);

  if (EXIT_OK == ret) {
    if (0 != _tfopen_s(&pFile, pFilename, _T("wb"))) {
      _tprintf(_T("*** Failed to open board config. save file '%s' for writing.\n"), pFilename);
      ret = EXIT_WRITE_FIRMWARE_FAILED;
      goto done;
    }

    nWritten = fwrite(buffer, 1, boardConfigLengthLimit, pFile);
    if (nWritten != (size_t)boardConfigLengthLimit) {
      _tprintf(_T("*** Failed to write expected number of bytes to board config. save file '%s'.\n"), pFilename);
      ret = EXIT_WRITE_FIRMWARE_FAILED;
      goto done;
    }
  }

done:
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilSaveVpdCommand(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  const TCHAR* pFilename)
{
  const unsigned int vpdLengthLimit = VPD_LENGTH_LIMIT;
  FILE* pFile = NULL;
  uint8_t buffer[VPD_LENGTH_LIMIT];
  size_t nWritten;
  bool_t bWasInServiceMode;
  int ret = EXIT_OK;

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, TRUE);
  if (EXIT_OK != ret) {
    goto done;
  }

  bWasInServiceMode = avr2utilIsInServiceMode(pDevice);

  ret = avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, TRUE, FALSE);
  if (EXIT_OK != ret) {
    goto done;
  }

  ret = avr2utilReadApplicationFlash(pDevice, bVerbose, VPD_BASE_ADDR, buffer, vpdLengthLimit);

  avr2utilAutoEnterExitServiceMode(pDevice, bVerbose, bWasInServiceMode, FALSE);

  if (EXIT_OK == ret) {
    if (0 != _tfopen_s(&pFile, pFilename, _T("wb"))) {
      _tprintf(_T("*** Failed to open VPD save file '%s' for writing.\n"), pFilename);
      ret = EXIT_WRITE_FIRMWARE_FAILED;
      goto done;
    }

    nWritten = fwrite(buffer, 1, vpdLengthLimit, pFile);
    if (nWritten != (size_t)vpdLengthLimit) {
      _tprintf(_T("*** Failed to write expected number of bytes to VPD save file '%s'.\n"), pFilename);
      ret = EXIT_WRITE_FIRMWARE_FAILED;
      goto done;
    }
  }

done:
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}
