#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>

#include "exit_code.h"
#include "i2c.h"

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

int
avr2utilI2cReadToFileCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  unsigned int length,
  const TCHAR* pFilename)
{
  uint8_t command[6];
  uint8_t response[4];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  size_t position, nWritten;
  int ret = EXIT_OK;

  if (bus > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }
  if (dev > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }
  if (address > 0xFFU) {
    _tprintf(_T("*** Address must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_ADDRESS;
  }

  pBuffer = (uint8_t*)malloc(length);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate data buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The i2c-read-to-file command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  for (position = 0; position < length; position++) {
    command[0] = 0x05; /* User extended command group */
    command[1] = 0x80; /* I2CRead command */
    command[2] = (uint8_t)bus; /* I2C bus number */
    command[3] = (uint8_t)dev; /* I2C device number */
    command[4] = (uint8_t)address; /* Address within device */
    command[5] = (uint8_t)1; /* Read 1 byte */
    expResponse[0] = 0x05; /* User extended command group */
    expResponse[1] = 0x00; /* I2CRead response */

    memset(response, 0, sizeof(response));
    ret = avr2utilSendAvr2Command(
      pDevice,
      bVerbose,
      0,
      command,
      sizeof(command),
      expResponse,
      response,
      sizeof(response),
      &responseLength);
    if (EXIT_OK != ret) {
      goto out;
    }

    avr2Status = response[2];
    if (0 != avr2Status) {
      const TCHAR* pReason;

      switch (avr2Status) {
      case 1:
        pReason = _T("General failure");
        break;

      default:
        pReason = _T("** UNEXPECTED ERROR CODE **");
        break;
      }
      _tprintf(_T("*** AVR2 returned status 0x%02X for byte %llu => %s\n"),
        (unsigned int)response[2], (unsigned long long)position, pReason);
      ret = EXIT_AVR2_BAD_STATUS;
      goto out;
    }

    pBuffer[position] = response[3];

    address++;
  }

out:
  if (0 != _tfopen_s(&pFile, pFilename, _T("wt"))) {
    _tprintf(_T("*** Failed to open file '%s' for writing.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }

  nWritten = fwrite(pBuffer, 1, position, pFile);
  if (nWritten != (size_t)position) {
    _tprintf(_T("*** Failed to write expected number of bytes to file '%s'.\n"), pFilename);
    ret = EXIT_I2C_WRITE_FILE_FAILED;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilI2cVerifyWithFileCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  const TCHAR* pFilename)
{
  const unsigned int maxErrorReported = 20;
  uint8_t command[6];
  uint8_t response[4];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  FILE* pFile = NULL;
  uint8_t* pFileBuffer = NULL;
  uint8_t* pReadBuffer = NULL;
  long length, position;
  unsigned int currentAddress;
  size_t nRead;
  unsigned int errorCount = 0;
  int ret = EXIT_OK;

  if (bus > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }
  if (dev > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }
  if (address > 0xFFU) {
    _tprintf(_T("*** Address must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_ADDRESS;
  }

  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to open file '%s' for reading.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  _tprintf(_T("File '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  pReadBuffer = (uint8_t*)malloc(length);
  if (NULL == pReadBuffer) {
    _tprintf(_T("*** Failed to allocate read data buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  pFileBuffer = (uint8_t*)malloc(length);
  if (NULL == pFileBuffer) {
    _tprintf(_T("*** Failed to allocate file data buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  nRead = fread(pFileBuffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The i2c-verify-with-file command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  currentAddress = address;
  for (position = 0; position < length; position++) {
    command[0] = 0x05; /* User extended command group */
    command[1] = 0x80; /* I2CRead command */
    command[2] = (uint8_t)bus; /* I2C bus number */
    command[3] = (uint8_t)dev; /* I2C device number */
    command[4] = (uint8_t)currentAddress; /* Address within device */
    command[5] = (uint8_t)1; /* Read 1 byte */
    expResponse[0] = 0x05; /* User extended command group */
    expResponse[1] = 0x00; /* I2CRead response */

    memset(response, 0, sizeof(response));
    ret = avr2utilSendAvr2Command(
      pDevice,
      bVerbose,
      0,
      command,
      sizeof(command),
      expResponse,
      response,
      sizeof(response),
      &responseLength);
    if (EXIT_OK != ret) {
      goto done;
    }

    avr2Status = response[2];
    if (0 != avr2Status) {
      const TCHAR* pReason;

      switch (avr2Status) {
      case 1:
        pReason = _T("General failure");
        break;

      default:
        pReason = _T("** UNEXPECTED ERROR CODE **");
        break;
      }
      _tprintf(_T("*** AVR2 returned status 0x%02X for byte %ld => %s\n"), (unsigned int)response[2], position, pReason);
      ret = EXIT_AVR2_BAD_STATUS;
      goto done;
    }

    pReadBuffer[position] = response[3];

    currentAddress++;
  }

  currentAddress = address;
  for (position = 0; position < length; position++) {
    uint8_t expected = pFileBuffer[position];
    uint8_t actual = pReadBuffer[position];

    if (actual != expected) {
      if (errorCount < maxErrorReported) {
        if (expected != actual) {
          _tprintf(_T("*** Verification error at address 0x%lX; expected=0x%02X actual=0x%02X\n"),
            (unsigned long)currentAddress, (unsigned int)expected, (unsigned int)actual);
        }
      } else if (errorCount == maxErrorReported) {
        _tprintf(_T("*** More than %u verification errors detected.\n"), maxErrorReported);
      }
      errorCount++;
    }
    currentAddress++;
  }

  if (errorCount) {
    _tprintf(_T("*** Total of %u verification errors detected.\n"), errorCount);
    ret = EXIT_VERIFY_FAILED;
  } else {
    _tprintf(_T("Verified OK.\n"));
  }

done:
  if (NULL != pReadBuffer) {
    free(pReadBuffer);
  }
  if (NULL != pFileBuffer) {
    free(pFileBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilI2cWriteFromFileCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  const TCHAR* pFilename)
{
  uint8_t command[7];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  long length, position;
  size_t nRead;
  int ret = EXIT_OK;

  if (bus > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }
  if (dev > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }
  if (address > 0xFFU) {
    _tprintf(_T("*** Address must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_ADDRESS;
  }

  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to open file '%s' for reading.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  _tprintf(_T("File '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  pBuffer = (uint8_t*)malloc(length);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate data buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  nRead = fread(pBuffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from file '%s'.\n"), pFilename);
    ret = EXIT_I2C_READ_FILE_FAILED;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The i2c-write-from-file command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  for (position = 0; position < length; position++) {
    command[0] = 0x05; /* User extended command group */
    command[1] = 0x82; /* I2CWrite command */
    command[2] = (uint8_t)bus; /* I2C bus number */
    command[3] = (uint8_t)dev; /* I2C device number */
    command[4] = (uint8_t)address; /* Address within device */
    command[5] = (uint8_t)1; /* Write 1 byte */
    command[6] = pBuffer[position]; /* The byte to write */
    expResponse[0] = 0x05; /* User extended command group */
    expResponse[1] = 0x02; /* I2CWrite response */

    memset(response, 0, sizeof(response));
    ret = avr2utilSendAvr2Command(
      pDevice,
      bVerbose,
      0,
      command,
      sizeof(command),
      expResponse,
      response,
      sizeof(response),
      &responseLength);
    if (EXIT_OK != ret) {
      goto done;
    }

    avr2Status = response[2];
    if (0 != avr2Status) {
      const TCHAR* pReason;

      switch (avr2Status) {
      case 1:
        pReason = _T("General failure");
        break;

      default:
        pReason = _T("** UNEXPECTED ERROR CODE **");
        break;
      }
      _tprintf(_T("*** AVR2 returned status 0x%02X for byte %ld => %s\n"), (unsigned int)response[2], position, pReason);
      ret = EXIT_AVR2_BAD_STATUS;
      goto done;
    }

    address++;
  }

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

int
avr2utilI2cReadCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  unsigned int length)
{
  uint8_t command[6];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint8_t* pResponse = NULL;
  unsigned int position, expResponseLength;
  int ret = EXIT_OK;

  if (bus > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }
  if (dev > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }
  if (address > 0xFFU) {
    _tprintf(_T("*** Address must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_ADDRESS;
  }
  if (length > 0xFFU) {
    _tprintf(_T("*** Byte count to return must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_COUNT;
  }

  expResponseLength = length + 3;
  pResponse = (uint8_t*)malloc(expResponseLength);
  if (NULL == pResponse) {
    _tprintf(_T("*** Failed to allocate response buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The i2c-read command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x80; /* I2CRead command */
  command[2] = (uint8_t)bus; /* I2C bus number */
  command[3] = (uint8_t)dev; /* I2C device number */
  command[4] = (uint8_t)address; /* Address within device */
  command[5] = (uint8_t)length; /* Number of bytes to read, returned in response */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x00; /* I2CRead response */

  memset(pResponse, 0, expResponseLength);
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    pResponse,
    expResponseLength,
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = pResponse[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("General failure");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"),
      (unsigned int)pResponse[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  for (position = 0; position < length; position++) {
    _tprintf(_T("%02X "), pResponse[position + 3]);
  }
  _tprintf(_T("\n"));

done:
  if (NULL != pResponse) {
    free(pResponse);
  }

  return ret;
}

int
avr2utilI2cWriteCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int bus,
  unsigned int dev,
  unsigned int address,
  unsigned int length,
  const uint8_t* pBytes)
{
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint8_t* pCommand = NULL;
  unsigned int position, commandLength;
  int ret = EXIT_OK;

  if (bus > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }
  if (dev > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }
  if (address > 0xFFU) {
    _tprintf(_T("*** Address must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_ADDRESS;
  }
  if (length > 0xFFU) {
    _tprintf(_T("*** Byte count to write must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_COUNT;
  }

  commandLength = length + 6;
  pCommand = (uint8_t*)malloc(commandLength);
  if (NULL == pCommand) {
    _tprintf(_T("*** Failed to allocate command buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The i2c-write command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  pCommand[0] = 0x05; /* User extended command group */
  pCommand[1] = 0x82; /* I2CWrite command */
  pCommand[2] = (uint8_t)bus; /* I2C bus number */
  pCommand[3] = (uint8_t)dev; /* I2C device number */
  pCommand[4] = (uint8_t)address; /* Address within device */
  pCommand[5] = (uint8_t)length;
  for (position = 0; position < length; position++) {
    pCommand[position + 6] = pBytes[position];
  }
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x02; /* I2CWrite response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    pCommand,
    commandLength,
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("General failure");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

done:
  if (NULL != pCommand) {
    free(pCommand);
  }

  return ret;
}

int
avr2utilI2cWriteReadCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int bus,
  unsigned int dev,
  unsigned int writeCount,
  unsigned int readCount,
  const uint8_t* pBytes)
{
  uint8_t expResponse[2];
  uint32_t expectedResponseLength, responseLength;
  uint8_t avr2Status;
  uint8_t* pCommand = NULL;
  uint8_t* pResponse = NULL;
  unsigned int position, commandLength;
  unsigned int actualTransferCount, expectedTransferCount, minimumReadTransferCount;
  unsigned int actualReadCount;
  int ret = EXIT_OK;

  if (bus > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }
  if (dev > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }
  if (writeCount > 0xFFU) {
    _tprintf(_T("*** Byte count to write must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_COUNT;
  }
  if (readCount > 0xFFU) {
    _tprintf(_T("*** Byte count to read must be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_COUNT;
  }

  commandLength = writeCount + 6;
  pCommand = (uint8_t*)malloc(commandLength);
  if (NULL == pCommand) {
    _tprintf(_T("*** Failed to allocate command buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  expectedResponseLength = readCount + 4;
  pResponse = (uint8_t*)malloc(expectedResponseLength);
  if (NULL == pResponse) {
    _tprintf(_T("*** Failed to allocate response buffer.\n"));
    ret = EXIT_ALLOCATION_FAILED;
    goto done;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The i2c-write-read command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    ret = EXIT_WRONG_MODE;
    goto done;
  }

  pCommand[0] = 0x05; /* User extended command group */
  pCommand[1] = 0x94; /* I2CWriteRead command */
  pCommand[2] = (uint8_t)bus; /* I2C bus number */
  pCommand[3] = (uint8_t)dev; /* I2C device number */
  pCommand[4] = (uint8_t)writeCount;
  pCommand[5] = (uint8_t)readCount;
  for (position = 0; position < writeCount; position++) {
    pCommand[position + 6] = pBytes[position];
  }
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x14; /* I2CWriteRead response */

  memset(pResponse, 0, expectedResponseLength);
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    pCommand,
    commandLength,
    expResponse,
    pResponse,
    expectedResponseLength,
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = pResponse[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid I2C bus number");
      break;

    case 2:
      pReason = _T("I2C slave NACKed, or transaction timed out.");
      break;

    case 3:
      pReason = _T("Invalid write byte count.");
      break;

    case 4:
      pReason = _T("Invalid read byte count.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)pResponse[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  actualTransferCount = pResponse[3];

  expectedTransferCount = 0;
  minimumReadTransferCount = 0;
  if (writeCount) {
    minimumReadTransferCount += 1 + writeCount;
    expectedTransferCount += 1 + writeCount;
  }
  if (readCount) {
    minimumReadTransferCount += 1;
    expectedTransferCount += 1 + readCount;
  }

  if (actualTransferCount > expectedTransferCount) {
    _tprintf(_T("+++ Reponse from uC indicates that it transferred more than requested number of bytes; actual %u vs. expected %u\n"), actualTransferCount, expectedTransferCount);
    actualTransferCount = expectedTransferCount;
  } else if (actualTransferCount < expectedTransferCount) {
    _tprintf(_T("+++ Reponse from uC indicates that it transferred less than requested number of bytes; actual %u vs. expected %u\n"), actualTransferCount, expectedTransferCount);
  }

  if (writeCount) {
    if (actualTransferCount < 1 + writeCount) {
      _tprintf(_T("+++ Reponse from uC indicates that write phase was truncated.\n"));
    }
  }

  if (readCount) {
    if (actualTransferCount > minimumReadTransferCount) {
      if (actualTransferCount < expectedTransferCount) {
        _tprintf(_T("+++ Reponse from uC indicates that read phase was truncated.\n"));
      }
      actualReadCount = actualTransferCount - minimumReadTransferCount;
      for (position = 0; position < actualReadCount; position++) {
        _tprintf(_T("%02X "), pResponse[position + 4]);
      }
      _tprintf(_T("\n"));
    } else {
      _tprintf(_T("+++ Reponse from uC indicates that no bytes were successfully read.\n"));
    }
  }

done:
  if (NULL != pResponse) {
    free(pResponse);
  }

  if (NULL != pCommand) {
    free(pCommand);
  }

  return ret;
}
