/*
** lmk61e2_regmap.c - Functions for reading and writing LMK61E2 register map (.txt) files.
**
** Such files consist of a number of lines of the form
**
**   Rnn<TAB>0xaadd
**
** where:
**
**   <TAB> is a tab character, although we accept any number of consecutive white space characters
**   nn is the base 10 register index
**   aa is the base 16 register index (again), which must be the same value as nn
**   dd is the base 16 register data
**
** (C) Copyright 2022 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>
# define _stprintf_s0(buf, siz, fmt) _stprintf_s(buf, siz, fmt)
# define _stprintf_s1(buf, siz, fmt, arg1) _stprintf_s(buf, siz, fmt, arg1)
# define _stprintf_s2(buf, siz, fmt, arg1, arg2) _stprintf_s(buf, siz, fmt, arg1, arg2)
# define _stprintf_s3(buf, siz, fmt, arg1, arg2, arg3) _stprintf_s(buf, siz, fmt, arg1, arg2, arg3)

#else

/* Linux or VxWorks */
# define _T(x) x
# define _TEOF EOF
# define _TINT char
# define _tprintf printf
# define _fgettc fgetc
# define _istdigit isdigit
# define _istprint isprint
# define _istspace isspace
# define _totlower tolower
# define _stscanf_s sscanf
# define _stprintf_s0(buf, siz, fmt) sprintf(buf, fmt)
# define _stprintf_s1(buf, siz, fmt, arg1) sprintf(buf, fmt, arg1)
# define _stprintf_s2(buf, siz, fmt, arg1, arg2) sprintf(buf, fmt, arg1, arg2)
# define _stprintf_s3(buf, siz, fmt, arg1, arg2, arg3) sprintf(buf, fmt, arg1, arg2, arg3)
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>

#include <sdk_common.h>
#include <platform_endian.h>

#include "lmk61e2_regmap.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define OFFSET_OF(t, f) ((size_t)((uintptr_t)&(((t*)0)->f) - (uintptr_t)((t*)0)))

#define MAX_LINE_LENGTH (32) /* Maximum line length excluding CR / LF, excluding NUL terminator */

#define MAX_MESSAGE_LENGTH (255) /* Maximum number of characters in a message, excluding NUL terminator */

#define MAX_BAD_LINE_MESSAGE (10) /* Maximum number of messages about malformed lines to output */

static LMK61E2RegMapError
readLine(
  FILE* pFile,
  TCHAR* pLineBuffer,
  unsigned int* pLength,
  bool_t *pbBadCharacter)
{
  unsigned int pos = 0;
  LMK61E2RegMapError ret = LMK61E2RegMapOk;
  bool_t bBadCharacter = FALSE;

  while (!feof(pFile)) {
    _TINT c;
    
    c = _fgettc(pFile);
    if (c == _TEOF) {
      if (ferror(pFile)) {
        ret = LMK61E2RegMapReadError;
      }
      break;
    } else if (c == _T('\n')) {
      break;
    } else if (c == _T('\r')) {
      /* Skip CR characters */
    } else {
      if (_istprint(c)) {
        if (pos < MAX_LINE_LENGTH) {
          pLineBuffer[pos++] = (TCHAR)c;
        }
      } else {
        /* Skip other nonprintable characters and record as warning */
        bBadCharacter = TRUE;
      }
    }
  }

  pLineBuffer[pos] = _T('\0');

  *pLength = pos;
  *pbBadCharacter = bBadCharacter;

  return ret;
}

static bool_t parseLine(
  const TCHAR* pLine,
  uint8_t* pRegIndex,
  uint8_t* pRegValue)
{
  typedef enum _State { LeadingSpace, ExpectR, ExpectRegIndex, Separator, Expect0xOrNumber, ExpectRegValue, Done } State;
  State state = LeadingSpace;
  TCHAR regIndexBuf[9];
  unsigned int nRegIndexBufPos = 0;
  TCHAR regValueBuf[9];
  unsigned int nRegValueBufPos = 0;
  TCHAR c, cLower, cNext;
  unsigned int nPos = 0;
  bool_t bRegValueIsHex = FALSE;
  unsigned int nTmpValue, nTmpRegIndex, nCheckRegIndex;

  while (Done != state) {
    c = pLine[nPos];
    if (_T('\0') == c) {
      break;
    }
    cLower = _totlower(c);
    cNext = pLine[nPos + 1];

    switch (state) {
    case LeadingSpace:
      if (_istspace(c)) {
        nPos++;
      } else {
        state = ExpectR;
      }
      break;

    case ExpectR:
      if (_totlower(c) == _T('r')) {
        nPos++;
        state = ExpectRegIndex;
      } else {
        return FALSE; /* Bad line - No 'R' in Rxx value */
      }
      break;

    case ExpectRegIndex:
      if (_istdigit(c)) {
        if (nRegIndexBufPos < ARRAY_LENGTH(regIndexBuf) - 1) {
          regIndexBuf[nRegIndexBufPos] = c;
        }
        nRegIndexBufPos++;
        nPos++;
      } else if (_istspace(c)) {
        state = Separator;
      } else {
        return FALSE; /* Bad line - Not a base 10 digit */
      }
      break;

    case Separator:
      if (_istspace(c)) {
        nPos++;
      } else {
        state = Expect0xOrNumber;
      }
      break;

    case Expect0xOrNumber:
      if (_T('0') == c && _T('x') == _totlower(cNext)) {
        bRegValueIsHex = TRUE;
        nPos += 2;
      }
      state = ExpectRegValue;
      break;

    case ExpectRegValue:
      if (_istdigit(c)) {
        if (nRegValueBufPos < ARRAY_LENGTH(regValueBuf) - 1) {
          regValueBuf[nRegValueBufPos] = c;
        }
        nRegValueBufPos++;
        nPos++;
      } else if (bRegValueIsHex && cLower >= 'a' && cLower <= 'f') {
        if (nRegValueBufPos < ARRAY_LENGTH(regValueBuf) - 1) {
          regValueBuf[nRegValueBufPos] = cLower;
        }
        nRegValueBufPos++;
        nPos++;
      } else if (_istspace(c)) {
        state = Done;
      } else {
        return FALSE; /* Bad line - Not a base 10 or base 16 digit */
      }
      break;

    default:
      assert(FALSE); /* How did we get here? */
    }
  }
 
  if (ExpectRegValue != state && Done != state) {
    return FALSE; /* Bad line - Did not finish in required state */
  }

  if (0 == nRegIndexBufPos) {
    return FALSE; /* Bad line - Empty register index */
  }

  if (0 == nRegValueBufPos) {
    return FALSE; /* Bad line - Empty register value */
  }

  regIndexBuf[nRegIndexBufPos] = _T('\0');
  regValueBuf[nRegValueBufPos] = _T('\0');

  _stscanf_s(regIndexBuf, _T("%u"), &nTmpRegIndex);

  if (bRegValueIsHex) {
    _stscanf_s(regValueBuf, _T("%x"), &nTmpValue);
  } else {
    _stscanf_s(regValueBuf, _T("%u"), &nTmpValue);
  }

  nCheckRegIndex = (nTmpValue >> 8) & 0xFF;
  if (nCheckRegIndex != nTmpRegIndex) {
    return FALSE; /* Bad line - Rxx value does not match upper 16 bits of hex value */
  }

  /* Parsed line OK */
  *pRegIndex = (uint8_t)nTmpRegIndex;
  *pRegValue = (uint8_t)(nTmpValue & 0xFF);

  return TRUE;
}

LMK61E2RegMapError LMK61E2RegMapReadFromFile(
  FILE* pFile,
  LMK61E2RegMapMessageCallback* pMessageCallback,
  LMK61E2RegMap** ppRegMap)
{
  LMK61E2RegMapError eResult = LMK61E2RegMapOk;
  unsigned int nPass, nLineNumber;
  unsigned int nPairCount = 0, nPairCountAlloc = 0;
  unsigned int nBadLineCount;
  LMK61E2RegMap* pRegMap = NULL;
  TCHAR messageBuf[MAX_MESSAGE_LENGTH + 1];
  TCHAR lineBuf[MAX_LINE_LENGTH + 1];
  unsigned int nLineLength;
  bool_t bBadCharacter;

  for (nPass = 0; nPass < 2; nPass++) {
    if (fseek(pFile, 0, SEEK_SET)) {
      if (NULL != pMessageCallback) {
        _stprintf_s1(messageBuf, ARRAY_LENGTH(messageBuf), _T("Failed to seek to beginning of input file; errno=%d"), errno);
        pMessageCallback(messageBuf, LMK61E2RegMapMessageError);
      }

      eResult = LMK61E2RegMapSeekError;
      goto out;
    }

    nPairCount = 0;
    nLineNumber = 1;
    nBadLineCount = 0;

    while (!feof(pFile)) {
      uint8_t nRegIndex, nRegValue;

      eResult = readLine(pFile, lineBuf, &nLineLength, &bBadCharacter);
      if (LMK61E2RegMapOk != eResult) {
        goto out;
      }

      if (0 == nLineLength) {
        continue;
      }

#if 0
      if (0 == nPass) {
        _tprintf(_T("%s\n"), lineBuf);
      }
#endif
    
      if (parseLine(lineBuf, &nRegIndex, &nRegValue)) {
        if (1 == nPass) {
          if (nPairCount < nPairCountAlloc) { /* Check in case file changes somehow between passes */
            pRegMap->pairs[nPairCount].nRegIndex = nRegIndex;
            pRegMap->pairs[nPairCount].nRegValue = nRegValue;
          }
        }

        nPairCount++;
      } else {
        if (NULL != pMessageCallback && 1 == nPass) {
          if (nBadLineCount < MAX_BAD_LINE_MESSAGE) {
            _stprintf_s1(messageBuf, ARRAY_LENGTH(messageBuf), _T("Line %u of input file appears to be malformed - ignoring."), nLineNumber);
            pMessageCallback(messageBuf, LMK61E2RegMapMessageWarning);

            if (nBadLineCount == MAX_BAD_LINE_MESSAGE - 1) {
              _stprintf_s0(messageBuf, ARRAY_LENGTH(messageBuf), _T("NOTE: Further messages about malformed lines will be suppressed"));
              pMessageCallback(messageBuf, LMK61E2RegMapMessageInfo);
            }
          }
        }

        nBadLineCount++;
      }

      nLineNumber++;
    }

    if (0 == nPairCount) {
      eResult = LMK61E2RegMapBadFile;
      goto out;
    }

    if (0 == nPass) {
      size_t nBytesAlloc;

      nPairCountAlloc = nPairCount;
      if (nPairCountAlloc > LMK61E2_MAX_PAIR_COUNT) {
        if (NULL != pMessageCallback) {
          _stprintf_s2(
              messageBuf, ARRAY_LENGTH(messageBuf),
              _T("More than %u register / value pairs in input file - ignoring excess %u pair(s)."),
              LMK61E2_MAX_PAIR_COUNT, nPairCount - LMK61E2_MAX_PAIR_COUNT);
          pMessageCallback(messageBuf, LMK61E2RegMapMessageWarning);
        }

        nPairCountAlloc = LMK61E2_MAX_PAIR_COUNT;
      }

      nBytesAlloc = 2 + 2 * nPairCountAlloc;
      pRegMap = (LMK61E2RegMap*)malloc(nBytesAlloc);
      if (NULL == pRegMap) {
        if (NULL != pMessageCallback) {
          _stprintf_s3(
            messageBuf, ARRAY_LENGTH(messageBuf),
            _T("Failed to allocate %llu (0x%llX) bytes for holding register map data; errno=%d"),
            (unsigned long long)nBytesAlloc, (unsigned long long)nBytesAlloc, errno);
          pMessageCallback(messageBuf, LMK61E2RegMapMessageError);
        }

        return LMK61E2RegMapAllocationError;
      }

      pRegMap->nPairCount = (uint8_t)nPairCountAlloc;
    }
  }

out:
  if (LMK61E2RegMapOk == eResult) {
    *ppRegMap = pRegMap;
  } else {
    if (NULL != pRegMap) {
      free(pRegMap);
    }
  }

  return eResult;
}

LMK61E2RegMapError LMK61E2RegMapRead(
  const TCHAR* pPath,
  LMK61E2RegMapMessageCallback* pMessageCallback,
  LMK61E2RegMap** ppRegMap)
{
  LMK61E2RegMapError ret = LMK61E2RegMapOk;
  TCHAR messageBuf[MAX_MESSAGE_LENGTH + 1];
  FILE* pFile = NULL;
  int err;

#ifdef _MSC_VER
  err = _tfopen_s(&pFile, pPath, _T("rt"));
#else
# ifdef _WIN32
  pFile = _tfopen(pPath, "rb");
# else
  pFile = fopen(pPath, "r");
# endif
  err = (NULL == pFile) ? errno : 0;
#endif
  if (err) {
    switch (err) {
    case EACCES:
      return LMK61E2RegMapPermissionDenied;

    case ENOENT:
      return LMK61E2RegMapNotFound;

    default:
      if (NULL != pMessageCallback) {
        _stprintf_s1(messageBuf, ARRAY_LENGTH(messageBuf), _T("Failed to seek to beginning of input file; errno=%d"), errno);
        pMessageCallback(messageBuf, LMK61E2RegMapMessageError);
      }

      return LMK61E2RegMapOpenFailed;
    }
  }

  ret = LMK61E2RegMapReadFromFile(pFile, pMessageCallback, ppRegMap);

  fclose(pFile);

  return ret;
}

void LMK61E2RegMapDestroy(
  LMK61E2RegMap* pRegMap)
{
  free(pRegMap);
}

const TCHAR* LM61E2RegMapGetErrorDesc(
  LMK61E2RegMapError nError)
{
  const TCHAR* pszReason;

  switch (nError) {
  case LMK61E2RegMapOk:
    pszReason = _T("no error");
    break;

  case LMK61E2RegMapBadFile:
    pszReason = _T("bad register map file");
    break;

  case LMK61E2RegMapAllocationError:
    pszReason = _T("unable to allocate memory");
    break;

  case LMK61E2RegMapPermissionDenied:
    pszReason = _T("access denied");
    break;

  case LMK61E2RegMapNotFound:
    pszReason = _T("file not found");
    break;

  case LMK61E2RegMapOpenFailed:
    pszReason = _T("unexpected error opening file");
    break;

  case LMK61E2RegMapSeekError:
    pszReason = _T("seek error while reading file");
    break;

  case LMK61E2RegMapReadError:
    pszReason = _T("error while reading file");
    break;

  default:
    pszReason = _T("unknown error");
    break;
  }

  return pszReason;
}
