#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf
# define _TEOF EOF
# define _TINT int
# define _istspace isspace
# define _fgettc fgetc
# define _totlower tolower

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sdk_common.h>

#include "mez.h"
#include "exit_code.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/* Read a line from stdin, keeping at most first 'n' non-space characters */
static size_t getLine(
  TCHAR* pszLineBuf,
  size_t nBufSize)
{
  bool_t bInSpace = TRUE;
  size_t nStored = 0, nMaxStored = nBufSize - 1;

  for (;;) {
    _TINT c = _fgettc(stdin);
    if (_TEOF == c || _T('\n') == c) {
      break;
    }

    if (!_istspace(c)) {
      bInSpace = FALSE;
    }

    if (!bInSpace) {
      if (nStored < nMaxStored) {
        pszLineBuf[nStored++] = c;
      }
    }
  }

  pszLineBuf[nStored] = _T('\0');

  return nStored;
}

extern int
avr2utilForceMezPowerCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bNonVolatile,
  bool_t bConfirmed,
  unsigned int nMezIndex,
  unsigned int nReqMillivolt,
  unsigned int nMinMillivolt,
  unsigned int nMaxMillivolt)
{
  const unsigned int nByteMax = 0xFFU;
  const unsigned int nWordMax = 0xFFFFU;
  TCHAR szInput[2];
  uint8_t command[9];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  bool_t bUnforcing = FALSE;
  int ret = EXIT_OK;

  if (nReqMillivolt > nWordMax) {
    _tprintf(_T("*** Requested voltage of %u mV is not valid; maximum allowed value is %u.\n"), nReqMillivolt, nWordMax);
    return EXIT_INVALID_MEZ_VOLTAGE;
  }
  
  if (nMinMillivolt > nWordMax) {
    _tprintf(_T("*** Requested minimum voltage of %u mV is not valid; maximum allowed value is %u.\n"), nMinMillivolt, nWordMax);
    return EXIT_INVALID_MEZ_VOLTAGE;
  }
  
  if (nMaxMillivolt > nWordMax) {
    _tprintf(_T("*** Requested maximum voltage of %u mV is not valid; maximum allowed value is %u.\n"), nMaxMillivolt, nWordMax);
    return EXIT_INVALID_MEZ_VOLTAGE;
  }
  
  if (nMezIndex > nByteMax) {
    _tprintf(_T("*** Mezzanine index of %u is not valid; maximum allowed value is %u.\n"), nMezIndex, nByteMax);
    return EXIT_INVALID_MEZ_INDEX;
  }

  if (nReqMillivolt == 0 || nReqMillivolt == 0xFFFF) {
    nMinMillivolt = 0;
    nMaxMillivolt = 0;
    bUnforcing = TRUE;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The force-mez-power command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  if (!bUnforcing && !bConfirmed) {
    size_t n;

    _tprintf(_T("WARNING: Forcing Mezzanine I/O voltage may damage hardware if performed incorrectly.\n"));
    _tprintf(_T("To proceed AT OWN RISK, enter \"Y\" at the prompt: "));
    fflush(stdout);

    n = getLine(szInput, ARRAY_LENGTH(szInput));
    if (n < 1 || _totlower(szInput[0]) != _T('y')) {
      return EXIT_NOT_CONFIRMED;
    }
  }

  command[0] = 0x05; /* User extended command group */
  if (bNonVolatile) {
    command[1] = 0x99; /* ForceMezPower command */
  } else {
    command[1] = 0x8B; /* ForceMezPower NV command */
  }
  command[2] = (uint8_t)(nReqMillivolt >> 0);
  command[3] = (uint8_t)(nReqMillivolt >> 8);
  command[4] = (uint8_t)(nMinMillivolt >> 0);
  command[5] = (uint8_t)(nMinMillivolt >> 8);
  command[6] = (uint8_t)(nMaxMillivolt >> 0);
  command[7] = (uint8_t)(nMaxMillivolt >> 8);
  command[8] = (uint8_t)nMezIndex;
  expResponse[0] = 0x05; /* User extended command group */
  if (bNonVolatile) {
    expResponse[1] = 0x19; /* ForceMezPower NV response */
  } else {
    expResponse[1] = 0x0B; /* ForceMezPower response */
  }

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 0x1U:
      pReason = _T("Mezzanine index not valid for device.");
      break;

    case 0x2U:
      pReason = _T("Force mezzanine voltage feature not supported by device.");
      break;

    case 0x3U:
      pReason = _T("Requested voltage range is not valid.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  if (bUnforcing) {
    if (bNonVolatile) {
      _tprintf(_T("uC will stop forcing Mezzanine %u I/O voltage at next board power-on.\n"), nMezIndex);
    } else {
      _tprintf(_T("uC stopped forcing Mezzanine %u I/O voltage with immediate effect.\n"), nMezIndex);
    }
  } else {
    if (bNonVolatile) {
      _tprintf(_T("uC will force Mezzanine %u I/O voltage at board power-on.\n"), nMezIndex);
      _tprintf(_T("****************************************************************************\n"));
      _tprintf(_T("WARNING: Forced Mezzanine I/O voltage will persist across power cycles.\n"));
      _tprintf(_T("         If you change the Mezzanine card fitted, remember to check that the\n"));
      _tprintf(_T("         forced I/O voltage is suitable for the new Mezzanine card.\n"));
      _tprintf(_T("****************************************************************************\n"));
    } else {
      _tprintf(_T("uC now forcing Mezzanine %u I/O voltage with immediate effect.\n"), nMezIndex);
    }
  }

  if (bNonVolatile) {
    _tprintf(_T("Use \"display-sensors\" verb after power-cycling board to see actual Mezzanine %u I/O voltage.\n"), nMezIndex);
  } else {
    _tprintf(_T("Use \"display-sensors\" verb to see actual Mezzanine %u I/O voltage.\n"), nMezIndex);
  }

done:
  return ret;
}

extern int
avr2utilGetMezPowerCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bNonVolatile,
  unsigned int nMezIndex)
{
  const unsigned int nByteMax = 0xFFU;
  uint8_t command[3];
  uint8_t response[9];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint16_t nReqMillivolt, nMinMillivolt, nMaxMillivolt;
  int ret = EXIT_OK;

  if (nMezIndex > nByteMax) {
    _tprintf(_T("*** Mezzanine index of %u is not valid; maximum allowed value is %u.\n"), nMezIndex, nByteMax);
    return EXIT_INVALID_MEZ_INDEX;
  }
  
  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The force-mez-power command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  command[0] = 0x05; /* User extended command group */
  if (bNonVolatile) {
    command[1] = 0x9A; /* GetMezPower NV command */
  } else {
    command[1] = 0x96; /* GetMezPower command */
  }
  command[2] = (uint8_t)nMezIndex;
  expResponse[0] = 0x05; /* User extended command group */
  if (bNonVolatile) {
    expResponse[1] = 0x1A; /* GetMezPower NV response */
  } else {
    expResponse[1] = 0x16; /* GetMezPower response */
  }

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 0x1U:
      pReason = _T("Mezzanine index not valid for device.");
      break;

    case 0x2U:
      pReason = _T("Force mezzanine voltage feature not supported by device.");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  nReqMillivolt = ((uint16_t)response[3] << 0) | ((uint16_t)response[4] << 8);
  nMinMillivolt = ((uint16_t)response[5] << 0) | ((uint16_t)response[6] << 8);
  nMaxMillivolt = ((uint16_t)response[7] << 0) | ((uint16_t)response[8] << 8);

  if (bNonVolatile) {
    if (0 == nReqMillivolt || 0xFFFF == nReqMillivolt) {
      _tprintf(
        _T("Mezzanine %u voltage is NOT forced at power up; requested (min., max.) = %u (%u, %u) mV\n"),
        nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
    } else {
      _tprintf(
        _T("Mezzanine %u voltage is forced at power up to requested (min., max.): %u (%u, %u) mV\n"),
        nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
    }
  } else {
    if (0 == nReqMillivolt || 0xFFFF == nReqMillivolt) {
      _tprintf(
        _T("Mezzanine %u voltage currently NOT forced; requested (min., max.) = %u (%u, %u) mV\n"),
        nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
    } else {
      _tprintf(
        _T("Mezzanine %u voltage currently forced to requested (min., max.): %u (%u, %u) mV\n"),
        nMezIndex, nReqMillivolt, nMinMillivolt, nMaxMillivolt);
    }
  }

done:
  return ret;
}
