#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#include "exit_code.h"
#include "sensor.h"
#include "sensor_info.h"
#include "vpd.h"

#define SENSOR_PAGE_LENGTH_LIMIT (64U) /* Maximum possible size of sensor data page */

/* Black box pages 0 - 23 are for clobber records */
/* Black box pages 24 - 30 are for noclobber reords */
#define BLACK_BOX_NUM_CLOBBER_RECORD   (24U) /* Number of black box sensor data pages reserved for clobber records */
#define BLACK_BOX_NUM_NOCLOBBER_RECORD  (7U) /* Number of black box sensor data pages reserved for noclobber records */

static int
requestBlackBoxRecord(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nRecordIndex,
  uint8_t* pSensorPage) /* Must be at least SENSOR_PAGE_LENGTH_LIMIT (64) bytes in size */
{
  uint8_t command[3];
  uint8_t response[67];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  command[0] = 0x05; /* User Ext command group */
  command[1] = 0x8A; /* Request Black Box record command */
  command[2] = nRecordIndex;
  expResponse[0] = 0x05; /* User Ext command group */
  expResponse[1] = 0x0A; /* Request Black Box record response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Request Black Box record command\n"), avr2Status);
    return EXIT_AVR2_COMMAND_ERROR;
  }

  memcpy(pSensorPage, response + 3, SENSOR_PAGE_LENGTH_LIMIT);

  return EXIT_OK;
}

static void
displaySensorPage(
  uint32_t nDrawingNumber,
  uint8_t nBoardRevMajor,
  uint8_t buffer[64])
{
  const double zeroCelsiusInK = 273.15;
  double adcScale;
  uint32_t etc, ec;
  uint16_t adc, tmp;
  uint8_t mezStat, crdStat0, crdStat1, crdStat2, crdStat3, si5338Stat, si5338Err, adcStat, tmpStat, reserved63;
  SensorInfo sensorInfo;
  unsigned int i;

  etc = ((uint32_t)buffer[0] << 0) | ((uint32_t)buffer[1] << 8) | ((uint32_t)buffer[2] << 16) | ((uint32_t)buffer[3] << 24);
  _tprintf(_T("  ETC         = { 0x%02X 0x%02X 0x%02X 0x%02X } => %u (0x%08X)\n"),
    buffer[0], buffer[1], buffer[2], buffer[3], etc, etc);
  ec = ((uint32_t)buffer[4] << 0) | ((uint32_t)buffer[5] << 8) | ((uint32_t)buffer[6] << 16) | ((uint32_t)buffer[7] << 24);
  _tprintf(_T("  EC          = { 0x%02X 0x%02X 0x%02X 0x%02X } => %u (0x%08X)\n"),
    buffer[4], buffer[5], buffer[6], buffer[7], ec, ec);
  for (i = 0; i < 14; i++) {
    avr2utilGetSensorInfo(nDrawingNumber, nBoardRevMajor, i, &sensorInfo, &adcScale);
    adc = (uint16_t)(((uint16_t)buffer[8 + i * 2] << 0) | ((uint16_t)buffer[9 + i * 2] << 8));
    _tprintf(_T("  ADC%02u (#%02u) = { 0x%02X 0x%02X }           => 0x%04X => %.3f %s [%s]\n"),
      i, i, buffer[8 + i * 2], buffer[9 + i * 2], adc, (double)adc * adcScale,
      sensorInfo.pSensorUnit, sensorInfo.pSensorName);
    adcStat = (uint8_t)((i & 0x1U) ? (uint8_t)((buffer[54 + (i >> 1)] & 0xF0U) >> 4) : (uint8_t)((buffer[54 + (i >> 1)] & 0xFU) >> 0));
  }
  for (i = 0; i < 4; i++) {
    avr2utilGetSensorInfo(nDrawingNumber, nBoardRevMajor, i + 14, &sensorInfo, &adcScale);
    tmp = (uint16_t)(((uint16_t)buffer[36 + i * 2] << 0) | ((uint16_t)buffer[37 + i * 2] << 8));
    _tprintf(_T("  TMP%02u (#%02u) = { 0x%02X 0x%02X }           => 0x%04X => %.3f %s [%s]\n"),
      i, i + 14, buffer[36 + i * 2], buffer[37 + i * 2], tmp, (double)tmp * adcScale - zeroCelsiusInK,
      sensorInfo.pSensorUnit, sensorInfo.pSensorName);
  }
  for (i = 0; i < 4; i++) {
    mezStat = buffer[44 + i];
    _tprintf(_T("  MEZ_STAT%u   = 0x%02X                    => F/PresC%c PresB%c DataErr%c CsumErr%c I2cFail%c PwrOK%c VIOEn%c PresA%c\n"),
      i, mezStat,
      (mezStat & 0x80U) ? _T('+') : _T('-'), (mezStat & 0x40U) ? _T('+') : _T('-'),
      (mezStat & 0x20U) ? _T('+') : _T('-'), (mezStat & 0x10U) ? _T('+') : _T('-'),
      (mezStat & 0x08U) ? _T('+') : _T('-'), (mezStat & 0x04U) ? _T('+') : _T('-'),
      (mezStat & 0x02U) ? _T('+') : _T('-'), (mezStat & 0x01U) ? _T('+') : _T('-'));
  }
  crdStat0 = buffer[48];
  _tprintf(_T("  CRD_STAT0   = 0x%02X                    => PSU7OK%c PSU6OK%c PSU5OK%c PSU4OK%c PSU3OK%c PSU2OK%c PSU1OK%c PSU0OK%c\n"),
    crdStat0,
    (crdStat0 & 0x80U) ? _T('+') : _T('-'), (crdStat0 & 0x40U) ? _T('+') : _T('-'),
    (crdStat0 & 0x20U) ? _T('+') : _T('-'), (crdStat0 & 0x10U) ? _T('+') : _T('-'),
    (crdStat0 & 0x08U) ? _T('+') : _T('-'), (crdStat0 & 0x04U) ? _T('+') : _T('-'),
    (crdStat0 & 0x02U) ? _T('+') : _T('-'), (crdStat0 & 0x01U) ? _T('+') : _T('-'));
  crdStat1 = buffer[49];
  _tprintf(_T("  CRD_STAT1   = 0x%02X                    => PSUOffReq%c PSU_AUX%c Resvd5%c BrdCfgVal%c DONE3%c DONE2%c DONE1%c DONE0%c\n"),
    crdStat1,
    (crdStat1 & 0x80U) ? _T('+') : _T('-'), (crdStat1 & 0x40U) ? _T('+') : _T('-'),
    (crdStat1 & 0x20U) ? _T('+') : _T('-'), (crdStat1 & 0x10U) ? _T('+') : _T('-'),
    (crdStat1 & 0x08U) ? _T('+') : _T('-'), (crdStat1 & 0x04U) ? _T('+') : _T('-'),
    (crdStat1 & 0x02U) ? _T('+') : _T('-'), (crdStat1 & 0x01U) ? _T('+') : _T('-'));
  crdStat2 = buffer[50];
  _tprintf(_T("  CRD_STAT2   = 0x%02X                    => HostRst%c USBCon%c GAP%c GA=%u(0x%x)\n"),
    crdStat2,
    (crdStat2 & 0x80U) ? _T('+') : _T('-'), (crdStat2 & 0x40U) ? _T('+') : _T('-'),
    (crdStat2 & 0x20U) ? _T('+') : _T('-'), crdStat2 & 0x1FU, crdStat2 & 0x1FU);
  crdStat3 = buffer[51];
  _tprintf(_T("  CRD_STAT3   = 0x%02X                       (reserved)\n"), crdStat3);
  si5338Stat = buffer[52];
  _tprintf(_T("  SI5338_STAT = 0x%02X                    => CFG3OK%c CFG2OK%c CFG1OK%c CFG0OK%c Resvd3%c Resvd2%c IntRegd%c InitOK%c\n"),
    si5338Stat,
    (si5338Stat & 0x80U) ? _T('+') : _T('-'), (si5338Stat & 0x40U) ? _T('+') : _T('-'),
    (si5338Stat & 0x20U) ? _T('+') : _T('-'), (si5338Stat & 0x10U) ? _T('+') : _T('-'),
    (si5338Stat & 0x08U) ? _T('+') : _T('-'), (si5338Stat & 0x04U) ? _T('+') : _T('-'),
    (si5338Stat & 0x02U) ? _T('+') : _T('-'), (si5338Stat & 0x01U) ? _T('+') : _T('-'));
  si5338Err = buffer[53];
  _tprintf(_T("  SI5338_ERR  = 0x%02X                    => ErrType=%u ErrCode=%u\n"),
    si5338Err, (si5338Err >> 4) & 0xFU, (si5338Err >> 0) & 0xFU);
  for (i = 0; i < 14; i++) {
    adcStat = (uint8_t)((i & 0x1U) ? (uint8_t)((buffer[54 + (i >> 1)] & 0xF0U) >> 4) : (uint8_t)((buffer[54 + (i >> 1)] & 0xFU) >> 0));
    _tprintf(_T("  ADC_STAT%02u  = 0x%X                     => OK%c Critical%c Over%c Under%c\n"),
      i, adcStat,
      (adcStat & 0x08U) ? _T('+') : _T('-'), (adcStat & 0x04U) ? _T('+') : _T('-'),
      (adcStat & 0x02U) ? _T('+') : _T('-'), (adcStat & 0x01U) ? _T('+') : _T('-'));
  }
  for (i = 0; i < 4; i++) {
    tmpStat = (uint8_t)((i & 0x1U) ? (uint8_t)((buffer[61 + (i >> 1)] & 0xF0U) >> 4) : (uint8_t)((buffer[61 + (i >> 1)] & 0xFU) >> 0));
    _tprintf(_T("  TMP_STAT%02u  = 0x%X                     => OK%c Critical%c Over%c Under%c\n"),
      i, tmpStat,
      (tmpStat & 0x08U) ? _T('+') : _T('-'), (tmpStat & 0x04U) ? _T('+') : _T('-'),
      (tmpStat & 0x02U) ? _T('+') : _T('-'), (tmpStat & 0x01U) ? _T('+') : _T('-'));
  }
  reserved63 = buffer[63];
  _tprintf(_T("  RESVD63     = 0x%02X                       (reserved)\n"), reserved63);
}

/* Returns true if etc1 > etc2 */
static bool_t
recordIsLater(
  uint32_t etc1,
  uint32_t etc2)
{
  if (etc1 == 0xFFFFFFFF) {
    /* Invalid timestamp, etc2 must be later */
    return FALSE;
  }
  if (etc2 == 0xFFFFFFFF) {
    /* Invalid timestamp, etc1 must be later */
    return TRUE;
  }
  return (etc1 > etc2) ? TRUE : FALSE;
}

int
avr2utilDisplayBlackBoxCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int nTypes,
  unsigned int nMaxRecordDisplayed)
{
  typedef struct _Record {
    uint32_t nPageIndex;
    uint8_t page[SENSOR_PAGE_LENGTH_LIMIT];
  } Record;
  Record clobberRecords[BLACK_BOX_NUM_CLOBBER_RECORD];
  Record noclobberRecords[BLACK_BOX_NUM_NOCLOBBER_RECORD];
  Record tmpRecord[SENSOR_PAGE_LENGTH_LIMIT];
  uint8_t vpdBuffer[256]; /* Request VPD command requires buffer of 256 bytes */
  uint8_t nRecordIndex, nCompareIndex;
  Record* pRecord;
  Record* pCompareRecord;
  uint32_t etc, compareEtc;
  uint32_t nDrawingNumber;
  uint8_t nBoardRevMajor;
  bool_t bNeedBlankLine = FALSE;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The display-sensors-raw command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  /* In Application Mode, request VPD using command */
  ret = avr2utilRequestVpd(pDevice, bVerbose, vpdBuffer);
  if (ret) {
    return ret;
  }

  if (nTypes & 0x1) {
    for (nRecordIndex = 0; nRecordIndex < BLACK_BOX_NUM_CLOBBER_RECORD; nRecordIndex++) {
      ret = requestBlackBoxRecord(pDevice, bVerbose, nRecordIndex, clobberRecords[nRecordIndex].page);
      if (EXIT_OK != ret) {
        _tprintf(_T("*** Failed to read black-box clobber record with index %u.\n"), nRecordIndex);
        goto done;
      }
      clobberRecords[nRecordIndex].nPageIndex = nRecordIndex;
    }
  }

  if (nTypes & 0x2) {
    for (nRecordIndex = 0; nRecordIndex < BLACK_BOX_NUM_NOCLOBBER_RECORD; nRecordIndex++) {
      ret = requestBlackBoxRecord(pDevice, bVerbose, (uint8_t)(nRecordIndex + BLACK_BOX_NUM_CLOBBER_RECORD), noclobberRecords[nRecordIndex].page);
      if (EXIT_OK != ret) {
        _tprintf(_T("*** Failed to read black-box noclobber record with index %u.\n"), nRecordIndex + BLACK_BOX_NUM_CLOBBER_RECORD);
        goto done;
      }
      noclobberRecords[nRecordIndex].nPageIndex = nRecordIndex + BLACK_BOX_NUM_CLOBBER_RECORD;
    }
  }

done:
  nDrawingNumber = GET_VPD_VAL32(&vpdBuffer[4]);
  avr2utilGetBoardRevision(vpdBuffer, nDrawingNumber, &nBoardRevMajor, NULL);

  if (nTypes & 0x1) {
    /* Sort clobber records by timestamp, from latest to oldest. */
    for (nRecordIndex = 0; nRecordIndex < BLACK_BOX_NUM_CLOBBER_RECORD; nRecordIndex++) {
      pRecord = &clobberRecords[nRecordIndex];
      etc = ((uint32_t)pRecord->page[0] << 0) | ((uint32_t)pRecord->page[1] << 8) | ((uint32_t)pRecord->page[2] << 16) | ((uint32_t)pRecord->page[3] << 24);

      for (nCompareIndex = (uint8_t)(nRecordIndex + 1); nCompareIndex < BLACK_BOX_NUM_CLOBBER_RECORD; nCompareIndex++) {
        pCompareRecord = &clobberRecords[nCompareIndex];
        compareEtc = ((uint32_t)pCompareRecord->page[0] << 0) | ((uint32_t)pCompareRecord->page[1] << 8) | ((uint32_t)pCompareRecord->page[2] << 16) | ((uint32_t)pCompareRecord->page[3] << 24);

        if (recordIsLater(compareEtc, etc)) {
          memcpy(tmpRecord, pRecord, sizeof(Record));
          memcpy(pRecord, pCompareRecord, sizeof(Record));
          memcpy(pCompareRecord, tmpRecord, sizeof(Record));
          etc = compareEtc;
        }
      }
    }

    /* Display sorted clobber records. */
    for (nRecordIndex = 0; nRecordIndex < BLACK_BOX_NUM_CLOBBER_RECORD && (nMaxRecordDisplayed == 0 || nRecordIndex < nMaxRecordDisplayed); nRecordIndex++) {
      pRecord = &clobberRecords[nRecordIndex];
      etc = ((uint32_t)pRecord->page[0] << 0) | ((uint32_t)pRecord->page[1] << 8) | ((uint32_t)pRecord->page[2] << 16) | ((uint32_t)pRecord->page[3] << 24);
      if (0xFFFFFFFF == etc) {
        /* Max. timestamp => record not valid */
        break;
      }

      if (0 == nRecordIndex) {
        _tprintf(_T("-------------------------------------------------------------------------------\n"));
      }
      _tprintf(_T("Black box clobber record (page %u):\n"), pRecord->nPageIndex);
      displaySensorPage(nDrawingNumber, nBoardRevMajor, pRecord->page);
      _tprintf(_T("-------------------------------------------------------------------------------\n"));
    }

    if (0 == nRecordIndex) {
      _tprintf(_T("No black box clobber records found.\n"));
    } else {
      bNeedBlankLine = TRUE;
    }
  }

  if (nTypes & 0x2) {
    /* Sort noclobber records by timestamp, from latest to oldest. */
    for (nRecordIndex = 0; nRecordIndex < BLACK_BOX_NUM_NOCLOBBER_RECORD; nRecordIndex++) {
      pRecord = &noclobberRecords[nRecordIndex];
      etc = ((uint32_t)pRecord->page[0] << 0) | ((uint32_t)pRecord->page[1] << 8) | ((uint32_t)pRecord->page[2] << 16) | ((uint32_t)pRecord->page[3] << 24);

      for (nCompareIndex = (uint8_t)(nRecordIndex + 1); nCompareIndex < BLACK_BOX_NUM_NOCLOBBER_RECORD; nCompareIndex++) {
        pCompareRecord = &noclobberRecords[nCompareIndex];
        compareEtc = ((uint32_t)pCompareRecord->page[0] << 0) | ((uint32_t)pCompareRecord->page[1] << 8) | ((uint32_t)pCompareRecord->page[2] << 16) | ((uint32_t)pCompareRecord->page[3] << 24);

        if (recordIsLater(compareEtc, etc)) {
          memcpy(tmpRecord, pRecord, sizeof(Record));
          memcpy(pRecord, pCompareRecord, sizeof(Record));
          memcpy(pCompareRecord, tmpRecord, sizeof(Record));
          etc = compareEtc;
        }
      }
    }

    /* Display sorted noclobber records. */
    for (nRecordIndex = 0; nRecordIndex < BLACK_BOX_NUM_NOCLOBBER_RECORD && (nMaxRecordDisplayed == 0 || nRecordIndex < nMaxRecordDisplayed); nRecordIndex++) {
      pRecord = &noclobberRecords[nRecordIndex];
      etc = ((uint32_t)pRecord->page[0] << 0) | ((uint32_t)pRecord->page[1] << 8) | ((uint32_t)pRecord->page[2] << 16) | ((uint32_t)pRecord->page[3] << 24);
      if (0xFFFFFFFF == etc) {
        /* Max. timestamp => record not valid */
        break;
      }

      if (0 == nRecordIndex) {
        if (bNeedBlankLine) {
          _tprintf(_T("\n"));
          bNeedBlankLine = FALSE;
        }
        _tprintf(_T("-------------------------------------------------------------------------------\n"));
      }
      _tprintf(_T("Black box noclobber record (page %u):\n"), pRecord->nPageIndex);
      displaySensorPage(nDrawingNumber, nBoardRevMajor, pRecord->page);
      _tprintf(_T("-------------------------------------------------------------------------------\n"));
    }

    if (0 == nRecordIndex) {
      _tprintf(_T("No black box noclobber records found.\n"));
    }
  }

  return ret;
}

static int
requestSensors(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t* pSensorPage) /* Must be at least SENSOR_PAGE_LENGTH_LIMIT (64) bytes in size */
{
  uint8_t command[2];
  uint8_t response[67];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  command[0] = 0x05; /* User Ext command group */
  command[1] = 0x89; /* Request Sensors command */
  expResponse[0] = 0x05; /* User Ext command group */
  expResponse[1] = 0x09; /* Request Sensors response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Request Sensors command\n"), avr2Status);
    return EXIT_AVR2_COMMAND_ERROR;
  }

  memcpy(pSensorPage, response + 3, SENSOR_PAGE_LENGTH_LIMIT);

  return EXIT_OK;
}

int
avr2utilDisplaySensorsCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose)
{
  uint8_t buffer[SENSOR_PAGE_LENGTH_LIMIT]; /* Request Sensors command requires buffer of SENSOR_PAGE_LENGTH_LIMIT (64) bytes */
  uint8_t vpdBuffer[256]; /* Request VPD command requires buffer of 256 bytes */
  uint32_t nDrawingNumber;
  uint8_t nBoardRevMajor;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The display-sensors command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  /* In Application Mode, request VPD using command */
  ret = avr2utilRequestVpd(pDevice, bVerbose, vpdBuffer);
  if (ret) {
    goto done;
  }

  /* In Application Mode, request sensor data using command */
  ret = requestSensors(pDevice, bVerbose, buffer);

done:
  if (EXIT_OK == ret) {
    nDrawingNumber = GET_VPD_VAL32(&vpdBuffer[4]);
    avr2utilGetBoardRevision(vpdBuffer, nDrawingNumber, &nBoardRevMajor, NULL);

    _tprintf(_T("Human-readable Sensor Page:\n"));
    displaySensorPage(nDrawingNumber, nBoardRevMajor, buffer);
  }

  return ret;
}

int
avr2utilDisplaySensorsRawCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose)
{
  uint8_t sensorPage[SENSOR_PAGE_LENGTH_LIMIT]; /* Request Sensors command requires buffer of SENSOR_PAGE_LENGTH_LIMIT (64) bytes */
  unsigned int i, j;
  int ret = EXIT_OK;

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The display-sensors-raw command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  /* In Application Mode, request sensor data using command */
  ret = requestSensors(pDevice, bVerbose, sensorPage);

  if (EXIT_OK == ret) {
    _tprintf(_T("Raw Sensor Page:\n"));
    _tprintf(_T("     "));
    for (i = 0; i < 16; i++) {
      _tprintf(_T(" +%X"), i);
    }
    _tprintf(_T("\n"));
    for (i = 0; i < SENSOR_PAGE_LENGTH_LIMIT; i += 16) {
      _tprintf(_T("0x%02X:"), i);
      for (j = 0; j < 16; j++) {
        _tprintf(_T(" %02X"), sensorPage[i + j]);
      }
      _tprintf(_T(" "));
      for (j = 0; j < 16; j++) {
        if (isprint(sensorPage[i + j])) {
          printf("%c", sensorPage[i + j]);
        } else {
          printf(".");
        }
      }
      _tprintf(_T("\n"));
    }
  }

  return ret;
}

static int
overrideSensor(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t sensorIndex,
  uint32_t overrideValue,
  bool_t bEnable) 
{
  uint8_t command[8];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;
  
  command[0] = 0x05; /* User extended command group */
  command[1] = 0x81; /* OverrideSensor command */
  command[2] = (uint8_t)sensorIndex; /* Index of sensor to override (model-specific) */
  command[3] = (uint8_t)((overrideValue >> 0) & 0xFFU); /* Override Value bits [7:0] (byte 0) */
  command[4] = (uint8_t)((overrideValue >> 8) & 0xFFU); /* Override Value bits [15:8] (byte 1) */
  command[5] = (uint8_t)((overrideValue >> 16) & 0xFFU); /* Override Value bits [23:16] (byte 2) */
  command[6] = (uint8_t)((overrideValue >> 24) & 0xFFU); /* Override Value bits [31:24] (byte 3) */
  command[7] = (uint8_t)(bEnable ? 0x1U : 0U); /* Enable byte */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x01; /* OverrideSensor response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid sensor index");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    return EXIT_AVR2_BAD_STATUS;
  }

  return ret;
}

int
avr2utilOverrideSensorUnscaledCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int sensorIndex,
  unsigned int overrideValue)
{
  int ret = EXIT_OK;

  if (sensorIndex >= 0xFFU) {
    _tprintf(_T("*** Sensor index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The override-sensor command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }
  
  ret = overrideSensor(pDevice, bVerbose, (uint8_t)sensorIndex /* cast safe due to range check above */, overrideValue, TRUE);

  return ret;
}

int
avr2utilReleaseSensorCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int sensorIndex)
{
  int ret = EXIT_OK;

  if (sensorIndex >= 0xFFU) {
    _tprintf(_T("*** Sensor index must be in the range 0 to 255.\n"));
    return EXIT_INVALID_INDEX;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The release-sensor command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }
  
  ret = overrideSensor(pDevice, bVerbose, (uint8_t)sensorIndex /* cast safe due to range check above */, 0, FALSE /* Stop overriding */);

  return ret;
}
