#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>

#include "exit_code.h"
#include "smbus.h"

int
avr2utilGetSMBUSChipAddrCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bGetNonvolatile)
{
  uint8_t command[4];
  uint8_t response[4];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  uint8_t nChipAddr;
  int ret = EXIT_OK;

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x93; /* Get SMBUS chip address command */
  command[2] = (uint8_t)(bGetNonvolatile ? 0x2 : 0x1); /* 0x1 => get address volatile, 0x2 => get address nonvolatile */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x13; /* Set SMBUS chip address response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned error status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  nChipAddr = response[3];

  if (bGetNonvolatile) {
    if (0 == nChipAddr) {
      _tprintf(_T("Nonvolatile SMBUS chip address is 0.\n"));
      _tprintf(_T("The microcontroller will disable its SMBUS interface when it next reboots.\n"));
    } else {
      _tprintf(_T("Nonvolatile SMBUS chip address is 0x%02X.\n"), nChipAddr);
      _tprintf(_T("The microcontroller will enable its SMBUS interface when it next reboots.\n"));
    }
  } else {
    if (0 == nChipAddr) {
      _tprintf(_T("Current SMBUS chip address is 0.\n"));
      _tprintf(_T("The microcontroller's SMBUS interface is currently disabled.\n"));
    } else {
      _tprintf(_T("Current SMBUS chip address is 0x%02X.\n"), nChipAddr);
      _tprintf(_T("The microcontroller's SMBUS interface is currently enabled.\n"));
    }
  }

done:
  return ret;
}

int
avr2utilSetSMBUSChipAddrCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bSetNonvolatile,
  unsigned int nChipAddr)
{
  uint8_t command[4];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  if (nChipAddr > 0x7FU) {
    _tprintf(_T("*** SMBUS chip address be in the range 0 to 127(0x7F).\n"));
    return EXIT_INVALID_INDEX;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x92; /* Set SMBUS chip address command */
  command[2] = (uint8_t)(bSetNonvolatile ? 0x2 : 0x1); /* 0x1 => set address volatile, 0x2 => set address nonvolatile */
  command[3] = (uint8_t)nChipAddr; /* 7-bit SMBUS chip address; cast safe due to above range check */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x12; /* Set SMBUS chip address response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2CommandEx(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    /* If changing volatile SMBUS chip address, need to change address that we use between command and response... */
    (uint8_t)(bSetNonvolatile ? 0 : nChipAddr), /* Cast safe due to above range check */
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid SMBUS chip address");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned error status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  if (bSetNonvolatile) {
    if (0 == nChipAddr) {
      _tprintf(_T("The microcontroller's nonvolatile SMBUS chip address has been set to 0.\n"));
      _tprintf(_T("The microcontroller will disable its SMBUS interface when it next reboots.\n"));
    } else {
      _tprintf(_T("The microcontroller's nonvolatile SMBUS chip address has been set to 0x%02X.\n"), nChipAddr);
      _tprintf(_T("The microcontroller will enable its SMBUS interface when it next reboots.\n"));
    }

  } else {
    if (0 == nChipAddr) {
      _tprintf(_T("The microcontroller's volatile SMBUS chip address has been set to 0.\n"));
      _tprintf(_T("This has disabled the microcontroller's SMBUS interface.\n"));
    } else {
      _tprintf(_T("The microcontroller's volatile SMBUS chip address has been set to 0x%02X.\n"), nChipAddr);
      _tprintf(_T("This has enabled the microcontroller's SMBUS interface.\n"));
    }
    _tprintf(_T("This condition will persist until the microcontroller next reboots.\n"));
  }

  _tprintf(_T("The microcontroller can be made to reboot by either (a) removing and then\n"));
  _tprintf(_T("restoring all power including standby power, or (b) entering and exiting Service\n"));
  _tprintf(_T("Mode.\n"));

done:
  return ret;
}
