#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sdk_common.h>

#include "exit_code.h"
#include "srv_mode.h"

int
avr2utilEnterExitServiceMode(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bEnterNotExit,
  bool_t bTestOnly)
{
  const TCHAR* pDescription = bEnterNotExit ? _T("Enter Service Mode") : _T("Exit Service Mode");
  uint8_t command[2];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  bool_t bNowInServiceMode;
  int ret = EXIT_OK;

  if ((bEnterNotExit && !avr2utilAllowEnterServiceMode(pDevice)) || (!bEnterNotExit && !avr2utilAllowExitServiceMode(pDevice))) {
    _tprintf(_T("*** %s via %s is not supported.\n"), pDescription, avr2utilGetInterfaceName(pDevice));
    return EXIT_MODE_CHANGE_FAILED;
  }

  if (bTestOnly) {
    /* Caller just wants to know if mode change will be allowed */
    return EXIT_OK;
  }

  if (bEnterNotExit) {
    _tprintf(_T("Entering service mode: "));
  } else {
    _tprintf(_T("Exiting service mode: "));
  }
  fflush(stdout);

  command[0] = 0x03; /* User command group */
  expResponse[0] = 0x03; /* User command group */
  if (bEnterNotExit) {
    command[1] = 0x80; /* ServiceMode command */
    expResponse[1] = 0x00; /* ServiceMode acknowledge response */
  } else {
    command[1] = 0x82; /* ApplicationMode command */
    expResponse[1] = 0x02; /* ApplicationMode acknowledge response */
  }

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for %s command\n"),
      (unsigned int)avr2Status, pDescription);
    return EXIT_AVR2_COMMAND_ERROR;
  }

  /* Give the uC enough time to reboot itself */
  avr2utilRebootDelay(pDevice);

  /* Verify that the uC has switched to the opposite mode */
  avr2utilDetermineServiceMode(pDevice, bVerbose);

  bNowInServiceMode = avr2utilIsInServiceMode(pDevice);
  if (bEnterNotExit) {
    if (!bNowInServiceMode) {
      _tprintf(_T("*** uC did not enter Service Mode as requested.\n"));
      return EXIT_MODE_CHANGE_FAILED;
    }
  } else {
    if (bNowInServiceMode) {
      _tprintf(_T("*** uC did not exit Service Mode as requested.\n"));
      return EXIT_MODE_CHANGE_FAILED;
    }
  }

  _tprintf(_T("OK\n"));

  return EXIT_OK;
}

int
avr2utilAutoEnterExitServiceMode(
  DeviceHandle* pDevice,
  bool_t bVerbose,
  bool_t bEnterNotExit,
  bool_t bTestOnly)
{
  if (avr2utilIsInServiceMode(pDevice) == bEnterNotExit) {
    return EXIT_OK;
  }

  if (!avr2utilIsAutoEnterServiceMode(pDevice)) {
    _tprintf(_T("*** Automatically %s Service Mode via %s is not supported.\n"),
      bEnterNotExit ? _T("entering") : _T("exiting"), avr2utilGetInterfaceName(pDevice));
    return EXIT_MODE_CHANGE_FAILED;
  }

  return avr2utilEnterExitServiceMode(pDevice, bVerbose, bEnterNotExit, bTestOnly);
}
