#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif
#include <ctype.h>

#include <sdk_common.h>
#include <platform_endian.h>

#include "exit_code.h"
#include "firmware.h"
#include "fpga_info.h"
#include "model_info.h"
#include "vpd.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

int
avr2utilRequestVpd(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t* pVpdData) /* Must be at least 256 bytes in size */
{
  uint8_t command[2];
  uint8_t response[259];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = 0;

  command[0] = 0x05; /* User Ext command group */
  command[1] = 0x8C; /* Request VPD command */
  expResponse[0] = 0x05; /* User Ext command group */
  expResponse[1] = 0x0C; /* Request VPD response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Request VPD command\n"), avr2Status);
    return EXIT_AVR2_COMMAND_ERROR;
  }

  memcpy(pVpdData, response + 3, 256);

  return EXIT_OK;
}

void
avr2utilGetBoardRevision(
  const uint8_t *pVpd,
  uint32_t nDrawingNumber,
  uint8_t* pnBoardRevMajor, 
  uint8_t* pnBoardRevMinor)
{
  uint8_t nPcbRev = 0;

  switch (nDrawingNumber) {
  case DRAWING_NUM_ADMXRCKU1:
  case DRAWING_NUM_ADMXRCKU1AT8:
    nPcbRev = GET_VPD_VAL8(&pVpd[88]);
    break;

  case DRAWING_NUM_ADMPCIE8K5:
  case DRAWING_NUM_ADMPCIEUCC:
  case DRAWING_NUM_ADMPCIE8K5_FH:
  case DRAWING_NUM_ADMPCIE8V3:
  case DRAWING_NUM_ADMPCIE9V3:
  case DRAWING_NUM_ADMPCIE9V3G:
  case DRAWING_NUM_ADMPCIE9H3:
  case DRAWING_NUM_ADMPCIE9H7:
  case DRAWING_NUM_ADMPCIE9V5:
  case DRAWING_NUM_ADMPCIE9V6:
  case DRAWING_NUM_ADMSDEVBASE:
  case DRAWING_NUM_ADMVPX39V2:
  case DRAWING_NUM_ADMPCIE9V7:
  case DRAWING_NUM_ADMPA100:
  case DRAWING_NUM_ADMPA101:
  case DRAWING_NUM_ADMPCIETRC01:
  case DRAWING_NUM_ADMPCIE9V8:
  case DRAWING_NUM_ADMAD01474:
    nPcbRev = GET_VPD_VAL8(&pVpd[12]);
    break;

  case DRAWING_NUM_ADMVPX39Z2:
    nPcbRev = GET_VPD_VAL8(&pVpd[62]);
    break;

  case DRAWING_NUM_ADMXRC9R1:
  case DRAWING_NUM_ADMXRC9Z1:
    nPcbRev = GET_VPD_VAL8(&pVpd[64]);
    break;

  case DRAWING_NUM_ADCXMCSTANDALONE:
	  nPcbRev = GET_VPD_VAL8(&pVpd[14]);
    break;

  default:
    break;
  }

  if (NULL != pnBoardRevMajor) {
    *pnBoardRevMajor = (uint8_t)((nPcbRev >> 4) & 0xFU);
  }
  if (NULL != pnBoardRevMinor) {
    *pnBoardRevMinor = (uint8_t)((nPcbRev >> 0) & 0xFU);
  }
}

int
avr2utilDisplayVpdCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose)
{
  const unsigned int vpdLengthLimit = VPD_LENGTH_LIMIT;
  const unsigned int Gi = 0x40000000U;
  const unsigned int Mi = 0x100000U;
  uint8_t buffer[256]; /* Request VPD command requires buffer of 256 bytes */
  int ret = EXIT_OK;
  const TCHAR* pModelName;
  uint8_t checksum, calcChecksum;
  uint8_t pcbRev;
  uint16_t version, length, highestVersionRecognized = 0;
  uint32_t drawingNumber, serialNumber;
  char fpgaStepping[9], fpgaSpeedGrade[5];
  uint16_t fpgaDevice;
  uint8_t fpgaTempGrade;
  uint8_t ddr4Class;
  unsigned int i, bank;

  memset(buffer, 0, 256);

  if (avr2utilIsInServiceMode(pDevice)) {
    /* In Service Mode, get VPD by reading AVR Flash */
    ret = avr2utilReadApplicationFlash(pDevice, bVerbose, VPD_BASE_ADDR, buffer, vpdLengthLimit);
  } else {
    /* In Application Mode, request VPD using command */
    ret = avr2utilRequestVpd(pDevice, bVerbose, buffer);
  }

  if (ret) {
    goto done;
  }

  version = GET_VPD_VAL16(&buffer[0]);
  length = GET_VPD_VAL16(&buffer[2]);
  drawingNumber = GET_VPD_VAL32(&buffer[4]);
  pModelName = avr2utilGetModelName(drawingNumber);
  serialNumber = GET_VPD_VAL32(&buffer[8]);

  _tprintf(_T("Human-readable VPD:\n"));
  _tprintf(_T("  Data structure version             %u (0x%04X)\n"), (unsigned int)version, (unsigned int)version);
  _tprintf(_T("  Data structure length              %u (0x%04X)\n"), (unsigned int)length, (unsigned int)length);
  _tprintf(_T("  Drawing number                     %u (0x%08X) => %s\n"), drawingNumber, drawingNumber, pModelName);
  _tprintf(_T("  Serial number                      %u (0x%08X)\n"), serialNumber, serialNumber);

  switch (drawingNumber) {
  case DRAWING_NUM_ADMXRCKU1:
  case DRAWING_NUM_ADMXRCKU1AT8:
    {
      const TCHAR* coolingTypeDescs[] = { _T("Air-cooled commerical"), _T("/AC1 - air-cooled industrial"), _T("/CC1 - conduction-cooled industrial") };
      const TCHAR* coatingTypeDescs[] = { _T("No coating"), _T("/A - Acrylic (Humiseal 1B31)"), _T("/P - Polyurethane (Conathane CE-1155)") };
      const TCHAR* pn4ConnectorDescs[] = { _T("NOT FITTED"), _T("FITTED") };
      const TCHAR* pCoolingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pCoatingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pPn4ConnectorDesc = _T("UNRECOGNIZED");
      uint8_t coolingType, coatingType, pn4Connector;
      uint32_t modifications32, si5338RefClkFreq;
      uint16_t dramFreq[4];
      uint8_t dramOrder[4];
      uint64_t dramSizeBytes;

      highestVersionRecognized = 0;
      modifications32 = GET_VPD_VAL32(&buffer[12]);
      _tprintf(_T("  Modifications bitmask              %u (0x%08X)\n"), modifications32, modifications32);
      if (DRAWING_NUM_ADMXRCKU1AT8 == drawingNumber) {
        _tprintf(_T("                                     => SW1 (DIP switch): %s\n"), (modifications32 & 0x1U) ? _T("FITTED") : _T("NOT FITTED"));
      }
      coolingType = GET_VPD_VAL8(&buffer[16]);
      pCoolingTypeDesc = (coolingType < ARRAY_LENGTH(coolingTypeDescs)) ? coolingTypeDescs[coolingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Cooling type                       %u (0x%02X) => %s\n"),
        (unsigned int)coolingType, (unsigned int)coolingType, pCoolingTypeDesc);
      coatingType = GET_VPD_VAL8(&buffer[17]);
      pCoatingTypeDesc = (coatingType < ARRAY_LENGTH(coatingTypeDescs)) ? coatingTypeDescs[coatingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Coating type                       %u (0x%02X) => %s\n"),
        (unsigned int)coatingType, (unsigned int)coatingType, pCoatingTypeDesc);
      pn4Connector = GET_VPD_VAL8(&buffer[18]);
      pPn4ConnectorDesc = (pn4Connector < ARRAY_LENGTH(pn4ConnectorDescs)) ? pn4ConnectorDescs[pn4Connector] : _T("UNRECOGNIZED");
      _tprintf(_T("  Pn4 connector                      %u (0x%02X) => %s\n"),
        (unsigned int)pn4Connector, (unsigned int)pn4Connector, pPn4ConnectorDesc);
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[20]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      for (bank = 0; bank < 4; bank++) {
        dramFreq[bank] = GET_VPD_VAL16(&buffer[40 + 8 * bank]);
        dramOrder[bank] = GET_VPD_VAL8(&buffer[42 + 8 * bank]);
        _tprintf(_T("  DDR4 SDRAM bank %u max. frequency   %u (0x%04X) => %u MHz / DDR4-%u\n"),
          bank, (unsigned int)dramFreq[bank], (unsigned int)dramFreq[bank],
          (unsigned int)(dramFreq[bank] + 5) / 10, (unsigned int)(dramFreq[bank] + 2) / 5);
        dramSizeBytes = ((uint64_t)1 << dramOrder[bank]) /* no. of words */ * 4U; /* bytes per word */
        _tprintf(_T("  DDR4 SDRAM bank %u size             %u (0x%02X) => "),
          bank, (unsigned int)dramOrder[bank], (unsigned int)dramOrder[bank]);
        if (dramSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
        }
      }
      memcpy(fpgaStepping, &buffer[72], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[80], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[84]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[86]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
      pcbRev = GET_VPD_VAL8(&buffer[88]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
    }
    break;

  case DRAWING_NUM_ADMPCIE8K5:
  case DRAWING_NUM_ADMPCIEUCC:
  case DRAWING_NUM_ADMPCIE8K5_FH:
  case DRAWING_NUM_ADMPCIE8V3:
  case DRAWING_NUM_ADMPCIE9V3:
  case DRAWING_NUM_ADMPCIE9V3G:
    {
      uint8_t modifications8;
      uint32_t si5338RefClkFreq, emcclkFreq, fabricClkFreq;
      uint16_t dramFreq[4];
      uint8_t dramOrder[4], dramRanks[4];
      uint8_t customBuildCode, customBuildRevision;

      highestVersionRecognized = 0;
      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"),
        (unsigned int)modifications8, (unsigned int)modifications8);
      if (drawingNumber == DRAWING_NUM_ADMPCIE8K5 ||
          drawingNumber == DRAWING_NUM_ADMPCIEUCC ||
          drawingNumber == DRAWING_NUM_ADMPCIE8K5_FH) {
        _tprintf(_T("                                     => RS232 converter:          %s\n"), (modifications8 & 0x1U) ? _T("FITTED") : _T("NOT FITTED"));
        _tprintf(_T("                                     => SI5328 jitter attenuator: %s\n"), (modifications8 & 0x2U) ? _T("FITTED") : _T("NOT FITTED"));
        _tprintf(_T("                                     => FireFly modules:          %s\n"), (modifications8 & 0x4U) ? _T("FITTED") : _T("NOT FITTED"));
      } else if (drawingNumber == DRAWING_NUM_ADMPCIE8V3) {
        _tprintf(_T("                                     => RS232 converter:          %s\n"), (modifications8 & 0x1U) ? _T("FITTED") : _T("NOT FITTED"));
        _tprintf(_T("                                     => zQSFP modules:            %s\n"), (modifications8 & 0x2U) ? _T("FITTED") : _T("NOT FITTED"));
        _tprintf(_T("                                     => FireFly modules:          %s\n"), (modifications8 & 0x4U) ? _T("FITTED") : _T("NOT FITTED"));
      }
      if (drawingNumber == DRAWING_NUM_ADMPCIE8K5 ||
          drawingNumber == DRAWING_NUM_ADMPCIEUCC ||
          drawingNumber == DRAWING_NUM_ADMPCIE8K5_FH || 
          drawingNumber == DRAWING_NUM_ADMPCIE8V3 ||
          drawingNumber == DRAWING_NUM_ADMPCIE9V3)
      {
        ddr4Class = (uint8_t)((modifications8 >> 4) & 0xFU);
        switch (ddr4Class) {
        case 0:
          _tprintf(_T("                                     => DDR4-2400 CL16 (0000)\n"));
          break;

        case 1:
          _tprintf(_T("                                     => DDR4-2400 CL17 (0001)\n"));
          break;

        default:
          _tprintf(_T("                                     => UNRECOGNISED DDR4 CHIP CLASS %u(0x%X)\n"),
            (unsigned int)ddr4Class, (unsigned int)ddr4Class);
          break;
        }
      }
      if (drawingNumber == DRAWING_NUM_ADMPCIE8K5 ||
          drawingNumber == DRAWING_NUM_ADMPCIEUCC ||
          drawingNumber == DRAWING_NUM_ADMPCIE8K5_FH ||
          drawingNumber == DRAWING_NUM_ADMPCIE9V3 ||
          drawingNumber == DRAWING_NUM_ADMPCIE9V3G)
      {
        customBuildCode = GET_VPD_VAL8(&buffer[14]);
        customBuildRevision = GET_VPD_VAL8(&buffer[15]);
        _tprintf(_T("  Custom build code, revision        %u, %u => "),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
        if (0 == customBuildCode) {
          _tprintf(_T("Standard build\n"));
        } else {
          _tprintf(_T("/C%u.%u\n"),
            (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
        }
      }
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[16]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      emcclkFreq = GET_VPD_VAL32(&buffer[36]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      if (drawingNumber == DRAWING_NUM_ADMPCIE8K5 ||
          drawingNumber == DRAWING_NUM_ADMPCIEUCC ||
          drawingNumber == DRAWING_NUM_ADMPCIE8K5_FH)
      {
        fabricClkFreq = GET_VPD_VAL32(&buffer[40]);
        _tprintf(_T("  Fabric clock frequency             %u (0x%08X) Hz\n"), fabricClkFreq, fabricClkFreq);
      }
      if (drawingNumber == DRAWING_NUM_ADMPCIE8K5 ||
          drawingNumber == DRAWING_NUM_ADMPCIE8V3 ||
          drawingNumber == DRAWING_NUM_ADMPCIEUCC ||
          drawingNumber == DRAWING_NUM_ADMPCIE8K5_FH ||
          drawingNumber == DRAWING_NUM_ADMPCIE9V3) {
        for (bank = 0; bank < 2; bank++) {
          uint64_t dramSizeBytes;

          dramFreq[bank] = GET_VPD_VAL16(&buffer[48 + 8 * bank]);
          dramOrder[bank] = GET_VPD_VAL8(&buffer[50 + 8 * bank]);
          dramRanks[bank] = GET_VPD_VAL8(&buffer[51 + 8 * bank]);
          _tprintf(_T("  DDR4 SDRAM bank %u max. frequency   %u (0x%04X) => %u MHz / DDR4-%u\n"),
            bank, (unsigned int)dramFreq[bank], (unsigned int)dramFreq[bank],
            (unsigned int)((dramFreq[bank] + 5) / 10), (unsigned int)((dramFreq[bank] + 2) / 5));
          dramSizeBytes = ((uint64_t)1 << dramOrder[bank]) /* no. of words */ * 8U; /* bytes per word */
          _tprintf(_T("  DDR4 SDRAM bank %u size             %u (0x%02X) => "),
            bank, (unsigned int)dramOrder[bank], (unsigned int)dramOrder[bank]);
          if (dramSizeBytes >= Gi) {
            _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
          } else {
            _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
          }
          _tprintf(_T("  DDR4 SDRAM bank %u ranks            %u (0x%02X) => "),
            (unsigned int)bank, (unsigned int)dramRanks[bank], (unsigned int)dramRanks[bank]);
          if (dramRanks[bank] & 0x80U) {
            _tprintf(_T("multi-die, %u dies\n"), (unsigned int)(dramRanks[bank] & 0x7F));
            if ((dramRanks[bank] & 0x7F) < 2) {
              _tprintf(_T("*** ERROR: Multi-die flag 0x80 is present, but no. of dies < 2\n"));
            }
          } else {
            if ((dramRanks[bank] & 0x7FU) == 1) {
              _tprintf(_T("single-die\n"));
            } else if ((dramRanks[bank] & 0x7FU) == 2) {
              _tprintf(_T("traditional twin-die\n"));
            } else {
              _tprintf(_T("traditional %u dies\n"), (unsigned int)(dramRanks[bank] & 0x7F));
              _tprintf(_T("*** ERROR: Multi-die flag 0x80 is absent, but no. of dies > 2\n"));
            }
          }
        }
      }
      memcpy(fpgaStepping, &buffer[64], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[72], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[76]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[78]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
    }
    break;

  case DRAWING_NUM_ADMPCIE9H3:
  case DRAWING_NUM_ADMPCIE9H7:
    {
      uint8_t modifications8;
      uint32_t si5338RefClkFreq, emcclkFreq;
      uint16_t hbmFreq;
      uint8_t hbmDieOrder, hbmDieCount;
      uint64_t hbmDieSizeBytes;
      uint8_t customBuildCode, customBuildRevision;

      highestVersionRecognized = 0;
      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"),
        (unsigned int)modifications8, (unsigned int)modifications8);
      if (DRAWING_NUM_ADMPCIE9H7 == drawingNumber) {
        _tprintf(_T("                                     => SI5328 jitter attenuator: %s\n"), (modifications8 & 0x1U) ? _T("FITTED") : _T("NOT FITTED"));
        _tprintf(_T("                                     => Active cooling fans:      %s\n"), (modifications8 & 0x2U) ? _T("FITTED") : _T("NOT FITTED"));
      }
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[16]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      emcclkFreq = GET_VPD_VAL32(&buffer[20]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      hbmFreq = GET_VPD_VAL16(&buffer[24]);
      hbmDieCount = GET_VPD_VAL8(&buffer[26]);
      hbmDieOrder = GET_VPD_VAL8(&buffer[27]);
      _tprintf(_T("  HBM max. frequency                 %u (0x%04X) => %u MHz\n"),
        (unsigned int)hbmFreq, (unsigned int)hbmFreq, (unsigned int)(hbmFreq / 10));
      hbmDieSizeBytes = (uint64_t)1 << hbmDieOrder;
      _tprintf(_T("  HBM die capacity                   %u (0x%02X) => "),
        (unsigned int)hbmDieOrder, (unsigned int)hbmDieOrder);
      if (hbmDieSizeBytes >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(hbmDieSizeBytes / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(hbmDieSizeBytes / Mi));
      }
      _tprintf(_T("  HBM die count                      %u (0x%02X)\n"),
        (unsigned int)hbmDieCount, (unsigned int)hbmDieCount);
      memcpy(fpgaStepping, &buffer[32], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[40], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[44]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[46]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
    }
    break;

  case DRAWING_NUM_ADMPCIE9V5:
  case DRAWING_NUM_ADMPCIE9V6:
    {
      uint8_t modifications8;
      uint32_t si5338RefClkFreq, emcclkFreq;
      uint8_t customBuildCode, customBuildRevision;

      highestVersionRecognized = 0;
      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"),
        (unsigned int)modifications8, (unsigned int)modifications8);
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[16]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      emcclkFreq = GET_VPD_VAL32(&buffer[20]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      memcpy(fpgaStepping, &buffer[32], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[40], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[44]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[46]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
    }
    break;

  case DRAWING_NUM_ADMSDEVBASE:
    {
      uint32_t si5338RefClkFreq;
      uint8_t modifications8, customBuildCode, customBuildRevision, fpgaSpaceGrade;

      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"),
        (unsigned int)modifications8, (unsigned int)modifications8);
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[16]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      memcpy(fpgaStepping, &buffer[64], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[72], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[76]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[78]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
      fpgaSpaceGrade = GET_VPD_VAL8(&buffer[79]);
      _tprintf(_T("  FPGA space grade                   %u (0x%02X) => %s\n"),
        (unsigned int)fpgaSpaceGrade, (unsigned int)fpgaSpaceGrade, avr2utilGetFpgaSpaceGradeString(fpgaSpaceGrade));
    }
    break;

  case DRAWING_NUM_ADMVPX39Z2:
    {
      const TCHAR* coolingTypeDescs[] = { _T("Air-cooled commerical"), _T("/AC1 - air-cooled industrial"), _T("/CC1 - conduction-cooled industrial") };
      const TCHAR* coatingTypeDescs[] = { _T("No coating"), _T("/A - Acrylic (Humiseal 1B31)"), _T("/P - Polyurethane (Conathane CE-1155)") };
      const TCHAR* pCoolingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pCoatingTypeDesc = _T("UNRECOGNIZED");
      uint8_t psQspiFlashOrder, customBuildCode, customBuildRevision, dramOrder, dramRanks;
      uint8_t psEmmcFlashOrder, psSataFlashOrder, coolingType, coatingType;
      uint16_t modifications16, dramFreq;
      uint32_t si5338RefClkFreq;
      uint64_t psQspiFlashSize, dramSizeBytes, psEmmcFlashSize, psSataFlashSize;

      modifications16 = GET_VPD_VAL16(&buffer[12]);
      _tprintf(_T("  Modifications bitmask              %u (0x%08X)\n"), modifications16, modifications16);
      _tprintf(_T("                                     => Optics: %s\n"), (modifications16 & 0x1U) ? _T("FITTED") : _T("NOT FITTED"));
      _tprintf(_T("                                     => VCC_PSBATT: %s\n"), (modifications16 & 0x2U) ? _T("ENABLED") : _T("DISABLED"));
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[16]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      for (bank = 0; bank < 2; bank++) {
        psQspiFlashOrder = GET_VPD_VAL8(&buffer[28 + 2 * bank]);
        psQspiFlashSize = (uint64_t)((uint32_t)1 << psQspiFlashOrder);
        _tprintf(_T("  PS QSPI Flash bank %u size Log 2    %u (0x%02X) => "), bank, psQspiFlashOrder, psQspiFlashOrder);
        if (psQspiFlashSize >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(psQspiFlashSize / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(psQspiFlashSize / Mi));
        }
      }
      dramFreq = GET_VPD_VAL16(&buffer[32]);
      dramOrder = GET_VPD_VAL8(&buffer[34]);
      dramRanks = GET_VPD_VAL8(&buffer[35]);
      _tprintf(_T("  PS DDR4 SDRAM max. frequency       %u (0x%04X) => %u MHz / DDR4-%u\n"),
        (unsigned int)dramFreq, (unsigned int)dramFreq,
        (unsigned int)((dramFreq + 5) / 10), (unsigned int)((dramFreq + 2) / 5));
      dramSizeBytes = ((uint64_t)1 << dramOrder) /* no. of words */ * 8U; /* bytes per word */
      _tprintf(_T("  PS DDR4 SDRAM 64b word size log 2  %u (0x%02X) => "),
        (unsigned int)dramOrder, (unsigned int)dramOrder);
      if (dramSizeBytes >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
      }
      _tprintf(_T("  PS DDR4 SDRAM ranks                %u (0x%02X) => "),
        (unsigned int)dramRanks, (unsigned int)dramRanks);
      if (dramRanks & 0x80U) {
        _tprintf(_T("multi-die, %u dies\n"), (unsigned int)(dramRanks & 0x7F));
        if ((dramRanks & 0x7F) < 2) {
          _tprintf(_T("*** ERROR: Multi-die flag 0x80 is present, but no. of dies < 2\n"));
        }
      } else {
        if ((dramRanks & 0x7FU) == 1) {
          _tprintf(_T("single-die\n"));
        } else if ((dramRanks & 0x7FU) == 2) {
          _tprintf(_T("traditional twin-die\n"));
        } else {
          _tprintf(_T("traditional %u dies\n"), (unsigned int)(dramRanks & 0x7F));
          _tprintf(_T("*** ERROR: Multi-die flag 0x80 is absent, but no. of dies > 2\n"));
        }
      }
      psEmmcFlashOrder = GET_VPD_VAL8(&buffer[40]);
      psEmmcFlashSize = (uint64_t)((uint64_t)1 << psEmmcFlashOrder);
      _tprintf(_T("  PS EMMC Flash size log 2           %u (0x%02X) => "), psEmmcFlashOrder, psEmmcFlashOrder);
      if (psEmmcFlashSize >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(psEmmcFlashSize / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(psEmmcFlashSize / Mi));
      }
      psSataFlashOrder = GET_VPD_VAL8(&buffer[42]);
      psSataFlashSize = (uint64_t)((uint64_t)1 << psSataFlashOrder);
      _tprintf(_T("  PS SATA Flash size log 2           %u (0x%02X) => "), psSataFlashOrder, psSataFlashOrder);
      if (psSataFlashSize >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(psSataFlashSize / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(psSataFlashSize / Mi));
      }
      coolingType = GET_VPD_VAL8(&buffer[44]);
      pCoolingTypeDesc = (coolingType < ARRAY_LENGTH(coolingTypeDescs)) ? coolingTypeDescs[coolingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Cooling type                       %u (0x%02X) => %s\n"),
        (unsigned int)coolingType, (unsigned int)coolingType, pCoolingTypeDesc);
      coatingType = GET_VPD_VAL8(&buffer[45]);
      pCoatingTypeDesc = (coatingType < ARRAY_LENGTH(coatingTypeDescs)) ? coatingTypeDescs[coatingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Coating type                       %u (0x%02X) => %s\n"),
        (unsigned int)coatingType, (unsigned int)coatingType, pCoatingTypeDesc);
      memcpy(fpgaStepping, &buffer[46], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[54], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[58]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[60]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
      pcbRev = GET_VPD_VAL8(&buffer[62]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
    }
    break;

  case DRAWING_NUM_ADMXRC9R1:
    {
      const TCHAR* coolingTypeDescs[] = {
          _T("Air-cooled commerical"),
          _T("/AC1 - air-cooled industrial"),
          _T("/CC1 - conduction-cooled industrial")
      };
      const TCHAR* coatingTypeDescs[] = {
        _T("No coating"),
        _T("/A - Acrylic"),
        _T("/P - Polyurethane") };
      const TCHAR* rfConnectorDescs[] = {
        _T("Right-angle Nicomatic (part CMM342D000F51-0020-240002)"),
        _T("Straight Nicomatic pins (part 22-1300-12)"),
        _T("Not fitted")
      };
      const TCHAR* adcTransformerDescs[] = {
        _T("MiniCircuits TCM2-43X+"),
        _T("MiniCircuits TCM2-672X+"),
        _T("MiniCircuits TCM2-63WX+")
      };
      const TCHAR* dacTransformerDescs[] = {
        _T("MiniCircuits TCM2-43X+"),
        _T("MiniCircuits TCM2-672X+"),
        _T("MiniCircuits TCM2-63WX+")
      };
      const TCHAR* fpgaPackageDescs[] = { _T("FFVE1156"), _T("FFSE1156") };
      const TCHAR* pCoolingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pCoatingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pRfConnectorDesc = _T("UNRECOGNIZED");
      const TCHAR* pAdcTransformerDesc = _T("UNRECOGNIZED");
      const TCHAR* pDacTransformerDesc = _T("UNRECOGNIZED");
      const TCHAR* pFpgaPackageDesc = _T("UNRECOGNIZED");
      uint8_t customBuildCode, customBuildRevision, dramOrder, dramRanks, fpgaPackage, cpldRev;
      uint8_t coolingType, coatingType, rfConnectorType;
      uint16_t dramFreq;
      uint32_t val32, modifications32, si5338RefClkFreq, psRefClkFreq;
      uint64_t dramSizeBytes;
      unsigned int bank;

      modifications32 = GET_VPD_VAL32(&buffer[12]);
      _tprintf(_T("  Modifications bitmask              %u (0x%08X)\n"), modifications32, modifications32);
      _tprintf(_T("                                     => Pn6 connector: %s\n"), (modifications32 & 0x1U) ? _T("NOT FITTED") : _T("FITTED"));
      _tprintf(_T("                                     => Pn4 connector: %s\n"), (modifications32 & 0x2U) ? _T("NOT FITTED") : _T("FITTED"));
      val32 = (modifications32 >> 3) & 0x3;
      if (val32 < ARRAY_LENGTH(adcTransformerDescs)) {
        pAdcTransformerDesc = adcTransformerDescs[val32];
      }
      _tprintf(_T("                                     => ADC RF Transformer: %s\n"), pAdcTransformerDesc);
      val32 = (modifications32 >> 10) & 0x3;
      if (val32 < ARRAY_LENGTH(dacTransformerDescs)) {
        pDacTransformerDesc = dacTransformerDescs[val32];
      }
      _tprintf(_T("                                     => DAC RF Transformer: %s\n"), pDacTransformerDesc);
      rfConnectorType = (uint8_t)((modifications32 >> 5) & 0x3);
      pRfConnectorDesc = (rfConnectorType < ARRAY_LENGTH(rfConnectorDescs)) ? rfConnectorDescs[rfConnectorType] : _T("UNRECOGNIZED");
      _tprintf(_T("                                     => RF connector: %s\n"), pRfConnectorDesc);
      ddr4Class = (uint8_t)((modifications32 >> 6) & 0xFU);
      switch (ddr4Class) {
      case 0:
        _tprintf(_T("                                     => DDR4-2400 CL16 (0000)\n"));
        break;

      case 1:
        _tprintf(_T("                                     => DDR4-2400 CL17 (0001)\n"));
        break;

      default:
        _tprintf(_T("                                     => UNRECOGNISED DDR4 CHIP CLASS %u(0x%X)\n"), (unsigned int)ddr4Class, (unsigned int)ddr4Class);
        break;
      }
      customBuildCode = GET_VPD_VAL8(&buffer[16]);
      customBuildRevision = GET_VPD_VAL8(&buffer[17]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      coolingType = GET_VPD_VAL8(&buffer[18]);
      pCoolingTypeDesc = (coolingType < ARRAY_LENGTH(coolingTypeDescs)) ? coolingTypeDescs[coolingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Cooling type                       %u (0x%02X) => %s\n"),
        (unsigned int)coolingType, (unsigned int)coolingType, pCoolingTypeDesc);
      coatingType = GET_VPD_VAL8(&buffer[19]);
      pCoatingTypeDesc = (coatingType < ARRAY_LENGTH(coatingTypeDescs)) ? coatingTypeDescs[coatingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Coating type                       %u (0x%02X) => %s\n"),
        (unsigned int)coatingType, (unsigned int)coatingType, pCoatingTypeDesc);
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[20]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      psRefClkFreq = GET_VPD_VAL32(&buffer[24]);
      _tprintf(_T("  PS ref. clock frequency            %u (0x%08X) Hz\n"), psRefClkFreq, psRefClkFreq);
      dramFreq = GET_VPD_VAL16(&buffer[28]);
      dramOrder = GET_VPD_VAL8(&buffer[30]);
      dramRanks = GET_VPD_VAL8(&buffer[31]);
      _tprintf(_T("  PS DDR4 SDRAM max. frequency       %u (0x%04X) => %u MHz / DDR4-%u\n"),
        (unsigned int)dramFreq, (unsigned int)dramFreq,
        (unsigned int)((dramFreq + 5) / 10), (unsigned int)((dramFreq + 2) / 5));
      dramSizeBytes = ((uint64_t)1 << dramOrder) /* no. of words */ * 4U; /* bytes per word */
      _tprintf(_T("  PS DDR4 SDRAM 32b word size log 2  %u (0x%02X) => "),
        (unsigned int)dramOrder, (unsigned int)dramOrder);
      if (dramSizeBytes >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
      }
      for (bank = 0; bank < 2; bank++) {
        dramFreq = GET_VPD_VAL16(&buffer[32 + 4 * bank]);
        dramOrder = GET_VPD_VAL8(&buffer[34 + 4 * bank]);
        dramRanks = GET_VPD_VAL8(&buffer[35 + 4 * bank]);
        _tprintf(_T("  PL DDR4 SDRAM bank %u max. freq.    %u (0x%04X) => %u MHz / DDR4-%u\n"),
          bank, (unsigned int)dramFreq, (unsigned int)dramFreq,
          (unsigned int)((dramFreq + 5) / 10), (unsigned int)((dramFreq + 2) / 5));
        dramSizeBytes = ((uint64_t)1 << dramOrder) /* no. of words */ * 1U; /* bytes per word */
        _tprintf(_T("  PL DDR4 SDRAM bank %u size log 2    %u (0x%02X) => "),
          bank, (unsigned int)dramOrder, (unsigned int)dramOrder);
        if (dramSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
        }
        _tprintf(_T("  PL DDR4 SDRAM bank %u ranks         %u (0x%02X) => "),
          bank, (unsigned int)dramRanks, (unsigned int)dramRanks);
        if (dramRanks & 0x80U) {
          _tprintf(_T("multi-die, %u dies\n"), (unsigned int)(dramRanks & 0x7F));
          if ((dramRanks & 0x7F) < 2) {
            _tprintf(_T("*** ERROR: Multi-die flag 0x80 is present, but no. of dies < 2\n"));
          }
        } else {
          if ((dramRanks & 0x7FU) == 1) {
            _tprintf(_T("single-die\n"));
          } else if ((dramRanks & 0x7FU) == 2) {
            _tprintf(_T("traditional twin-die\n"));
          } else {
            _tprintf(_T("traditional %u dies\n"), (unsigned int)(dramRanks & 0x7F));
            _tprintf(_T("*** ERROR: Multi-die flag 0x80 is absent, but no. of dies > 2\n"));
          }
        }
      }
      val32 = GET_VPD_VAL32(&buffer[40]);
      _tprintf(_T("  LMK04208 ref. clock frequency      %u (0x%X) Hz\n"), val32, val32);
      val32 = GET_VPD_VAL32(&buffer[44]);
      _tprintf(_T("  LMK04208 VCXO clock frequency      %u (0x%X) Hz\n"), val32, val32);
      memcpy(fpgaStepping, &buffer[48], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[56], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[60]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[62]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
      fpgaPackage = GET_VPD_VAL8(&buffer[63]);
      pFpgaPackageDesc = (fpgaPackage < ARRAY_LENGTH(fpgaPackageDescs)) ? fpgaPackageDescs[fpgaPackage] : _T("UNRECOGNIZED");
      _tprintf(_T("  FPGA package                       %u (0x%02X) => %s\n"),
        (unsigned int)fpgaPackage, (unsigned int)fpgaPackage, pFpgaPackageDesc);
      pcbRev = GET_VPD_VAL8(&buffer[64]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      cpldRev = GET_VPD_VAL8(&buffer[65]);
      _tprintf(_T("  CPLD revision                      %u (0x%02X) => %u.%u\n"),
        (unsigned int)cpldRev, (unsigned int)cpldRev, (unsigned int)((cpldRev >> 4) & 0xFU), (unsigned int)(cpldRev & 0xFU));
    }
    break;

  case DRAWING_NUM_ADMXRC9Z1:
    {
      const TCHAR* coolingTypeDescs[] = {
          _T("Air-cooled commerical"),
          _T("/AC1 - air-cooled industrial"),
          _T("/CC1 - conduction-cooled industrial")
      };
      const TCHAR* coatingTypeDescs[] = {
        _T("No coating"),
        _T("/A - Acrylic"),
        _T("/P - Polyurethane") };
      const TCHAR* pCoolingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pCoatingTypeDesc = _T("UNRECOGNIZED");
      uint8_t customBuildCode, customBuildRevision, dramOrder;
      uint8_t coolingType, coatingType, emmcOrder, qspiOrder;
      uint16_t dramFreq;
      uint32_t modifications32, si5338RefClkFreq, psRefClkFreq;
      uint64_t dramSizeBytes, emmcSizeBytes, qspiSizeBytes;
      unsigned int bank;

      modifications32 = GET_VPD_VAL32(&buffer[12]);
      _tprintf(_T("  Modifications bitmask              %u (0x%08X)\n"), modifications32, modifications32);
      _tprintf(_T("                                     => Pn6 connector: %s\n"), (modifications32 & 0x1U) ? _T("NOT FITTED") : _T("FITTED"));
      _tprintf(_T("                                     => Pn4 connector: %s\n"), (modifications32 & 0x2U) ? _T("NOT FITTED") : _T("FITTED"));
      customBuildCode = GET_VPD_VAL8(&buffer[16]);
      customBuildRevision = GET_VPD_VAL8(&buffer[17]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      coolingType = GET_VPD_VAL8(&buffer[18]);
      pCoolingTypeDesc = (coolingType < ARRAY_LENGTH(coolingTypeDescs)) ? coolingTypeDescs[coolingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Cooling type                       %u (0x%02X) => %s\n"),
        (unsigned int)coolingType, (unsigned int)coolingType, pCoolingTypeDesc);
      coatingType = GET_VPD_VAL8(&buffer[19]);
      pCoatingTypeDesc = (coatingType < ARRAY_LENGTH(coatingTypeDescs)) ? coatingTypeDescs[coatingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Coating type                       %u (0x%02X) => %s\n"),
        (unsigned int)coatingType, (unsigned int)coatingType, pCoatingTypeDesc);
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[20]);
      _tprintf(_T("  SI5338 ref. clock frequency        %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      psRefClkFreq = GET_VPD_VAL32(&buffer[24]);
      _tprintf(_T("  PS ref. clock frequency            %u (0x%08X) Hz\n"), psRefClkFreq, psRefClkFreq);
      dramFreq = GET_VPD_VAL16(&buffer[28]);
      dramOrder = GET_VPD_VAL8(&buffer[30]);
      _tprintf(_T("  PS DDR4 SDRAM max. frequency       %u (0x%04X) => %u MHz / DDR4-%u\n"),
        (unsigned int)dramFreq, (unsigned int)dramFreq,
        (unsigned int)((dramFreq + 5) / 10), (unsigned int)((dramFreq + 2) / 5));
      dramSizeBytes = ((uint64_t)1 << dramOrder) /* no. of words */ * 4U; /* bytes per word */
      _tprintf(_T("  PS DDR4 SDRAM 32b word size log 2  %u (0x%02X) => "),
        (unsigned int)dramOrder, (unsigned int)dramOrder);
      if (dramSizeBytes >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
      }
      for (bank = 0; bank < 2; bank++) {
        dramFreq = GET_VPD_VAL16(&buffer[32 + 4 * bank]);
        dramOrder = GET_VPD_VAL8(&buffer[34 + 4 * bank]);
        _tprintf(_T("  PL DDR4 SDRAM bank %u max. freq.    %u (0x%04X) => %u MHz / DDR4-%u\n"),
          bank, (unsigned int)dramFreq, (unsigned int)dramFreq,
          (unsigned int)((dramFreq + 5) / 10), (unsigned int)((dramFreq + 2) / 5));
        dramSizeBytes = ((uint64_t)1 << dramOrder) /* no. of words */ * 1U; /* bytes per word */
        _tprintf(_T("  PL DDR4 SDRAM bank %u size log 2    %u (0x%02X) => "),
          bank, (unsigned int)dramOrder, (unsigned int)dramOrder);
        if (dramSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
        }
      }
      emmcOrder = GET_VPD_VAL8(&buffer[40]);
      emmcSizeBytes = (uint64_t)1 << emmcOrder;
      _tprintf(_T("  PS EMMC memory size log 2          %u (0x%02X) => "), (unsigned int)emmcOrder, (unsigned int)emmcOrder);
      if (0 == emmcSizeBytes) {
        _tprintf(_T("NOT FITTED\n"));
      } else {
        if (emmcSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(emmcSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(emmcSizeBytes / Mi));
        }
      }
      qspiOrder = GET_VPD_VAL8(&buffer[42]);
      qspiSizeBytes = (uint64_t)1 << qspiOrder;
      _tprintf(_T("  PS QSPI flash memory size log 2    %u (0x%02X) => "), (unsigned int)qspiOrder, (unsigned int)qspiOrder);
      if (qspiSizeBytes >= Gi) {
        _tprintf(_T("%u GiB\n"), (unsigned int)(qspiSizeBytes / Gi));
      } else {
        _tprintf(_T("%u MiB\n"), (unsigned int)(qspiSizeBytes / Mi));
      }
      memcpy(fpgaStepping, &buffer[48], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[56], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[60]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[62]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
      pcbRev = GET_VPD_VAL8(&buffer[64]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
    }
    break;

  case DRAWING_NUM_ADMVPX39V2:
    {
      const TCHAR* coolingTypeDescs[] = { _T("Air-cooled commerical"), _T("/AC1 - air-cooled industrial"), _T("/CC1 - conduction-cooled industrial") };
      const TCHAR* coatingTypeDescs[] = { _T("No coating"), _T("/A - Acrylic (Humiseal 1B31)"), _T("/P - Polyurethane (Conathane CE-1155)") };
      const TCHAR* pCoolingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pCoatingTypeDesc = _T("UNRECOGNIZED");
      uint8_t modifications8, coolingType, coatingType;
      uint8_t customBuildCode, customBuildRevision;
      uint32_t si5338RefClkFreq, emcclkFreq;
      uint16_t dramFreq[4];
      uint8_t dramOrder[4];
      uint64_t dramSizeBytes;

      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%08X)\n"), modifications8, modifications8);
      _tprintf(_T("                                     => FireFly connector: %s\n"), (modifications8 & 0x1U) ? _T("FITTED") : _T("NOT FITTED"));
      _tprintf(_T("                                     => SI5338 ref. clock: %s\n"), (modifications8 & 0x2U) ? _T("From VPX conn.") : _T("Internal"));
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      coolingType = GET_VPD_VAL8(&buffer[16]);
      pCoolingTypeDesc = (coolingType < ARRAY_LENGTH(coolingTypeDescs)) ? coolingTypeDescs[coolingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Cooling type                       %u (0x%02X) => %s\n"),
        (unsigned int)coolingType, (unsigned int)coolingType, pCoolingTypeDesc);
      coatingType = GET_VPD_VAL8(&buffer[17]);
      pCoatingTypeDesc = (coatingType < ARRAY_LENGTH(coatingTypeDescs)) ? coatingTypeDescs[coatingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Coating type                       %u (0x%02X) => %s\n"),
        (unsigned int)coatingType, (unsigned int)coatingType, pCoatingTypeDesc);
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[32]);
      _tprintf(_T("  Internal SI5338 ref. clock freq.   %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      emcclkFreq = GET_VPD_VAL32(&buffer[36]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      for (bank = 0; bank < 4; bank++) {
        dramFreq[bank] = GET_VPD_VAL16(&buffer[48 + 4 * bank]);
        dramOrder[bank] = GET_VPD_VAL8(&buffer[50 + 4 * bank]);
        _tprintf(_T("  DDR4 SDRAM bank %u max. frequency   %u (0x%04X) => %u MHz / DDR4-%u\n"),
          bank, (unsigned int)dramFreq[bank], (unsigned int)dramFreq[bank],
          (unsigned int)(dramFreq[bank] + 5) / 10, (unsigned int)(dramFreq[bank] + 2) / 5);
        dramSizeBytes = ((uint64_t)1 << dramOrder[bank]) /* no. of words */ * 8U; /* bytes per word */
        _tprintf(_T("  DDR4 SDRAM bank %u size             %u (0x%02X) => "),
          bank, (unsigned int)dramOrder[bank], (unsigned int)dramOrder[bank]);
        if (dramSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
        }
      }
      memcpy(fpgaStepping, &buffer[64], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[72], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[76]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[78]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
    }
    break;

  case DRAWING_NUM_ADMPCIE9V7:
    {
      const TCHAR* coolingTypeDescs[] = { _T("Air-cooled commerical"), _T("/AC1 - air-cooled industrial"), _T("/CC1 - conduction-cooled industrial") };
      const TCHAR* coatingTypeDescs[] = { _T("No coating"), _T("/A - Acrylic (Humiseal 1B31)"), _T("/P - Polyurethane (Conathane CE-1155)") };
      const TCHAR* pCoolingTypeDesc = _T("UNRECOGNIZED");
      const TCHAR* pCoatingTypeDesc = _T("UNRECOGNIZED");
      uint8_t modifications8, coolingType, coatingType;
      uint8_t customBuildCode, customBuildRevision;
      uint32_t si5338RefClkFreq, emcclkFreq;
      uint16_t dramFreq[4];
      uint8_t dramOrder[4];
      uint64_t dramSizeBytes;

      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%08X)\n"), modifications8, modifications8);
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      coolingType = GET_VPD_VAL8(&buffer[16]);
      pCoolingTypeDesc = (coolingType < ARRAY_LENGTH(coolingTypeDescs)) ? coolingTypeDescs[coolingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Cooling type                       %u (0x%02X) => %s\n"),
        (unsigned int)coolingType, (unsigned int)coolingType, pCoolingTypeDesc);
      coatingType = GET_VPD_VAL8(&buffer[17]);
      pCoatingTypeDesc = (coatingType < ARRAY_LENGTH(coatingTypeDescs)) ? coatingTypeDescs[coatingType] : _T("UNRECOGNIZED");
      _tprintf(_T("  Coating type                       %u (0x%02X) => %s\n"),
        (unsigned int)coatingType, (unsigned int)coatingType, pCoatingTypeDesc);
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[32]);
      _tprintf(_T("  Internal SI5338 ref. clock freq.   %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      emcclkFreq = GET_VPD_VAL32(&buffer[36]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      for (bank = 0; bank < 4; bank++) {
        dramFreq[bank] = GET_VPD_VAL16(&buffer[48 + 4 * bank]);
        dramOrder[bank] = GET_VPD_VAL8(&buffer[50 + 4 * bank]);
        _tprintf(_T("  DDR4 SDRAM bank %u max. frequency   %u (0x%04X) => %u MHz / DDR4-%u\n"),
          bank, (unsigned int)dramFreq[bank], (unsigned int)dramFreq[bank],
          (unsigned int)(dramFreq[bank] + 5) / 10, (unsigned int)(dramFreq[bank] + 2) / 5);
        dramSizeBytes = ((uint64_t)1 << dramOrder[bank]) /* no. of words */ * 8U; /* bytes per word */
        _tprintf(_T("  DDR4 SDRAM bank %u size             %u (0x%02X) => "),
          bank, (unsigned int)dramOrder[bank], (unsigned int)dramOrder[bank]);
        if (dramSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
        }
      }
      memcpy(fpgaStepping, &buffer[64], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[72], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[76]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[78]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
    }
    break;

  case DRAWING_NUM_ADMPA100:
  case DRAWING_NUM_ADMPA101:
    {
      uint8_t modifications8;
      uint8_t customBuildCode, customBuildRevision;
      uint32_t si5338RefClkFreq, emcclkFreq;
      uint16_t dramFreq[2];
      uint8_t dramOrder[2];
      uint64_t dramSizeBytes;

	  pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"), modifications8, modifications8);
      _tprintf(_T("                                     => Heatsink: %s\n"), (modifications8 & 0x1U) ? _T("2-SLOT") : _T("1-SLOT"));
      _tprintf(_T("                                     => Fans: %s\n"), (modifications8 & 0x2U) ? _T("FITTED") : _T("NOT FITTED"));
      _tprintf(_T("                                     => Top shroud: %s\n"), (modifications8 & 0x4U) ? _T("FITTED") : _T("NOT FITTED"));
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      si5338RefClkFreq = GET_VPD_VAL32(&buffer[32]);
      _tprintf(_T("  Internal SI5338 ref. clock freq.   %u (0x%08X) Hz\n"), si5338RefClkFreq, si5338RefClkFreq);
      emcclkFreq = GET_VPD_VAL32(&buffer[36]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      for (bank = 0; bank < 2; bank++) {
        dramFreq[bank] = GET_VPD_VAL16(&buffer[48 + 4 * bank]);
        dramOrder[bank] = GET_VPD_VAL8(&buffer[50 + 4 * bank]);
        _tprintf(_T("  DDR4 SDRAM bank %u max. frequency   %u (0x%04X) => %u MHz / DDR4-%u\n"),
          bank, (unsigned int)dramFreq[bank], (unsigned int)dramFreq[bank],
          (unsigned int)(dramFreq[bank] + 5) / 10, (unsigned int)(dramFreq[bank] + 2) / 5);
        dramSizeBytes = ((uint64_t)1 << dramOrder[bank]) /* no. of words */ * 8U; /* bytes per word */
        _tprintf(_T("  DDR4 SDRAM bank %u size             %u (0x%02X) => "),
          bank, (unsigned int)dramOrder[bank], (unsigned int)dramOrder[bank]);
        if (dramSizeBytes >= Gi) {
          _tprintf(_T("%u GiB\n"), (unsigned int)(dramSizeBytes / Gi));
        } else {
          _tprintf(_T("%u MiB\n"), (unsigned int)(dramSizeBytes / Mi));
        }
      }
      memcpy(fpgaStepping, &buffer[64], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[72], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[76]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[78]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
	}
	break;

  case DRAWING_NUM_ADCXMCSTANDALONE:
    {
      const TCHAR* ppTempGradeStrings[3] = { _T("C (Commercial)"), _T("I (Industrial)"), _T("E (Extended)") };
      const TCHAR* pTempGradeString = _T("UNRECOGNIZED");
      const TCHAR* ppXmcConnTypeStrings[3] = { _T("VITA 42 XMC"), _T("VITA 61 XMC2"), _T("VITA 88 XMC+") };
      const TCHAR* pXmcConnTypeString = _T("UNRECOGNIZED");
      uint8_t customBuildCode, customBuildRevision;
      uint8_t boardTempGrade;
      uint8_t modifications8, xmcConnType;

  	  customBuildCode = GET_VPD_VAL8(&buffer[12]);
      customBuildRevision = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
	    pcbRev = GET_VPD_VAL8(&buffer[14]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      boardTempGrade = GET_VPD_VAL8(&buffer[15]);
      if (boardTempGrade < ARRAY_LENGTH(ppTempGradeStrings)) {
        pTempGradeString = ppTempGradeStrings[boardTempGrade];
      }
      _tprintf(_T("  Board temperature grade            %u (0x%02X) => %s\n"),
        (unsigned int)boardTempGrade, (unsigned int)boardTempGrade, pTempGradeString);
      modifications8 = GET_VPD_VAL8(&buffer[16]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"), modifications8, modifications8);
      xmcConnType = (uint8_t)(modifications8 & 0x3);
      if (xmcConnType < ARRAY_LENGTH(ppXmcConnTypeStrings)) {
        pXmcConnTypeString = ppXmcConnTypeStrings[xmcConnType];
      }
      _tprintf(_T("                                     => XMC connector type: %s\n"), pXmcConnTypeString);
	  }
    break;

  case DRAWING_NUM_ADMPCIETRC01:
    {
      uint8_t customBuildCode, customBuildRevision, modifications8;

	    pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"), modifications8, modifications8);
  	  customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      memcpy(fpgaStepping, &buffer[32], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[40], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[44]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[46]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
	  }
    break;

  case DRAWING_NUM_ADMPCIE9V8:
  case DRAWING_NUM_ADMAD01474:
    {
      uint8_t modifications8;
      uint8_t customBuildCode, customBuildRevision;
      uint32_t emcclkFreq;

      pcbRev = GET_VPD_VAL8(&buffer[12]);
      _tprintf(_T("  PCB revision                       %u (0x%02X) => %u.%u\n"),
        (unsigned int)pcbRev, (unsigned int)pcbRev, (unsigned int)((pcbRev >> 4) & 0xFU), (unsigned int)(pcbRev & 0xFU));
      modifications8 = GET_VPD_VAL8(&buffer[13]);
      _tprintf(_T("  Modifications bitmask              %u (0x%02X)\n"), modifications8, modifications8);
      customBuildCode = GET_VPD_VAL8(&buffer[14]);
      customBuildRevision = GET_VPD_VAL8(&buffer[15]);
      _tprintf(_T("  Custom build code, revision        %u, %u => "),
        (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      if (0 == customBuildCode) {
        _tprintf(_T("Standard build\n"));
      } else {
        _tprintf(_T("/C%u.%u\n"),
          (unsigned int)customBuildCode, (unsigned int)customBuildRevision);
      }
      emcclkFreq = GET_VPD_VAL32(&buffer[20]);
      _tprintf(_T("  EMCCLK clock frequency             %u (0x%08X) Hz\n"), emcclkFreq, emcclkFreq);
      memcpy(fpgaStepping, &buffer[32], 8);
      fpgaStepping[8] = '\0';
      _tprintf(_T("  FPGA stepping                      \""));
      printf("%s", fpgaStepping);
      _tprintf(_T("\"\n"));
      memcpy(fpgaSpeedGrade, &buffer[40], 4);
      fpgaSpeedGrade[4] = '\0';
      _tprintf(_T("  FPGA speed grade                   \""));
      printf("%s", fpgaSpeedGrade);
      _tprintf(_T("\"\n"));
      fpgaDevice = GET_VPD_VAL16(&buffer[44]);
      _tprintf(_T("  FPGA device                        %u (0x%04X) => %s\n"),
        (unsigned int)fpgaDevice, (unsigned int)fpgaDevice, avr2utilGetFpgaDeviceString(fpgaDevice));
      fpgaTempGrade = GET_VPD_VAL8(&buffer[46]);
      _tprintf(_T("  FPGA temperature grade             %u (0x%02X) => %s\n"),
        (unsigned int)fpgaTempGrade, (unsigned int)fpgaTempGrade, avr2utilGetFpgaTempGradeString(fpgaTempGrade));
    }
    break;

  default:
    _tprintf(_T("WARNING: Unable to decode remaining data for UNRECOGNIZED drawing number.\n"));
    _tprintf(_T("         Use the 'display-vpd-raw' command in order to display raw VPD.\n"));
    break;
  }

  if (length >= 5 && length <= VPD_LENGTH_LIMIT) {
    unsigned int nCheck = length - 1U;

    checksum = GET_VPD_VAL8(&buffer[length - 1]);
    _tprintf(_T("  Data structure checksum            %u (0x%02X)\n"),
      (unsigned int)checksum, (unsigned int)checksum);
    calcChecksum = 0;
    for (i = 0; i < nCheck; i++) {
      calcChecksum = (uint8_t)(calcChecksum + GET_VPD_VAL8(&buffer[i]));
    }
    calcChecksum = (uint8_t)-calcChecksum;
    if (calcChecksum != checksum) {
      _tprintf(_T("*** ERROR: Calculated checksum %u 0x%02X does not match actual checksum %u (0x%02X)\n"),
        (unsigned int)calcChecksum, (unsigned int)calcChecksum, (unsigned int)checksum, (unsigned int)checksum);
    }
  } else {
    _tprintf(_T("*** ERROR: Unable to verify checksum because data structure length %u (0x%04X) is invalid\n"),
      (unsigned int)length, (unsigned int)length);
  }

  if (version > highestVersionRecognized) {
    _tprintf(_T("WARNING: Data structure version %u is not recognized by this version of avr2util.\n"), (unsigned int)version);
    _tprintf(_T("         Highest data structure version recognised is %u.\n"), (unsigned int)highestVersionRecognized);
    _tprintf(_T("         Some fields of the VPD might not be displayed.\n"));
  }

done:
  return ret;
}

int
avr2utilDisplayVpdRawCommand(
  const DeviceHandle* pDevice,
  bool_t bVerbose)
{
  const unsigned int vpdLengthLimit = VPD_LENGTH_LIMIT;
  uint8_t buffer[256]; /* Request VPD command requires buffer of 256 bytes */
  unsigned int i, j;
  int ret = EXIT_OK;

  memset(buffer, 0, 256);

  if (avr2utilIsInServiceMode(pDevice)) {
    /* In Service Mode, get VPD by reading AVR Flash */
    ret = avr2utilReadApplicationFlash(pDevice, bVerbose, VPD_BASE_ADDR, buffer, vpdLengthLimit);
  } else {
    /* In Application Mode, request VPD using command */
    ret = avr2utilRequestVpd(pDevice, bVerbose, buffer);
  }

  if (ret) {
    goto done;
  }

  _tprintf(_T("Raw VPD:\n"));
  _tprintf(_T("     "));
  for (i = 0; i < 16; i++) {
    _tprintf(_T(" +%X"), i);
  }
  _tprintf(_T("\n"));
  for (i = 0; i < vpdLengthLimit; i += 16) {
    _tprintf(_T("0x%02X:"), i);
    for (j = 0; j < 16; j++) {
      _tprintf(_T(" %02X"), buffer[i + j]);
    }
    _tprintf(_T(" "));
    for (j = 0; j < 16; j++) {
      if (isprint(buffer[i + j])) {
        printf("%c", buffer[i + j]);
      } else {
        printf(".");
      }
    }
    _tprintf(_T("\n"));
  }

done:
  return ret;
}
